%% Copyright 2014 Keith Clark, Peter Robinson
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Error display
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


'$display_goal_for_error'(OG, OG) :-
    !,
    writeL(stderr, [OG]).
'$display_goal_for_error'(OG, OGCopy) :-
    OG = (A = B), A == B, !,
    writeL(stderr, [OGCopy]).
'$display_goal_for_error'(OG, OGCopy) :-
    writeL(stderr, [OGCopy, " (equivalently ", OG, ")"]).

'$display_with_type'(F, _CallType) :-
    (F = '$$var$$' ; F = '$$var$$'(_)),
    !.
'$display_with_type'(_F, CallType) :-
    var(CallType), !,
    writeL(stderr, [nl_]).
'$display_with_type'(F, durative(CallType)) :-
    !,
    writeL(stderr, ["  ( with ", wr_(durative(F : CallType)), ")", nl_]).
'$display_with_type'(F, discrete(CallType)) :-
    !,
    writeL(stderr, ["  ( with ", wr_(discrete(F : CallType)), ")", nl_]).
'$display_with_type'(F, CallType) :-
    writeL(stderr, ["  ( with ", wr_(F : CallType), ")", nl_]).
    
%%'$display_error'(T) :- errornl('$display_error'(T)),fail.
'$display_error'(_) :-
    writeL(stderr, [nl_]),
    global_state_increment(num_errors, _), fail.

'$display_error'(no_type_error(OG, Context, OGCopy, ContextCopy)) :-
    writeL([no_type_error(OG, Context, OGCopy, ContextCopy), nl_]),
    fail.
'$display_error'(builtin_redefinition(F, Kind)) :-
    !,
    writeL(stderr, ["Error: ", F, " is a builtin ", Kind,
                    " and can't be redefined", nl_,nl_]),
    fail.
'$display_error'(type_as_term_error(Type, Clause)) :-
    !,
    writeL(stderr, ["Error: ", Type,
                    " is a type and can't be used as a term in", nl_,
                   Clause, nl_]),
    fail.
'$display_error'(not_mode_error(Vars, tr_prog, OG, OGCopy,'$tr'(Head, Rule))) :-
    !,
    writeL(stderr, ["Mode Error: ", Vars, " might not be ground when calling ",nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "of the rule", nl_, Rule, nl_]),
    writeL(stderr, ["of the TR procedure ", Head, nl_]).
'$display_error'(not_mode_error(Vars, Kind, OG, OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Mode Error: ", Vars, " might not be ground when calling ",nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "in the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(globals_locals_error(Var, Kind, OG, OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Error: The variable ", Var, " has been quantified earlier and so can't be reused in", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "in the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(locals_globals_error(Vars, Kind, OG, OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Error: The variables ", Vars, " are free variables and so",
                   nl_,"can't be used as quantified variables in", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "in the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(not_underscore_mode_error(Vars, Kind, OG, OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Mode Error: The variables ", Vars, " are required to be underscore variables in ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "in the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(forall_quant_vars_error(Vars, Kind, OG, OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Error: The variables ", Vars, " are required to appear both before and after the implication of ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "in the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(forall_nonlocal_quant_vars_error(Vars, Kind, OG, OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Error: The quantified variables ", Vars, " occur outside the quantification of ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "in the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(mode_error_list_constr(T, Vars, Term, Kind, OG,
					OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Error: The variables ", Vars, " in template term ", T, " of",nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "will not have or be given ground values", nl_,
                    "so the list ", Term, " will not be ground", nl_,
                    "in the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(mode_error('!', Arg, function, OG, OGCopy, ClauseCopy,
                            _CallType)) :-
    !,
    writeL(stderr, ["Mode Error: ", Arg, " will not be ground in the function result ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "of the function rule",nl_,ClauseCopy, nl_,
                    "as it is not in the rule head or given a ground value in the commit test", nl_]).
'$display_error'(mode_error('!', Arg, tr_prog, OG, OGCopy, '$tr'(Head, Rule),
                            _CallType)) :-
    OG = '$time'(T),
    OGCopy = '$time'(TCopy),
    !,
    writeL(stderr, ["Mode Error: ", Arg, " will not comply with mode !",nl_,
		    "as it might not be ground in the time expression ", nl_]),
    '$display_goal_for_error'(T, TCopy),
    writeL(stderr, [nl_, "of the rule", nl_, Rule, nl_]),
    writeL(stderr, ["of the TR procedure ", Head, nl_]).
'$display_error'(mode_error('!', Arg, tr_prog, OG, OGCopy, '$tr'(Head, Rule),
                            _CallType)) :-
    !,
    writeL(stderr, ["Mode Error: ", Arg, " will not comply with mode !",nl_,
		    "as it might not be ground when calling ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "of the rule", nl_, Rule, nl_]),
    writeL(stderr, ["of the TR procedure ", Head, nl_]).
 '$display_error'(mode_error('!', Arg, relation_query, OG, OGCopy, _ClauseCopy,
                            CallType)) :-
    !,
    functor(OG, F, _),
    writeL(stderr, ["Mode Error: ", Arg, " will not comply with mode !",nl_,
		    "as it might not be ground when calling ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    '$display_with_type'(F, CallType).
'$display_error'(mode_error('!', Arg, expression_query, OG, OGCopy, ClauseCopy,
                            CallType)) :-
    !,
    functor(OG, F, _),
    writeL(stderr, ["Mode Error: ", Arg, " will not comply with mode !",nl_,
		    "as it might not be ground when calling ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    '$display_with_type'(F, CallType),
    writeL(stderr,  [nl_,"of the query", nl_, ClauseCopy, nl_]).
'$display_error'(mode_error('!', Arg, _Kind, _OG, '$time'(_, _, _), 
                           '$tr'(Head, Rule), _CallType)) :-
    !,
    writeL(stderr, ["Mode Error: ", Arg, " will not comply with mode !",nl_,
		    "as it might not be ground on the left hand side of the rule ", nl_,
                    wr_(Rule), nl_]),
    writeL(stderr, ["of the TR procedure ", Head, nl_]).
'$display_error'(mode_error('!', Arg, Kind, OG, OGCopy, ClauseCopy,
                            CallType)) :-
    !,
    functor(OG, F, _),
    writeL(stderr, ["Mode Error: ", Arg, " will not comply with mode !",nl_,
		    "as it might not be ground when calling ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    '$display_with_type'(F, CallType),      
     writeL(stderr, ["of the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(mode_error(doubleq_bound, V, _Kind, Clause)) :-
    !,
    writeL(stderr, ["Mode Error: the variable ", V, " has mode ?? but is instantiated in",
                    nl_, Clause, nl_]).
'$display_error'(mode_error(unbound_by_body, Unbound, _Kind, Clause)) :-
    !,
    writeL(stderr, ["Mode Error: the variables ", Unbound,
		     " are required to be ground by the body of ", nl_,
		     Clause,
		     "but are not", nl_]).
'$display_error'(type_error(Term, Type, '$unknown', relation_query, OG, OGCopy,
                            _ClauseCopy, CallType)) :-
    !,
    functor(OG, F, _),
    writeL(stderr, ["Type Error: ", Term, " is not of type ", wr_(Type),
		    " in condition ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    '$display_with_type'(F, CallType).
'$display_error'(type_error(Term, Type, T, relation_query, OG, OGCopy,
                            _ClauseCopy, CallType)) :-
    functor(OG, F, _), atom(F),
    !,
    writeL(stderr, ["Type Error: ", Term, " has type ", wr_(T),
		    " but is required to be of type ", wr_(Type),
		    " in condition ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    '$display_with_type'(F, CallType).
'$display_error'(type_error(Term, Type, T, relation_query, OG, OGCopy,
                            _ClauseCopy, _CallType)) :-
    !,
    writeL(stderr, ["Type Error: ", Term, " has type ", wr_(T),
		    " but is required to be of type ", wr_(Type),
		    " in condition ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_]).
'$display_error'(type_error(Term, Type, '$unknown', function, OG, OGCopy,
                            ClauseCopy, _CallType)) :-
    !,
    writeL(stderr, ["Type Error: ", Term, " is not of type ", wr_(Type),
		    " in ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "of the function rule", nl_, ClauseCopy, nl_]).
'$display_error'(type_error(Term, Type, T, function, OG, OGCopy,
                            ClauseCopy, _CallType)) :-
    !,
    writeL(stderr, ["Type Error: ", Term, " has type ", wr_(T),
		    " but is required to be of type ", wr_(Type),
		    " in ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_, "of the function rule", nl_, ClauseCopy, nl_]).
'$display_error'(type_error(Term, Type, '$unknown', Kind, '$head'(_Head),
			    _OGCopy, ClauseCopy, _CallType)) :-
    !,
    writeL(stderr, ["Type Error: ", Term, " is not of type ", wr_(Type), nl_,
		    "in the head of the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(type_error(Term, Type, T, Kind, '$head'(_Head), _OGCopy,
                            ClauseCopy, _CallType)) :-
    !,
    writeL(stderr, ["Type Error: ", Term, " has type ", wr_(T),
		    " but is required to be of type ", wr_(Type), nl_,
		    "in the head of the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(type_error(Term, Type, '$unknown', _Kind, OG, OGCopy,
                            '$tr'(Head, Rule), CallType)) :-
    !,
    functor(OG, F, _),
    writeL(stderr, ["Type Error: ", Term, " is not of type ", wr_(Type),
		    " in condition ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    '$display_with_type'(F, CallType),
    writeL(stderr, ["of the rule", nl_, wr_(Rule), nl_,
                    "of the TR procedure ", Head, nl_]).
'$display_error'(type_error(Term, Type, '$unknown', Kind, OG, OGCopy,
                            ClauseCopy, CallType)) :-
    !,
    functor(OG, F, _),
    writeL(stderr, ["Type Error: ", Term, " is not of type ", wr_(Type),
		    " in condition ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    '$display_with_type'(F, CallType),
    writeL(stderr, ["of the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(type_error(Term, Type, T, expression_query, OG, OGCopy,
                            _ClauseCopy, _CallType)) :-
    !,
    writeL(stderr, ["Type Error: ", Term, " has type ", wr_(T),
		    " but is required to be of type ", wr_(Type),
		    " in the expression", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    writeL(stderr, [nl_]).
'$display_error'(type_error(Term, Type, T1, tr_prog, OG, OGCopy,
                            '$tr'(Head, Rule), _CallType)) :-
    OG = '$time'(T),
    OGCopy = '$time'(TCopy),
    !,
    writeL(stderr, ["Type Error: ", Term, " has type ", wr_(T1),
		    " but is required to be of type ", wr_(Type), nl_,
		    "in the the time expression", nl_]),
    '$display_goal_for_error'(T, TCopy),
    writeL(stderr,  [nl_, "of the rule", nl_, wr_(Rule), nl_]),
    writeL(stderr, ["of the TR procedure ", Head, nl_]).
'$display_error'(type_error(Term, Type, T1, tr_prog, OG, OGCopy,
                            '$tr'(Head, Rule), _CallType)) :-
    OG = '$repeat'(T),
    OGCopy = '$repeat'(TCopy),
    !,
    writeL(stderr, ["Type Error: ", Term, " has type ", wr_(T1),
		    " but is required to be of type ", wr_(Type), nl_,
		    "in the the wait-repeat expression", nl_]),
    '$display_goal_for_error'(T, TCopy),
    writeL(stderr,  [nl_, "of the rule", nl_, wr_(Rule), nl_]),
    writeL(stderr, ["of the TR procedure ", Head, nl_]).
'$display_error'(type_error(Term, Type, T, _, OG, OGCopy,
                            '$tr'(Head, Rule), CallType)) :-
    !,
    functor(OG, F, _),
    writeL(stderr, ["Type Error: ", Term, " has type ", wr_(T),
		    " but is required to be of type ", wr_(Type),
		    " in  ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    '$display_with_type'(F, CallType),
    writeL(stderr, ["of the rule ", nl_, wr_(Rule), nl_]),
    writeL(stderr, ["of the TR procedure ", Head, nl_]).
'$display_error'(type_error(Term, Type, T, Kind, OG, OGCopy,
                            ClauseCopy, CallType)) :-
    !,
    functor(OG, F, _),
    writeL(stderr, ["Type Error: ", Term, " has type ", wr_(T),
		    " but is required to be of type ", wr_(Type),
		    " in  ", nl_]),
    '$display_goal_for_error'(OG, OGCopy),
    '$display_with_type'(F, CallType),
    writeL(stderr, [" of the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(inconsistent_unify_type_error(Term, Type1, Type2, _,
					       ClauseCopy)) :-
    !,
    writeL(stderr, ["Type Error: ",  Term,
                    " cannot be both of type ",
                    wr_(Type1)," and ", wr_(Type2),
                    " in", nl_, ClauseCopy, nl_]).
'$display_error'(inconsistent_type_error(Term, Type1, Type2, relation_query, OG,
                                         OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Type Error:  argument ",  Term,
                    " cannot be both of type ",
                    wr_(Type1)," and ", wr_(Type2),nl_,
                    "in "]),
    '$display_goal_for_error'(OG, OGCopy),
      writeL(stderr, [" of the query", nl_, ClauseCopy, nl_]).
'$display_error'(inconsistent_type_error(Term, Type1, Type2, expression_query, OG,
                                         OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Type Error:  argument ",  Term,
                    " cannot be both of type ",
                    wr_(Type1)," and ", wr_(Type2),nl_,
                    "in "]),
    '$display_goal_for_error'(OG, OGCopy),
      writeL(stderr, [" of the query", nl_, ClauseCopy, nl_]).
'$display_error'(inconsistent_type_error(Term, Type1, Type2, Kind, OG,
                                         OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Type Error:  argument ",  Term,
                    " cannot be both of type ",
                    wr_(Type1)," and ", wr_(Type2),nl_,
                    "in "]),
    '$display_goal_for_error'(OG, OGCopy),
      writeL(stderr, [" of the ", Kind, " rule", nl_, ClauseCopy, nl_]).
/*
  '$display_error'(inconsistent_types_error(Term1, Term2, Kind, OG,
                                          OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Type Error:  terms ",  wr_(Term1), 
                    " cannot be both of type ",
                    wr_(Type1)," and ", wr_(Type2),nl_," and ", wr_(Term2),nl_,
                    " have incompatible types in "]),
    '$display_goal_for_error'(OG, OGCopy),
      writeL(stderr, [" of the ", Kind, " rule", nl_, ClauseCopy, nl_]).
*/
'$display_error'(inconsistent_type_error_head(Term, Type1, Type2, Kind,
					      ClauseCopy)) :-
    !,
    writeL(stderr, ["Type Error: ",  Term,
                    " cannot be both of type ",
                    wr_(Type1)," and ", wr_(Type2),nl_,
                    "in the head of the ", Kind, " rule", nl_, ClauseCopy, nl_]).
'$display_error'(head_output_type_error(Var, T, Type, _, '$head'(_), Kind,
					_OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Type Error:  when ",  Var:Type,
		    " is being generated",nl_,
		    "it might be assigned a value of type ",
		     T, " in the head"]),
    writeL(stderr, [nl_, "but is not followed by a call ", type(Var, Type),
		    " in the ",
		    Kind,
		    " rule", nl_, ClauseCopy,nl_]).
'$display_error'(head_output_type_error(Var, T, Type, '+?', _OG, _Kind,
					_OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Type Error:  when ",  Var:Type,
		    " of the head is being generated",nl_,
		    "it might be assigned a value of type ",
		     T, " in the body of", nl_,  ClauseCopy,nl_]).
'$display_error'(head_output_type_error(Var, T, Type, _Mode, '$head'(_), Kind,
					_OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Type Error:  when ",  Var, ":", Type,
		    " is not given", nl_,
		    "it will be assigned a value of type ",
		    T, " in the head."]),
    writeL(stderr, [nl_, "The body of the ", Kind, " rule", nl_,
		   ClauseCopy,nl_, "requires a call ",
		    type(Var, Type), nl_]).
'$display_error'(head_output_type_error(Var, T, Type, _Mode, OG, Kind,
					OGCopy, ClauseCopy)) :-
    !,
    writeL(stderr, ["Type Error:  when ",  Var, ":", Type,
		    " is being generated", nl_,
		    "it might be assigned a value of type ",
		    T, " in "]),
    '$display_goal_for_error'(OG, OGCopy), 
    writeL(stderr, [nl_, "but is not immediately followed by a call ",
		    type(Var, Type),
		    " in the ",
		    Kind,
		    " rule", nl_, ClauseCopy,nl_]).
		    
'$display_error'(handle_message_wrong_kind) :-
    !,
    writeL(stderr, ["Error: handle_message is not declared as an action ",
                    nl_]),
    fail.
'$display_error'(handle_message_wrong_type(Type, Vars)) :-
    !,
    thaw_term(Type),
    '$vars2names'(Type, Vars),
    writeL(stderr, ["Error: handle_message must be declared with type", nl_,
                    "([term?, handle]) ~>> - it has been declared as type",
                    nl_, Type, nl_]),
    fail.
'$display_error'(handle_percepts_wrong_kind) :-
    !,
    writeL(stderr, ["Error: handle_percepts is not declared as an action ",
                    nl_]),
    fail.
'$display_error'(handle_percepts_wrong_type(Type, Vars)) :-
    !,
    thaw_term(Type),
    '$vars2names'(Type, Vars),
    writeL(stderr, ["Error: handle_percepts must be declared with type", nl_,
                    "([percept_message_type]) ~>> - it has been declared as type",
                    nl_, Type, nl_]),
    fail.
'$display_error'(controls_wrong_kind) :-
    !,
    writeL(stderr, ["Error: mod_controls is not declared as a relation ",
                    nl_]),
    fail.
'$display_error'(controls_wrong_type(Type, Vars)) :-
    !,
    thaw_term(Type),
    '$vars2names'(Type, Vars),
    writeL(stderr, ["Error: mod_controls must be declared with type", nl_,
                    "([control_action, control_action]) ~>> - it has been declared as type",
                    nl_, Type, nl_]),
    fail.
/*
'$display_error'(controls_no_defn) :-
    !,
    writeL(stderr, ["Error: controls must be declared with type", nl_,
                    "([control_action]) ~>>", nl_,
                    "and given a definition.", nl_]),
    fail.
*/
'$display_error'(must_be_variable(G1, G, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", G1 , " is not a variable in ", G,
		    " of ", nl_, Clause, nl_]),
    fail.
  

'$display_error'(kind_mismatch(Clause, Vars)) :-
    !,
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: The definition", nl_, Clause, nl_,
                    "does not match the declaration", nl_]).
'$display_error'(tr_head_nonvar(A, Head, Vars)) :-
    !,
    '$vars2names'(Head, Vars),
    writeL(stderr, ["Error: the argument ", A, " of the head ", nl_,
		    Head,
		    " of the TR procedure is not a (unique) variable", nl_]).
'$display_error'(relhead_mode(_Mode, Vars, FVars, IF(Head, OrigBody))) :-
    !,
      thaw_term((Head, OrigBody)),
      '$vars2names'((Head, OrigBody), Vars),
    writeL(stderr, ["Mode Error: the variables ", FVars,
		     " are required to be ground by the body of ", nl_,
		     IF(Head, OrigBody), nl_,
		     "but are not", nl_]),
    freeze_term(Vars).
'$display_error'(relhead0_mode(_Mode, Vars, FVars, Head)) :-
    !,
    (
      thaw_term(Vars),
      '$vars2names'(Head, Vars),
      writeL(stderr, ["Mode Error: the variables ", FVars,
                      " occur in non-input arguments in the fact", nl_,
                      Head, nl_,
                      "but is unbound", nl_]),
      fail
    ;
      true
    ).

'$display_error'(tr_while_until_mode(Vars,
                             '$tr'(Head, Rule), FVars, _OG)) :-
    !,
    (
      thaw_term((Vars, Head, Rule)),
      '$vars2names'((Head, Rule), Vars),
      writeL(stderr, ["Mode Error: the variables ", FVars,
                      " occur unbound on the left hand side of the rule ", nl_,
                      Rule, nl_,
                      "of the TR procedure ", Head, nl_]),
      fail
    ;
      true
    ).
'$display_error'(tr_act_mode(rhs, Vars,
                             '$tr'(Head, Rule), FVars, _OG)) :-
    !,
    (
      thaw_term((Vars, Head, Rule)),
      '$vars2names'((Head, Rule), Vars),
      writeL(stderr, ["Mode Error: the variables ", FVars,
                      " occur unbound on the right hand side of the rule ", nl_,
                      Rule, nl_,
                      "of the TR procedure ", Head, nl_]),
      fail
    ;
      true
    ).
'$display_error'(tr_act_mode(input, Vars,
                             '$tr'(Head, Rule), FVars, OG)) :-
    !,
    (
      thaw_term((Vars, Head, Rule)),
      '$vars2names'((Head, Rule), Vars),
      writeL(stderr, ["Mode Error: the variables ", FVars,
                      " occur unbound in the goal ",
                      nl_, OG, " of the rule ", nl_, Rule, nl_,
                      "of the TR procedure ", Head, nl_]),
      fail
    ;
      true
    ).
'$display_error'(rel_act_mode(unbound(T), RelAct,
                              Vars, IF(Head, OrigBody), _, FVars, OG)) :-
    !,
    (
      thaw_term((Vars, OrigBody)),
      '$vars2names'((Head, OrigBody), Vars),
      writeL(stderr, ["Mode Error: the variables ", FVars,
                      " occur unbound in argument ", T, " in the goal ",
                      nl_, OG, nl_, "of the ", RelAct , " rule", nl_,
                      IF(Head, OrigBody), nl_]),
      fail
    ;
      true
    ).
'$display_error'(rel_act_mode(input(T), RelAct,
                              Vars, IF(Head, OrigBody), _, FVars, OG)) :-
    !,
    (
      thaw_term((Vars, OrigBody)),
      '$vars2names'((Head, OrigBody), Vars),
      writeL(stderr, ["Mode Error: the variables ", FVars,
                      " occur unbound in input argument ", T, " in the goal ",
                      nl_, OG, nl_, "of the ", RelAct, " rule" , nl_,
                      IF(Head, OrigBody), nl_]),
      fail
    ;
      true
    ).
'$display_error'(rel_act_mode(output(T), RelAct,
                              Vars, IF(Head, OrigBody), _, _FVars, OG)) :-
    !,
    (
      thaw_term((Vars, OrigBody)),
      '$vars2names'((Head, OrigBody), Vars),
      writeL(stderr, ["Mode Error: the input argument ", T,
                      " is not a variable in the goal ",
                      nl_, OG, nl_, "of the ", RelAct, " rule" , nl_,
                      IF(Head, OrigBody), nl_]),
      fail
    ;
      true
    ).
'$display_error'(rel_act_mode(output(T), RelAct,
                              Vars, IF(Head, OrigBody), _, _FVars, OG)) :-
    !,
    (
      thaw_term((Vars, OrigBody)),
      '$vars2names'((Head, OrigBody), Vars),
      writeL(stderr, ["Mode Error: the input argument ", T,
                      " is not a variable in the goal ",
                      nl_, OG, nl_, "of the ", RelAct, " rule" , nl_,
                      IF(Head, OrigBody), nl_]),
      fail
    ;
      true
    ).
'$display_error'(rel_act_mode(function_call(F), RelAct,
                              Vars, IF(Head, OrigBody), _, FVars, OG)) :-
    !,
    (
      thaw_term((Vars, OrigBody)),
      '$vars2names'((Head, OrigBody), Vars),
      writeL(stderr, ["Mode Error: the variables ", FVars,
                      " are not bound in the higher-order function call ",
                      F, " of",
                      nl_, OG, nl_, "of the ", RelAct, " rule" , nl_,
                      IF(Head, OrigBody), nl_]),
      fail
    ;
      true
    ).
'$display_error'(rel_act_mode(Mode, RelAct, Vars, IF(Head, OrigBody), _, FVars, OG)) :-
    !,
    (
      thaw_term((Vars, OrigBody)),
      '$vars2names'((Head, OrigBody), Vars),
     errornl(vvv___(OG)),
      writeL(stderr, ["Mode Error: ", FVars, " is not of mode ",
                      Mode,
                      " in the goal ",
                      nl_, OG, nl_, "of the ", RelAct, " rule" , nl_,
                      IF(Head, OrigBody), nl_]),
      fail
    ;
      true
    ).
'$display_error'(rel_query_inout_mode(Vars, Kind, _, A, OG)) :-
    !,
    (
      '$do_unifs'(Vars),
      writeL(stderr, ["Mode Error: the argument ", A,
                      " is neither a variable nor a ground term in ",
                      nl_, "an annotated argument in the goal ",
                               nl_, OG, nl_, "of the query ", nl_,
                               Kind, nl_]),
      fail
    ;
      true
    ).
'$display_error'(rel_query_mode(Kind, Vars, FVars, OG)) :-
    !,
      '$do_unifs'(Vars),
      writeL(stderr, ["Mode Error: the variables ", FVars,
                               " occur unbound in input arguments in the goal ",
                               nl_, OG, nl_, "of the query ", nl_,
                               Kind, nl_]).
'$display_error'(expression_query_mode(Kind, Vars, FVars, _OG)) :-
    !,
      '$do_unifs'(Vars),
      writeL(stderr, ["Mode Error: the variables ", FVars,
                               " occur unbound in input arguments in the expression ", nl_,
                               Kind, nl_]).
'$display_error'(function_rel_mode(Vars, Kind, FVars, OG)) :-
    !,
    (
      thaw_term(Kind),
      '$vars2names'(Kind, Vars),
      writeL(stderr, ["Mode Error: the variables ", FVars,
                               " occur unbound in input arguments in the goal ",
                               nl_, OG, nl_, "of the function ", nl_,
                               Kind, nl_]),
      fail
    ;
      true
    ).

'$display_error'(function_mode(Vars, Orig, FVars)) :-
    !,
    (
      thaw_term((Vars, Orig)),
      '$vars2names'(Orig, Vars),
      writeL(stderr, ["Mode Error: the variables ", FVars,
                               " occur unbound in input arguments of ",nl_,
                               "the body of the function ", nl_,
                               Orig, nl_]),
      fail
    ;
      true
    ).

'$display_error'(tr_rule_type(Vars, Index, Head, OrigRule)) :-
    !,
    (
      thaw_term((Head, OrigRule)),
      '$vars2names'((Head, OrigRule), Vars),
      writeL(stderr, ["Type Error: rule ", nl_, OrigRule, nl_,
                               "at index ", Index, " of the TR program ",
                               Head,nl_,
                               "has a type error",nl_]),
      fail
    ;
      true
    ).
'$display_error'(rel_type(Vars, OrigRule, OG)) :-
    !,
    (
      thaw_term(OrigRule),
      '$vars2names'(OrigRule, Vars),
      writeL(stderr, ["Type Error: in argument of ", OG, nl_,
                               " in relation clause ",
                               nl_, OrigRule, nl_]),
      fail
    ;
      true
    ).
'$display_error'(relhead_type(Thawed, OrigRule)) :-
    !,
    writeL(stderr, ["Type Error: in argument of the head of relation clause ",
                             nl_, OrigRule, nl_]),
    freeze_term(Thawed).
'$display_error'(acthead_type(Thawed, OrigRule)) :-
    !,
    writeL(stderr, ["Type Error: in argument of the head of action clause ",
                             nl_, OrigRule, nl_]),
    freeze_term(Thawed).
'$display_error'(type_type_error(TypeTest, Thawed, OrigRule, Term, Type)) :-
    !,
     writeL(stderr, ["Type Error: argument ",  Term, " cannot be of type ",
			      wr_(Type),nl_,
			      " in ", TypeTest, " of",
			      nl_, OrigRule, nl_]),
    freeze_term(Thawed).
'$display_error'(type_type_query_error(TypeTest, Thawed, Term, Type)) :-
    !,
     writeL(stderr, ["Type Error: argument ",  Term, " cannot be of type ",
			      wr_(Type), " in ", TypeTest, nl_]),
    freeze_term(Thawed).
'$display_error'(relhead_term_type(relation, Thawed, OrigRule, Term, Type)) :-
    !,
    (
     ground(Type)
    ->
     writeL(stderr, ["Type Error: argument ",  Term, " is not of type ",
			      wr_(Type),nl_,
			      " in the head of relation clause ",
			      nl_, OrigRule, nl_])
    ;
     writeL(stderr, ["Type Error: argument ",  Term,
			     " is not of polymorphic type ",
                             wr_(Type),nl_,
			     " in the head of relation clause ",
                             nl_, OrigRule, nl_])
    ),
    freeze_term(Thawed).
'$display_error'(relhead_term_type(action, Thawed, OrigRule, Term, Type)) :-
    !,
    (
     ground(Type)
    ->
     writeL(stderr, ["Type Error: argument ",  Term, " is not of type ",
			      wr_(Type),nl_,
			      " in the head of action clause ",
			      nl_, OrigRule, nl_])
    ;
     writeL(stderr, ["Type Error: argument ",  Term,
			     " is not of polymorphic type ",
                             wr_(Type),nl_,
			     " in the head of action clause ",
                             nl_, OrigRule, nl_])
    ),
    freeze_term(Thawed).
'$display_error'(relhead_term_type(belief, Thawed, OrigRule, Term, Type)) :-
    !,
     writeL(stderr, ["Type Error: argument ",  Term, " is not of type ",
			      wr_(Type),nl_,
			      " in the belief/percept relation",
			      nl_, OrigRule, nl_]),

    freeze_term(Thawed).
'$display_error'(tr_term_type(Thawed, '$tr'(Head, Rule), Term, Type, Pred)) :-
    !,
    (
      Pred = '$time_'(_T)
    ->
      writeL(stderr, ["Type Error:  argument ",  Term, " is not of type ",
                      wr_(Type),nl_,
                      " in a time constraint of the rule ", nl_, Rule,nl_,
                      "of the TR procedure ", Head,nl_])
    ;
      writeL(stderr, ["Type Error:  argument ",  Term, " is not of type ",
                      wr_(Type),nl_,
                      " in the body condition ", Pred,
                      " of the rule ", nl_, Rule,nl_,
                      "of the TR procedure ", Head,nl_])
    ),
      freeze_term(Thawed).
'$display_error'(tr_action_term_type(Thawed, '$tr'(Head, Rule),
				     Term, Type, Pred)) :-
      writeL(stderr, ["Type Error:  argument ",  Term, " is not of type ",
                      wr_(Type),nl_,
                      " in the action ", Pred,
                      " of the rule ", nl_, Rule,nl_,
                      "of the TR procedure ", Head,nl_]),
      freeze_term(Thawed).
    
'$display_error'(tr_goal_term_type(Thawed, '$tr'(Head, Rule), Term, Type, Pred)) :-
    !,
    writeL(stderr, ["Type Error:  argument ",  Term, " is not of type ",
                    wr_(Type),nl_,
                    " in the action goal ", Pred,
                    " of the rule ", nl_, Rule,nl_,
                    "of the TR procedure ", Head,nl_]),
    freeze_term(Thawed).
'$display_error'(tr_goal_term_type(Thawed, '$tr'(Head, Rule), Term, Type, Pred)) :-
    !,
    writeL(stderr, ["Type Error:  argument ",  Term, " is not of type ",
                    wr_(Type),nl_,
                    " in the action goal ", Pred,
                    " of the rule ", nl_, Rule,nl_,
                    "of the TR procedure ", Head,nl_]),
    freeze_term(Thawed).
'$display_error'(tr_ts_time_type(Thawed, '$tr'(Head, Rule), Term, Type)) :-
    !,
    writeL(stderr, ["Type Error: ",  Term, " is not of type ",
                    wr_(Type),nl_,
                    " in the timed sequence on the right hand side of the rule ", nl_, Rule,nl_,
                    "of the TR procedure ", Head,nl_]),
    freeze_term(Thawed).
'$display_error'(wait_repeat_time_type(Thawed, '$tr'(Head, Rule),
                                       Term, Type)) :-
    !,
    writeL(stderr, ["Type Error: ",  Term, " is not of type ",
                    wr_(Type),nl_,
                    " in the wait/repeat on the right hand side of the rule ",
                    nl_, Rule,nl_,
                    "of the TR procedure ", Head,nl_]),
    freeze_term(Thawed).
'$display_error'(wait_time_type(Thawed, '$tr'(Head, Rule),
                                       Term, Type)) :-
    !,
    writeL(stderr, ["Type Error: ",  Term, " is not of type ",
                    wr_(Type),nl_,
                    " in the wait on the right hand side of the rule ",
                    nl_, Rule,nl_,
                    "of the TR procedure ", Head,nl_]),
    freeze_term(Thawed).
'$display_error'(rel_term_type(Thawed, OrigRule, Term, Type, Pred)) :-
    !,
    (
     var(Type)
    ->
    writeL(stderr, ["Type Error: cannot type argument ",  Term, nl_,
			     " in the body condition ", Pred,
                             " of relation clause ",
                             nl_, OrigRule, nl_])
     ;
    writeL(stderr, ["Type Error:  argument ",  Term, " is not of type ",
                             wr_(Type),nl_,
			     " in the body condition ", Pred,
                             " of relation clause ",
                             nl_, OrigRule, nl_])
    ),
    freeze_term(Thawed).
'$display_error'(action_term_type(Thawed, OrigRule, Term, Type, Pred)) :-
    !,
    (
     var(Type)
    ->
    writeL(stderr, ["Type Error: cannot type argument ",  Term, nl_,
			     " in the body condition ", Pred,
                             " of action clause ",
                             nl_, OrigRule, nl_])
     ;
    writeL(stderr, ["Type Error: argument ",  Term, " is not of type ",
                             wr_(Type),nl_,
			     " in the body condition ", Pred,
                             " of action clause ",
                             nl_, OrigRule, nl_])
    ),
    freeze_term(Thawed).
'$display_error'(function_rel_type(Thawed, OrigRule, Term, Type, Pred)) :-
    !,
    (
     var(Type)
    ->
    writeL(stderr, ["Type Error: cannot type argument ",  Term, nl_,
			     " in the condition ", Pred,
                             " of the function definition ",
                             nl_, OrigRule, nl_])
     ;
    writeL(stderr, ["Type Error: argument ",  Term, " is not of type ",
                             wr_(Type),nl_,
			     " in the condition ", Pred,
                             " of the function definition ",
                             nl_, OrigRule, nl_])
    ),
    freeze_term(Thawed).
'$display_error'(function_term_type(Thawed, Rule, Term, Type)) :-
    !,
    (
     var(Type)
    ->
    writeL(stderr, ["Type Error: cannot type ",  Term, nl_,
			     " in the body of the function definition",
                             nl_, Rule, nl_])
     ;
     ground(Type)
    ->
    writeL(stderr, ["Type Error: ",  Term, " is not of type ",
                             wr_(Type),nl_,
			     " in the body of the function definition",
                             nl_, Rule, nl_])
    ;
    writeL(stderr, ["Type Error: ",  Term,
			     " is not of polymorphic type ",
                             wr_(Type),nl_,
			     " in the body of the function definition",
                             nl_, Rule, nl_])
      
    ),
    freeze_term(Thawed).

'$display_error'(funhead_type(Thawed, Rule, Term, Type)) :-
    writeL(stderr, ["Type Error: ",  Term, " is not of type ",
                             wr_(Type),nl_,
			     " in the head of function definition ",
                             nl_, Rule, nl_]),
    freeze_term(Thawed).

'$display_error'(relation_query_type(Query, Vars, _Pred, T, TT)) :-
    '$do_unifs'(Vars),
    writeL(stderr, ["Query type error: ", T, " is not of type ",
                             wr_(TT),nl_,
			     "in the query",
                             nl_, Query, nl_]).
'$display_error'(expression_query_type(Query, Vars, _Pred, T, TT)) :-
    '$do_unifs'(Vars),
    writeL(stderr, ["Query type error: ", T, " is not of type ",
                             wr_(TT),nl_,
			     "in the expression",
                             nl_, Query, nl_]).

'$display_error'(relation_query(Query)) :-
    writeL(stderr, ["Failure for ", Query, nl_, nl_]).

'$display_error'(overlapping_types(TName, RName)) :-
    !,
    writeL(stderr, ["Error: Types ", TName, " and ", RName, " overlap", nl_,
		    "Consider splitting the types into non-overlapping types and using type union", nl_]).
'$display_error'(invalid_type(Clause, Vars, T)) :-
    !,
    thaw_term(Clause),
   '$vars2names'(Clause, Vars),
     writeL(stderr, ["Error: ",T, " is an invalid type in ", nl_, Clause,nl_]),
     fail.
'$display_error'(invalid_type_overlapping_constr(Clause, Vars, _T,
						 Clause1, Vars1)) :-
    !,
    thaw_term(Clause),
   '$vars2names'(Clause, Vars),
    thaw_term(Clause1),
   '$vars2names'(Clause1, Vars1),
   Clause = '::='(T, _),
   Clause1 = '::='(T1, _),
     writeL(stderr, ["Error: ",T, " and ", T1, " overlap", nl_,
		    "Consider splitting the types into non-overlapping types and using type union", nl_]),
     fail.
'$display_error'(invalid_constructor_type(Clause, Vars, T)) :-
    !,
    thaw_term(Clause),
   '$vars2names'(Clause, Vars),
     writeL(stderr, ["Error: ",T, " is an invalid constructor in the enumeration", nl_, Clause,nl_]),
     fail.
'$display_error'(invalid_atom_enum_type(Clause, Vars, T)) :-
    !,
    thaw_term(Clause),
   '$vars2names'(Clause, Vars),
     writeL(stderr, ["Error: ",T, " is a type name in the enumeration", nl_, Clause,nl_]),
     fail.

'$display_error'(task_atomic_prim_check(_, Rule, TR)) :-
    !,
    writeL(stderr, ["Error: The action of the rule ", nl_,
                             Rule, nl_,
                             "of " , TR,
                             " is primitive and is called outside a task_atomic", nl_]).
'$display_error'(valid_tr_guard(_Vars, Index, TR)) :-
    !,
     writeL(stderr, ["Error: Rule ", Index, " of the TR program ",
                              TR, " contains an invalid guard ", nl_]).
'$display_error'(valid_tr_RHS(_, Index, TR)) :-
    !,
     writeL(stderr, ["Error: Rule ", Index, " of the TR program ",
                              TR, " contains an invalid action ", nl_]).
'$display_error'(valid_rel(Vars, Clause,CClause)) :-
    !,
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: The clause ", nl_]),
    (Clause = If(Head,Body) -> show_clause(If,Head,Body) ; writeq(Clause)), 
    writeL(stderr, [ "\n contains an invalid goal. Was compiled to: ",
                              nl_, CClause, nl_]).
'$display_error'(valid_rel_query(G, Vars)) :-
    '$do_unifs'(Vars),
    writeL(stderr, [w(G), " is an invalid goal",nl_]), !,
    fail.
'$display_error'(valid_expression_query(G, Vars)) :-
    '$do_unifs'(Vars),
    writeL(stderr, [G, " is an invalid expression",nl_]), !,
    fail.
'$display_error'(goal_mode(Vars, Goal, _, FVars, OG)) :-
    !,
    thaw_term(Goal),      
    '$vars2names'(Goal, Vars),
    writeL(stderr, ["Mode Error: the variables ", FVars,
             " occur unbound in input arguments in the goal ", nl_,
             OG, nl_]),
    fail.
/*
'$display_error'(eval_in_head('$tr_prog'(Head, _Rules), Clause, Vars)) :-
    !,
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Evaluation Error: the function ", A,
                             " occurs in the head ", Head, " of the TR definition",nl_,
                             Clause,nl_]),
    fail.
*/
'$display_error'(eval_in_head(A, Clause, Vars)) :-
    !,
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Evaluation Error: the function ", A,
                             " occurs in the head of the definition",nl_,
                             Clause,nl_]),
    fail.
'$display_error'(undeclared_relation(Pred, relation(Clause, Vars))) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    functor(Pred, P, N),
    writeL(stderr, ["Error: the goal ", P/N,
                             " is undeclared in", nl_,
                             Clause,nl_]),
    fail.
'$display_error'(bad_expression(Term, '$tr'(Head, Rule), Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    functor(Term, P, N),
    writeL(stderr, ["Error: ", P/N,
                             " is undeclared in the rule", nl_]),
    write_a_tr_rule(stderr, Rule),
    writeL(stderr, [nl_, "of the TR procedure ", Head, nl_]),
    (
      '$get_ultimate_functor'(Term, R),
      '$type_info'(R, Kind, _, _)
    ->
      writeL(stderr, ["However, there is a ", Kind, " declaration for ", R, nl_])
    ;
      true
    ),
    fail.
'$display_error'(bad_expression(Term, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    functor(Term, P, N),
    writeL(stderr, ["Error: ", P/N,
                             " is undeclared in", nl_,
                             Clause,nl_]),
    (
      '$get_ultimate_functor'(Term, R),
      '$type_info'(R, Kind, _, _, _)
    ->
      writeL(stderr, ["However, there is a ", Kind, " declaration for ", R, nl_])
    ;
      true
    ),
    fail.
'$display_error'(bad_show(Show, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", Show, " is not allowed in ", nl_, Clause, nl_]),
    fail.
    
'$display_error'(bad_goal(Term, '$tr'(Head, Rule), Vars)) :-
    !,
    thaw_term((Head,Rule)),
    '$vars2names'((Head,Rule), Vars),
    writeL(stderr, ["Error: ", Term,
                             " is undeclared in rule", nl_,
                             Rule,nl_, "of the TR procedure ", Head,nl_]),
    (
      '$get_ultimate_functor'(Term, R),
      '$type_info'(R, Kind, _, RT, _),
      functor(RT, R, N)
    ->
      writeL(stderr, ["However, there is a ", Kind, " declaration for ", R/N, nl_])
    ;
      true
    ),
    
    fail.
'$display_error'(bad_goal(Term, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", Term,
                             " is undeclared in", nl_,
                             Clause,nl_]),
    fail.
'$display_error'(bad_belief_goal(Term, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", Term,
                             " is not a declared belief/percept relation in", nl_,
                             Clause,nl_]),
    fail.
'$display_error'(bad_belief_global_goal(Term, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", Term,
                             " is a global value in", nl_,
                             Clause,nl_]),
    fail.
'$display_error'(bad_discrete_goal(Term, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", Term,
                             " is a discrete action in", nl_,
                             Clause,nl_]),
    fail.
'$display_error'(bad_durative_goal(Term, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", Term,
                             " is a durative action in", nl_,
                             Clause,nl_]),
    fail.
'$display_error'(bad_tr_prog_goal(Term, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", Term,
                             " is a TR procedure in", nl_,
                             Clause,nl_]),
    fail.
'$display_error'(non_action_in_tr_goal(G, '$tr'(Head, Rule), Vars)) :-
    !,
    thaw_term((Head,Rule)),
    '$vars2names'((Head,Rule), Vars),
    writeL(stderr, ["Error: ", G,
                             " occuring after ++ is not an action in", nl_,
                             Rule,nl_, "of the TR procedure ", Head,nl_]),
    fail.
'$display_error'(non_simple_action_in_tr_goal(G, '$tr'(Head, Rule), Vars)) :-
    !,
    thaw_term((Head,Rule)),
    '$vars2names'((Head,Rule), Vars),
    writeL(stderr, ["Error: ", G,
                             " occuring after ++ is not a simple action in", nl_,
                             Rule,nl_, "of the TR procedure ", Head,nl_]),
    fail.
'$display_error'(non_dis_dur_in_tr_actions(G, '$tr'(Head, Rule), Vars)) :-
    !,
    thaw_term((Head,Rule)),
    '$vars2names'((Head,Rule), Vars),
    writeL(stderr, ["Error: ", G,
                             " occurs in the TR actions and is neither durative nor discrete in ", nl_,
                             Rule,nl_, "of the TR procedure ", Head,nl_]),
    fail.
'$display_error'(tr_in_tr_actions(G, '$tr'(Head, Rule), Vars)) :-
    !,
    thaw_term((Head,Rule)),
    '$vars2names'((Head,Rule), Vars),
    writeL(stderr, ["Error: the TR procedure ", G,
                             " is mixed with other actions in ", nl_,
                             Rule,nl_, "of the TR procedure ", Head,nl_]),
    fail.
'$display_error'(action_in_rel(G, '$tr'(Head, Rule), Vars)) :-
    !,
    thaw_term((Head,Rule)),
    '$vars2names'((Head,Rule), Vars),
    writeL(stderr, ["Error: ", G,
                             " is an action - expecting relation in rule", nl_,
                             Rule,nl_, "of the TR procedure ", Head,nl_]),
    fail.
'$display_error'(action_in_rel(G, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", G,
                             " is an action - expecting relation in", nl_,
                             Clause,nl_]),
    fail.
'$display_error'(expression_in_relact(G, '$tr'(Head, Rule), Vars)) :-
    !,
    thaw_term((Head,Rule)),
    '$vars2names'((Head,Rule), Vars),
    writeL(stderr, ["Error: ", G,
                    " is an expression", nl_,
                    "- expecting relation or action in rule", nl_,
                    Rule,nl_, "of the TR procedure ", Head,nl_]),
    fail.
'$display_error'(expression_in_relact(G, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", G,
                    " is an expression", nl_,
                    "- expecting relation or action in", nl_,
                    Clause,nl_]),
    fail.
'$display_error'(not_a_global(A, G, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", A,
                             " is not a global value in ", G, " of",nl_,
                             Clause,nl_]),
    fail.
'$display_error'(undeclared_relation(Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: undeclared relation ", nl_,
                             Clause,nl_]),
    fail.
'$display_error'(undeclared_function(Clause, Vars)) :-
    !,
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: undeclared function ", nl_,
                             Clause,nl_]),
    fail.
'$display_error'(wrong_num_args(Term, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", Term,
                    " has the wrong number of arguments in", nl_,
                    Clause,nl_]),
    fail.
'$display_error'(non_atom_functor(F, Head, Clause, Vars)) :-
    !,
    thaw_term(Clause),
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", F, " of " , Head, " is not an atom", nl_, "in", nl_,
                             Clause,nl_]),
    fail.
'$display_error'(mismatched_function_types(Clause, Vars)) :-
    !,
    '$vars2names'(Clause, Vars),
    writeL(stderr, ["Error: ", nl_,
                             Clause,nl_, "does not match type declaration", nl_]),
    fail.
'$display_error'(already_declared(Term, Clause, Vars)) :-
    !,
    '$vars2names'(Clause, Vars),
    '$get_ultimate_functor'(Term, F),
    writeL(stderr, ["Error: ", Term,
                    " contains the wrong number of arguments in", nl_,
                    Clause,nl_,"The type of ", F, " is", nl_]),
    display_type(stderr, F),nl(stderr),
    
    fail.
'$display_error'(already_declared(Term,
                                  tr_action('$tr'(Head, Rule), Vars))) :-
    !,
    thaw_term('$tr'(Head, Rule)),
    '$vars2names'('$tr'(Head, Rule), Vars),
    '$get_ultimate_functor'(Term, F),
    writeL(stderr, ["Error: ", Term,
                    " contains the wrong number of arguments in the rule",nl_,
                    Rule,nl_,
                    "of the TR procedure ", Head,nl_,
                    "The type of ", F, " is", nl_]),
    display_type(stderr, F),nl(stderr),
     fail.

'$display_error'(E) :-
    writeL(stderr, ["unmatched error ", E, nl_]).

%process_mode_error(M, A, B, C) :- errornl(process_mode_error(M, A, B, C)),fail.
process_type_as_term_error(Type, Clause, Vars) :-
    !,
    '$vars2names'(Clause, Vars),
    '$update_error'(type_as_term_error(Type, Clause)),
    fail.

process_mode_error(doubleq_bound, V) :-
    !,
    ip_lookup(type_checker_context, _(_, Vars)),
    thaw_term(Vars),
    '$vars2names'([V], Vars),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(mode_error(doubleq_bound, V, ContextCopy)),
    fail.
process_mode_error(unbound_by_body, Unbound) :-
    !,
    ip_lookup(type_checker_context, _(_, Vars)),
    thaw_term(Vars),
    '$vars2names'(Unbound, Vars),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(mode_error(unbound_by_body, Unbound, ContextCopy)),
    fail.
process_mode_error(Mode, Arg) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    ip_lookup(call_type, CallType),
    '$update_error'(mode_error(Mode, Arg, OG, Context, OGCopy, ContextCopy, CallType)),
    fail.
process_no_type_error :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(no_type_error(OG, Context, OGCopy, ContextCopy)),
    fail.
process_not_mode_error(Vars) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(not_mode_error(Vars, OG, Context, OGCopy, ContextCopy)),
    fail.
process_globals_locals_error(Vars) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(globals_locals_error(Vars, OG, Context, OGCopy, ContextCopy)),
    fail.
process_locals_globals_error(Vars) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(locals_globals_error(Vars, OG, Context, OGCopy, ContextCopy)),
    fail.
process_not_underscore_mode_error(Vars) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(not_underscore_mode_error(Vars, OG, Context, OGCopy, ContextCopy)),
    fail.
process_forall_quant_vars_error(Vars) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(forall_quant_vars_error(Vars, OG, Context, OGCopy, ContextCopy)),
    fail.
process_forall_nonlocal_quant_vars_error(Vars) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(forall_nonlocal_quant_vars_error(Vars, OG, Context, OGCopy, ContextCopy)),
    fail.
process_mode_error_list_constr(T, Vars, Term) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(mode_error_list_constr(T, Vars, Term, OG, Context, OGCopy, ContextCopy)),
    fail.
   

process_mode_error(Mode, rel_type(_, T), FVars, OG) :-
    !,
    process_mode_error(Mode, T, FVars, OG).
process_mode_error(Mode, tr(Kind, Vars), FVars, OG) :-
    !,
    %global_state_lookup('$error_level',EL),
    '$update_error'(mode_error(Mode, tr(Kind, Vars), FVars, OG)),
    %global_state_increment('$error_level', _),
    fail.
process_mode_error(Mode, relation(Kind, Vars), FVars, OG) :-
    !,
    ip_lookup(call_type, CallType),
    %global_state_lookup('$error_level',EL),
    '$update_error'(mode_error(Mode, relation(Kind, Vars), FVars, OG, CallType)),
    %global_state_increment('$error_level', _),
    fail.
process_mode_error(Mode, action(Kind, Vars), FVars, OG) :-
    !,
    ip_lookup(call_type, CallType),
    %global_state_lookup('$error_level',EL),
    '$update_error'(mode_error(Mode, action(Kind, Vars), FVars, OG, CallType)),
    %global_state_increment('$error_level', _),
    fail.
process_mode_error(Mode, relhead_mode(Thawed), FVars, OG) :-
    !,
    %global_state_lookup('$error_level',EL),
    '$update_error'(mode_error(Mode, relhead_mode(Thawed), FVars, OG, _CallType)),
    %global_state_increment('$error_level', _),
    fail.
process_mode_error(Mode, relhead0_mode(Thawed), FVars, OG) :-
    !,
    ip_lookup(call_type, CallType),
    %global_state_lookup('$error_level',EL),
    '$update_error'(mode_error(Mode, relhead0_mode(Thawed), FVars, OG, CallType)),
    %global_state_increment('$error_level', _),
    fail.
process_mode_error(Mode, relation_query(Kind, Vars), FVars, OG) :-
    !,
    ip_lookup(call_type, CallType),
    %global_state_lookup('$error_level',EL),
    '$update_error'(mode_error(Mode, relation_query(Kind, Vars), FVars, OG, CallType)),
    %global_state_increment('$error_level', _),
    fail.
process_mode_error(Mode, expression_query(Kind, Vars), FVars, OG) :-
    !,
    %global_state_lookup('$error_level',EL),
    '$update_error'(mode_error(Mode, expression_query(Kind, Vars), FVars, OG)),
    %global_state_increment('$error_level', _),
    fail.
process_mode_error(Mode, function_goal_type(Kind, Vars), FVars, OG) :-
    !,
    %global_state_lookup('$error_level',EL),
    '$update_error'(mode_error(Mode, function_goal_type(Kind, Vars), FVars, OG)),
    %global_state_increment('$error_level', _),
    fail.
process_mode_error(X, A, B, C) :- errornl(unprocessed___XXXXX___(process_mode_error(X, A, B, C))),fail.

process_head_output_type_error(Var, T, Type, Mode) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(head_output_type_error(Var, T, Type, Mode,
					   OG, Context, OGCopy, ContextCopy)),
    fail.

%%process_type_error(Term, Type, T) :- errornl(process_type_error(Term, Type, T)),fail.
process_type_error(Term, Type, T) :-
    ip_lookup(current_predicate, OG),
    OG = '$head'(_Head), !,
    ip_lookup(type_checker_context, Context),
    ip_lookup(type_checker_copy_context, ContextCopy),
    ip_lookup(call_type, CallType),
    '$update_error'(type_error(Term, Type, T, OG, Context, OG,
			       ContextCopy, CallType)),
    fail.
process_type_error(Term, Type, T) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    ip_lookup(call_type, CallType),
    (
     '$error'(inconsistent_type_error(_, _, _, _, _, OGCopy, ContextCopy))
    ->
      true
    ;
     '$update_error'(type_error(Term, Type, T, OG, Context, OGCopy, ContextCopy, CallType))
    ), !, fail.

process_inconsistent_unify_type_error(Term, Type1, Type2) :-
    ip_lookup(type_checker_context, Context),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(inconsistent_unify_type_error(Term, Type1, Type2,
					 Context, ContextCopy)),
    fail.
process_inconsistent_type_error(Term, Type1, Type2) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(inconsistent_type_error(Term, Type1, Type2,
					 OG, Context, OGCopy, ContextCopy)),
    fail.
/*
process_inconsistent_types_error(Term1, Term2) :-
    ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(inconsistent_types_error(Term1, Term2,
					 OG, Context, OGCopy, ContextCopy)),
    fail.
*/
process_inconsistent_type_error_head(Term, Type1, Type2) :-
    %ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Context),
    %ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, ContextCopy),
    '$update_error'(inconsistent_type_error_head(Term, Type1, Type2,
					 Context, ContextCopy)),
    fail.

process_unification_failure_warning(A, B) :-
    %ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Kind(Clause, VarNames)),
    %ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, _Kind(ClauseCopy, VarNamesCopy)),
    '$set_var_names'(Clause, VarNames, _),
    '$set_var_names'(ClauseCopy, VarNamesCopy, _),
    writeL(stderr, ["Warning: the unification ", (A = B),
                    " always fails in the ", Kind, " rule", nl_,
                    ClauseCopy, nl_]),
    fail.
process_unification_failure_warning(_A, _B).

process_unification_succeeds_warning(A, B) :-
    %ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Kind(Clause, VarNames)),
    %ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, _Kind(ClauseCopy, VarNamesCopy)),
    '$set_var_names'(Clause, VarNames, _),
    '$set_var_names'(ClauseCopy, VarNamesCopy, _),
    writeL(stderr, ["Warning: the unification ", (A = B),
                    " always succeeds in the ", Kind, " rule", nl_,
                    ClauseCopy, nl_]),
    fail.
process_unification_succeeds_warning(_A, _B).

process_type_check_failure_warning :-
    %ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Kind(Clause, VarNames)),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, _Kind(ClauseCopy, VarNamesCopy)),
    '$set_var_names'(Clause, VarNames, _),
    '$set_var_names'(ClauseCopy, VarNamesCopy, _),
    writeL(stderr, ["Warning: the type check ", OGCopy,
                    " always fails in the ", Kind, " rule", nl_,
                    ClauseCopy, nl_]),
    fail.
process_type_check_failure_warning.

process_type_check_succeeds_warning :-
    %ip_lookup(current_predicate, OG),
    ip_lookup(type_checker_context, Kind(Clause, VarNames)),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup(type_checker_copy_context, _Kind(ClauseCopy, VarNamesCopy)),
    '$set_var_names'(Clause, VarNames, _),
    '$set_var_names'(ClauseCopy, VarNamesCopy, _),
    writeL(stderr, ["Warning: the type check ", OGCopy,
                    " always succeeds in the ", Kind, " rule", nl_,
                    ClauseCopy, nl_]),
    fail.
process_type_check_succeeds_warning.

   
process_error_context(Error) :-
    %global_state_lookup('$error_level',EL),
    '$update_error'(Error),
    %global_state_increment('$error_level', _),
    fail.

%%'$process_saved_error'(X) :- errornl('$process_saved_error'(X)), fail.
'$process_saved_error'(not_mode_error(Vars, OG, Kind(Clause, VarNames),
                                      OGCopy, _(ClauseCopy, VarNamesCopy))) :-
    once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(not_mode_error(Vars, Kind, OG, OGCopy, ClauseCopy))
         )), fail.
'$process_saved_error'(globals_locals_error(Vars, OG,
                                                 Kind(Clause, VarNames),
                                      OGCopy, _(ClauseCopy, VarNamesCopy))) :-
    once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(globals_locals_error(Vars, Kind, OG, OGCopy, ClauseCopy))
         )), fail.
'$process_saved_error'(locals_globals_error(Vars, OG,
                                                 Kind(Clause, VarNames),
                                      OGCopy, _(ClauseCopy, VarNamesCopy))) :-
    once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(locals_globals_error(Vars, Kind, OG, OGCopy, ClauseCopy))
         )), fail.
'$process_saved_error'(not_underscore_mode_error(Vars, OG,
                                                 Kind(Clause, VarNames),
                                      OGCopy, _(ClauseCopy, VarNamesCopy))) :-
    once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(not_underscore_mode_error(Vars, Kind, OG, OGCopy, ClauseCopy))
         )), fail.
'$process_saved_error'(forall_quant_vars_error(Vars, OG,
                                                 Kind(Clause, VarNames),
                                      OGCopy, _(ClauseCopy, VarNamesCopy))) :-
    once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(forall_quant_vars_error(Vars, Kind, OG, OGCopy, ClauseCopy))
         )), fail.
'$process_saved_error'(forall_nonlocal_quant_vars_error(Vars, OG,
                                                 Kind(Clause, VarNames),
                                      OGCopy, _(ClauseCopy, VarNamesCopy))) :-
    once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(forall_nonlocal_quant_vars_error(Vars, Kind, OG, OGCopy, ClauseCopy))
         )), fail.
'$process_saved_error'(mode_error_list_constr(T, Vars, Term, OG,
                                                 Kind(Clause, VarNames),
                                      OGCopy, _(ClauseCopy, VarNamesCopy))) :-
    once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(mode_error_list_constr(T, Vars, Term, Kind,
                                                  OG, OGCopy, ClauseCopy))
         )), fail.
'$process_saved_error'(mode_error(Mode, Arg, OG, Kind(Clause, VarNames),
                                  OGCopy, _(ClauseCopy, VarNamesCopy), CallType)) :-
    once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(mode_error(Mode, Arg, Kind, OG, OGCopy, ClauseCopy, CallType))
         )), fail.
'$process_saved_error'(mode_error(doubleq_bound, V,
				  Kind(Clause, VarNames))) :-
    once((
          '$set_var_names'(Clause, VarNames, _),
          '$display_error'(mode_error(doubleq_bound, V, Kind, Clause))
         )), fail.
'$process_saved_error'(mode_error(unbound_by_body, Unbound,
				  Kind(Clause, VarNames))) :-
    once((
          '$set_var_names'(Clause, VarNames, _),
          '$display_error'(mode_error(unbound_by_body, Unbound, Kind, Clause))
         )), fail.
'$process_saved_error'(type_error(Term, Type, T, OG, Kind(Clause, VarNames),
                                  OGCopy, _(ClauseCopy, VarNamesCopy),
                                  CallType)) :-
    once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(type_error(Term, Type, T, Kind, OG, OGCopy, ClauseCopy, CallType))
         )), fail.
'$process_saved_error'(inconsistent_unify_type_error(Term, Type1, Type2,
                                                Kind(Clause, VarNames),
                                                 _(ClauseCopy, VarNamesCopy))) :-
     once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(inconsistent_unify_type_error(Term, Type1, Type2, Kind,
						   ClauseCopy))
         )), fail.
'$process_saved_error'(inconsistent_type_error(Term, Type1, Type2,
                                                OG, Kind(Clause, VarNames),
                                                OGCopy,
                                                _(ClauseCopy, VarNamesCopy))) :-
     once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(inconsistent_type_error(Term, Type1, Type2, Kind, OG, OGCopy, ClauseCopy))
         )), fail.
/*
  '$process_saved_error'(inconsistent_types_error(Term1, Term2,
                                                OG, Kind(Clause, VarNames),
                                                OGCopy,
                                                _(ClauseCopy, VarNamesCopy))) :-
     once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(inconsistent_types_error(Term1, Term2, Kind, OG, OGCopy, ClauseCopy))
         )), fail.
*/
'$process_saved_error'(inconsistent_type_error_head(Term, Type1, Type2,
                                                Kind(Clause, VarNames),
                                                _(ClauseCopy, VarNamesCopy))) :-
     once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(inconsistent_type_error_head(Term, Type1, Type2, Kind, ClauseCopy))
         )), fail.

 '$process_saved_error'(head_output_type_error(Var, T, Type, Mode, OG,
					       Kind(Clause, VarNames), OGCopy,
					       _(ClauseCopy, VarNamesCopy))) :-
     once((
          '$set_var_names'(Clause, VarNames, _),
          '$set_var_names'(ClauseCopy, VarNamesCopy, _),
          '$display_error'(head_output_type_error(Var, T, Type, Mode, OG, Kind,
						  OGCopy, ClauseCopy))
         )), fail.
    
    
'$process_saved_error'(relhead_type(relation, VarNames, Clause)) :-
    once((
          '$set_var_names'(Clause, VarNames, Thawed),
          '$display_error'(relhead_type(Thawed, Clause))
         )), fail.
'$process_saved_error'(relhead_type(action, VarNames, Clause)) :-
    once((
          '$set_var_names'(Clause, VarNames, Thawed),
          '$display_error'(acthead_type(Thawed, Clause))
         )), fail.
'$process_saved_error'(rel_type(Pred, relation(Clause, Vars))) :-
    once((
          '$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(rel_type(Thawed, Clause, Pred))
         )), fail.
'$process_saved_error'(rel_type(Pred, function_goal_type(Clause, Vars))) :-
    once((
          '$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(function_rel_type(Thawed, Clause, Pred))
         )), fail.
'$process_saved_error'(type_type_error(T, TT, TypeTest,
                                       _(_RelAct, VarNames, Clause))) :-
    once((
          '$set_var_names'(Clause, VarNames, Thawed),
          '$display_error'(type_type_error(TypeTest, Thawed, Clause, T, TT))
         )), fail.
'$process_saved_error'(type_type_query_error(T, TT, TypeTest,
                                       _(_RelAct, VarNames))) :-
    once((
          freeze_term(TypeTest),
           '$set_var_names'(TypeTest, VarNames, Thawed),
          '$display_error'(type_type_query_error(TypeTest, Thawed, T, TT))
         )), fail.
'$process_saved_error'(inconsistent_type_error(T, T1, T2,
                                          rel_type(Pred,
                                               RelAct(Clause, Vars)))) :-
    once((
          %'$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(inconsistent_type_error(Vars, Clause, T,
						T1, T2, RelAct, Pred))
         )), fail.
/*
  '$process_saved_error'(inconsistent_types_error(Term1, Term2,
                                          rel_type(Pred,
                                               RelAct(Clause, Vars)))) :-
    once((
          %'$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(inconsistent_types_error(Vars, Clause, Term1, Term2,
                                                    RelAct, Pred))
         )), fail.
*/
 '$process_saved_error'(arg_type_error(T, TT,
                                     relhead_type(RelAct, VarNames, Clause))) :-
    once((
          '$set_var_names'(Clause, VarNames, Thawed),
          '$display_error'(relhead_term_type(RelAct, Thawed, Clause, T, TT))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT,
                                      rel_type(Pred,
                                               tr(Clause, Vars)))) :-
    once((
          '$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(tr_term_type(Thawed, Clause, T, TT, Pred))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT,
                                      rel_type(Pred,
                                               tr_goal(Clause, Vars)))) :-
    once((
          '$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(tr_goal_term_type(Thawed, Clause, T, TT, Pred))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT,
                                      rel_type(_Pred,
                                               tr_ts_time(Clause, Vars)))) :-
    once((
          '$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(tr_ts_time_type(Thawed, Clause, T, TT))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT,
                                      rel_type(_Pred,
                                               wait_repeat_time(Clause, Vars)))) :-
    once((
          '$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(wait_repeat_time_type(Thawed, Clause, T, TT))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT,
                                      rel_type(_Pred,
                                               wait_time(Clause, Vars)))) :-
    once((
          '$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(wait_time_type(Thawed, Clause, T, TT))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT,
                                      rel_type(Pred,
                                               tr_action(Clause, Vars)))) :-
    once((
          '$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(tr_action_term_type(Thawed, Clause, T, TT, Pred))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT,
                                      rel_type(Pred,
                                               relation(Clause, Vars)))) :-
    once((
          '$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(rel_term_type(Thawed, Clause, T, TT, Pred))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT,
                                      rel_type(Pred,
                                               action(Clause, Vars)))) :-
    once((
          '$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(action_term_type(Thawed, Clause, T, TT, Pred))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT,
                                      rel_type(Pred,
                                               function_goal_type(Clause,
                                                                  Vars)))) :-
    once((
          '$set_var_names'(Clause, Vars, Thawed),
          '$display_error'(function_rel_type(Thawed, Clause, T, TT, Pred))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT,
                                     funhead_type(Vars, Rule))) :-
    once((
          '$set_var_names'(Rule, Vars, Thawed),
          '$display_error'(funhead_type(Thawed, Rule, T, TT))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT,
                                     function_type(Vars, Rule))) :-
    once((
          '$set_var_names'(Rule, Vars, Thawed),
          '$display_error'(function_term_type(Thawed, Rule, T, TT))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT, rel_type(Pred, relation_query(Query, Vars)))) :-
    once((
         '$display_error'(relation_query_type(Query, Vars, Pred, T, TT))
         )), fail.
'$process_saved_error'(arg_type_error(T, TT, rel_type(Pred, expression_query(Query, Vars)))) :-
    once((
         '$display_error'(expression_query_type(Query, Vars, Pred, T, TT))
         )), fail.
'$process_saved_error'(function_type(Vars, Rule)) :-
    once((
          '$set_var_names'(Rule, Vars, Thawed),
          Rule = (_ -> T),
          '$display_error'(function_term_type(Thawed, Rule, T, _TT))
         )), fail.
'$process_saved_error'(relation_query(Query)) :-
    once(('$display_error'(relation_query(Query)))),
    fail.
'$process_saved_error'(mode_error(Mode, tr(Kind, Vars), FVars, OG)) :-
    once(( 
          '$display_error'(tr_act_mode(Mode, Vars, Kind, FVars, OG))
         )),
    fail.
'$process_saved_error'(mode_error(Mode, relation(Kind, Vars), FVars, OG)) :-
    once((
          '$display_error'(rel_act_mode(Mode, relation, Vars, Kind, _, FVars, OG))
         )),
    fail.
'$process_saved_error'(mode_error(Mode, action(Kind, Vars), FVars, OG)) :-
    once((
          '$display_error'(rel_act_mode(Mode, action, Vars, Kind, _, FVars, OG))
         )),
    fail.
'$process_saved_error'(mode_error(_, relation_query(Kind, Vars), FVars, OG)) :-
    once((
          '$display_error'(rel_query_mode(Kind, Vars, FVars, OG))
         )),
    fail.
'$process_saved_error'(mode_error(_, expression_query(Kind, Vars), FVars, OG)) :-
    once((
          '$display_error'(expression_query_mode(Kind, Vars, FVars, OG))
         )),
    fail.
'$process_saved_error'(mode_error(_, function_goal_type(Kind, Vars), FVars, OG)) :-
    once((
           '$display_error'(function_rel_mode(Vars, Kind, FVars, OG))
         )),
    fail.
'$process_saved_error'(mode_error(Mode, relhead_mode(Thawed), FVars, OG)) :-
    once((
           '$display_error'(relhead_mode(Mode, Thawed, FVars, OG))
         )),
    fail.
'$process_saved_error'(mode_error(Mode, relhead0_mode(Thawed), FVars, OG)) :-
    once((
           '$display_error'(relhead0_mode(Mode, Thawed, FVars, OG))
         )),
    fail.

'$process_saved_error'(_E).

'$update_error'(_E) :-
    ip_lookup(check_index_, N1),
    global_state_lookup(gs_check_index_, N2),
    (
      N1 < N2
    ->
      !
    ;
      global_state_set(gs_check_index_, N1),
      fail
    ).   
'$update_error'(Error) :-
    retract('$error'(_)), !,
    assert('$error'(Error)).
'$update_error'(Error) :-
    assert('$error'(Error)).
