%% Copyright 2014 Keith Clark, Peter Robinson
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.

%?- [qwrite].
?-op(990,xfx,'<=').
?-op(200,xf,'<=').
?-op(200,xf,'?->').
?-op(200,xf,'~>').
?-op(200,xf,'~>>').
%?-op(200,xf,'->').
?-op(200,fx,'^').
?-op(200,fx,'?').
?-op(920,xfx,'=>').
?-op(920,xfx,'~>').
?-op(920,xfx,'~>>').
%%%% !!!!
%?- op(qulog_table, 920,xfx,'->').

?- op(600,xfx,'..').

?-op(1010,xfx,'::').

?-op(906,fx,'@').
?-op(906,fx,'forall').
?-op(906,fx,'exists').
%?- op(1150, xf, '$').

?- op(910, xfy,  '&' ). 
?-op(700,xfx,'=?').   
?-op(700,xfx,'$=').   
?-op(500,xfy,'<>').
?-op(500,xfy,'++').   
?-op(600,xfx,'in').
?-op(600,xfx,'subs').

?- op(200, fy, 'not').
%?- op(907, fy, 'type').
?- op(190,fx, '#').
?- op(1101, xfx, '::=').
?- op(50, xfx, '?').
?- dynamic('$parser_info'/4).

read_qulog_term(Term) :-
    stream_property(0, line_number(StartLine)),
    interpreter_read_query_string(QString),
    stream_property(0, line_number(EndLine)),
    (
      string_length(QString, 0)
    ->
      Term = end_of_file,
      VarNames = []
    ;
      retractall('$parser_info'(_, _, _, _)),
      assert('$parser_info'(QString, StartLine, EndLine, stdin)),
      string2tokens(QString, Tokens, VarList),
      Tokens \= [],
      '$collect_var_tables'(VarList, _Variables, VarNames, _Singles, []),
     '$unify_on_names'(VarList),
      catch(
            (
              phrase(term(Term), Tokens, [])
            ->
              true
            ;
              report_error('Can\'t parse', Tokens),
              nl,
              fail
              ),
            syntax_error,
            fail
           )
    ).


interpreter_read_query(Query, VarNamesOut) :-
    stream_property(0, line_number(StartLine)),
    interpreter_read_query_string(QString),
    stream_property(0, line_number(EndLine)),
    (
      string_length(QString, 0)
    ->
      Query = end_of_file,
      VarNames = []
    ;
      retractall('$parser_info'(_, _, _, _)),
      assert('$parser_info'(QString, StartLine, EndLine, stdin)),
      string2tokens(QString, Tokens, VarList),
      Tokens \= [],
      '$collect_var_tables'(VarList, _Variables, VarNames, _Singles, []),
     '$unify_on_names'(VarList),
      remove_duplicates(VarNames, VarNamesOut),
      catch(
            (
              phrase(query(Query), Tokens, [])
            ->
              true
            ;
              report_error('Can\'t parse', Tokens),
              nl,
              fail
              ),
            syntax_error,
            fail
           )
    ).

'$unify_on_names'([]).
'$unify_on_names'([A|B]) :-
    member(A, B), !,
    '$unify_on_names'(B).
'$unify_on_names'([_A|B]) :-
    '$unify_on_names'(B).

strip_dot_whitespace(InString, OutString) :-

    (
      string_concat(A1, A2, InString),
      string_concat("%", _, A2)
    ->
      S0 = A1
    ;
      S0 = InString
    ),
    string_concat(S1, S2, S0),
    whitespaces(S2),
    string_concat(OutString, ".", S1), !.

whitespaces(S) :- string_length(S, 0), !.
whitespaces(S) :-
    string_concat(S1, " ", S), !,
    whitespaces(S1).
whitespaces(S) :-
    string_concat(S1, "\t", S), !,
    whitespaces(S1).
    
interpreter_read_query_string(QString) :-
    get_line(Line),
    (
      Line = -1
    ->
      QString = ""
    ;
      (
       strip_dot_whitespace(Line, QString)
      ->
        true
      ;
        string_concat(Line, " ", LineNL),
        peek(Char),
        interpreter_read_query_string_aux(Char, LineNL, QString)
      )
    ).

interpreter_read_query_string_aux('\n', Line, QString) :-
    !,
    get_line(_),
    QString = Line.
interpreter_read_query_string_aux(_, Line, QString) :-
    get_line(Line1),
    (
      Line1 = -1
    ->
      QString = Line
    ;
      strip_dot_whitespace(Line1, Line0)
    ->
      string_concat(Line0, " ", Line1NL),
      string_concat(Line, Line1NL, QString)
    ;
      string_concat(Line1, " ", Line1NL),
      string_concat(Line, Line1NL, NewLine),
      peek(Char),
      interpreter_read_query_string_aux(Char, NewLine, QString)
    ).

read_term_string(Stream, String, StartLine, EndLine) :-
    stream_property(Stream, line_number(StartLine)),
    get_line(Stream, Line),
    (
     Line = -1
    ->
     String = ""
    ;
     string_concat(Line, "\n", LineNL),
     peek(Stream, Char),
     read_term_string_aux(Char, Stream, LineNL, String)
    ),
    stream_property(Stream, line_number(EndLine)).


read_term_string_aux(' ', Stream, Line, String) :-
    !,
    get_line(Stream, Line1),
    (
     Line1 = -1
    ->
     String = Line
    ;
     string_concat(Line1, "\n", Line1NL),
     string_concat(Line, Line1NL, NewLine),
     peek(Stream, Char),
     read_term_string_aux(Char, Stream, NewLine, String)
    ).
read_term_string_aux('\t', Stream, Line, String) :-
    !,
    get_line(Stream, Line1),
    (
     Line1 = -1
    ->
     String = Line
    ;
     string_concat(Line1, "\n", Line1NL),
     string_concat(Line, Line1NL, NewLine),
     peek(Stream, Char),
     read_term_string_aux(Char, Stream, NewLine, String)
    ).
read_term_string_aux('\t', Stream, Line, String) :-
    !,
    get_line(Stream, Line1),
    (
     Line1 = -1
    ->
     String = Line
    ;
     string_concat(Line1, "\n", Line1NL),
     string_concat(Line, Line1NL, NewLine),
     peek(Stream, Char),
     read_term_string_aux(Char, Stream, NewLine, String)
    ).
read_term_string_aux('\n', Stream, Line, String) :-
    !,
    get_line(Stream, Line1),
    (
     Line1 = -1
    ->
     String = Line
    ;    
     %string_concat(Line1, "\n", Line1NL),
     %string_concat(Line, Line1NL, NewLine),
     peek(Stream, Char),
     read_term_string_aux(Char, Stream, Line, String)
    ).
read_term_string_aux('}', Stream, Line, String) :-
    !,
    get_line(Stream, Line1),
    string_concat(Line1, "\n", Line1NL),
    string_concat(Line, Line1NL, String).
    
    /*
    (
     Line1 = -1
    ->
     String = Line
    ;    
     string_concat(Line1, "\n", Line1NL),
     string_concat(Line, Line1NL, NewLine),
     peek(Stream, Char),
     read_term_string_aux(Char, Stream, NewLine, String)
    ).
      */
read_term_string_aux(_, _Stream, Line, String) :-
    String = Line.
/*
    string_concat(LineFS, "\n", Line), 
    (
      strip_dot_whitespace(LineFS,StringNoNL)
    ->
      string_concat(StringNoNL,"\n",String)
    ;
      string_concat(LineFS,"\n",String)
    ).
*/

string2tokens(String, Tokens, VarList) :-
    open_string(read(String), Stream),
    get_tokens(Stream, Tokens1),
    close(Stream),
    '$post_process_tokens'(Tokens1, Tokens, VarList).

/*
read_from_file(File, Tokens, VarList, Variables, VarNames, Singles) :-
    open(File, read, Stream),
    read_term_string(Stream, String),
    string2tokens(String, Tokens, VarList),
    '$collect_var_tables'(VarList, Variables, VarNames, Singles, []).
*/

'$graphics_chars'(A) :-
      atom_chars(A, [C|_]),
      '$parser_graphics'(C), !.

'$parser_graphics'('|').
'$parser_graphics'('-').
'$parser_graphics'('/').
'$parser_graphics'('+').
'$parser_graphics'('*').
'$parser_graphics'('<').
'$parser_graphics'('=').
'$parser_graphics'('>').
'$parser_graphics'('#').
'$parser_graphics'('@').
'$parser_graphics'('$').
'$parser_graphics'('\\').
'$parser_graphics'('^').
'$parser_graphics'('&').
'$parser_graphics'(';').
'$parser_graphics'('~').
'$parser_graphics'('`').
'$parser_graphics'(':').
'$parser_graphics'('.').
'$parser_graphics'('?').
'$parser_graphics'('!').

'$post_process_tokens'([], [], []).
'$post_process_tokens'([t(5, '.')], [], [])  :- !.
'$post_process_tokens'([t(5, '.'), t(13,_)], [], [])  :- !.
'$post_process_tokens'([t(4, Atom)|Rest], [op('`')|Output],
		       VarList) :-
    string_to_atom(S, Atom), string_concat("`", B, S),B \= "",
    !,
    string_to_atom(B, Atom1),
    '$post_process_tokens'([t(4, Atom1)|Rest], Output, VarList).
'$post_process_tokens'([t(4, '*-')|Rest], Output, VarList) :-
    !,
    '$post_process_tokens'([t(4, '*'), t(4, '-')|Rest], Output, VarList).
'$post_process_tokens'([t(4, '**-')|Rest], Output, VarList) :-
    !,
    '$post_process_tokens'([t(4, '**'), t(4, '-')|Rest], Output, VarList).
'$post_process_tokens'([t(4, '+-')|Rest], Output, VarList) :-
    !,
    '$post_process_tokens'([t(4, '+'), t(4, '-')|Rest], Output, VarList).
'$post_process_tokens'([t(4, '--')|Rest], Output, VarList) :-
    !,
    '$post_process_tokens'([t(4, '-'), t(4, '-')|Rest], Output, VarList).
'$post_process_tokens'([t(4, '='), t(7, '?')|Rest], [op('=?')|Output],
                       VarList) :-
    !,
    '$post_process_tokens'(Rest, Output, VarList).
'$post_process_tokens'([t(7, '!'), t(4, '?')|Rest], [op('!?')|Output],
                       VarList) :-
    !,
    '$post_process_tokens'(Rest, Output, VarList).
'$post_process_tokens'([t(7, '!'), t(4, '??')|Rest], [op('!??')|Output],
                       VarList) :-
    !,
    '$post_process_tokens'(Rest, Output, VarList).

'$post_process_tokens'([t(Kind, A)|Rest], [op(A)|PRest], Varlist) :-
    ( Kind = 4 ; Kind = 7 ),
    '$parser_reserved_atom'(A),
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(11, 0)|Rest], [op('!')|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(5, '.')|Rest], [op('.')|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
%'$post_process_tokens'([t(4, A)|_Rest], _Tokens, _Varlist) :-
%    '$graphics_chars'(A), !, fail.
'$post_process_tokens'([t(1, '['), t(1, ']')|Rest],
		       [atom([])|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(1, '(')|Rest], [ob('(')|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(1, ' (')|Rest], [ob(' (')|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(1, '[')|Rest], [ob('[')|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(1, '{')|Rest], [ob('{')|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(1, ')')|Rest], [cb(')')|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(1, ']')|Rest], [cb(']')|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(1, '}')|Rest], [cb('}')|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(1, P)|Rest], [punct(P)|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(4, A)|Rest], [atom(A)|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(6, A)|Rest], [var(_, A)|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(7, A)|Rest], [atom(A)|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(9, S)|Rest], [string(S)|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(13, _S)|Rest], PRest, Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(3, N)|Rest], [int(N)|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(14, N)|Rest], [num(N)|PRest], Varlist) :-
    !,
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([t(2, Value)|Rest], [Variable|PRest],
		       [Variable|Varlist]) :-
    !,
    Variable = var(Var,Value),
    '$readR_var'(Var,Value), 
    '$post_process_tokens'(Rest, PRest, Varlist).
'$post_process_tokens'([X|Rest], [X|PRest], Varlist) :-
    '$post_process_tokens'(Rest, PRest, Varlist).



'$parser_reserved_atom'('+').
'$parser_reserved_atom'('-').
'$parser_reserved_atom'('*').
'$parser_reserved_atom'('/').
'$parser_reserved_atom'('\\').
'$parser_reserved_atom'('//').
'$parser_reserved_atom'('/\\').
'$parser_reserved_atom'('\\/').
'$parser_reserved_atom'('>>').
'$parser_reserved_atom'('<<').
'$parser_reserved_atom'('**').
'$parser_reserved_atom'('?').
'$parser_reserved_atom'('??').
'$parser_reserved_atom'('!??').
'$parser_reserved_atom'('!?').
'$parser_reserved_atom'('^').
'$parser_reserved_atom'(':').
'$parser_reserved_atom'('::').
'$parser_reserved_atom'('::=').
'$parser_reserved_atom'('<=').
'$parser_reserved_atom'('~>').
'$parser_reserved_atom'('~>>').
'$parser_reserved_atom'('->').
'$parser_reserved_atom'('=>').
'$parser_reserved_atom'('~~').
'$parser_reserved_atom'('^^').
'$parser_reserved_atom'('~^').
'$parser_reserved_atom'('`').
'$parser_reserved_atom'('&').
'$parser_reserved_atom'(';').
'$parser_reserved_atom'('..').
'$parser_reserved_atom'('|').
'$parser_reserved_atom'('!').
'$parser_reserved_atom'('||').
'$parser_reserved_atom'('=').
'$parser_reserved_atom'('\\=').
'$parser_reserved_atom'('=?').
'$parser_reserved_atom'('$=').
'$parser_reserved_atom'('@').
'$parser_reserved_atom'('<>').
'$parser_reserved_atom'('?-').
'$parser_reserved_atom'('++').
'$parser_reserved_atom'('@..').
'$parser_reserved_atom'('#').
'$parser_reserved_atom'('in').
'$parser_reserved_atom'('>').
'$parser_reserved_atom'('>=').
'$parser_reserved_atom'('<').
'$parser_reserved_atom'('<=').
'$parser_reserved_atom'('@>').
'$parser_reserved_atom'('@>=').
'$parser_reserved_atom'('@<').
'$parser_reserved_atom'('@<=').
'$parser_reserved_atom'('@=').
'$parser_reserved_atom'('$').
'$parser_reserved_atom'(':=').
'$parser_reserved_atom'('+:=').
'$parser_reserved_atom'('-:=').
%'$parser_reserved_atom'('type').
'$parser_reserved_atom'('forall').
'$parser_reserved_atom'('exists').
'$parser_reserved_atom'('mod').
'$parser_reserved_atom'('union').
'$parser_reserved_atom'('inter').
'$parser_reserved_atom'('diff').
%'$parser_reserved_atom'('remember').
%'$parser_reserved_atom'('forget').
%'$parser_reserved_atom'('forgetall').
%'$parser_reserved_atom'('replace').
%'$parser_reserved_atom'('while').
%'$parser_reserved_atom'('until').


get_tokens(Stream, T) :-
    '$read_next_token'(Stream, T1, V1),
    get_tokens_aux(Stream, T1, V1, T).


get_tokens_aux(_Stream, T1, _V1, T) :- T1 = 8, !, T = [].
get_tokens_aux(Stream, T1, V1, T) :-
    T = [t(T1,V1)|Rest],
    '$read_next_token'(Stream, T2, V2),
    get_tokens_aux(Stream, T2, V2, Rest).

'$warn_of_single_vars'(T, Singles, VarNames) :-
    '$vars2names'(T, VarNames),
    '$warn_of_single_vars1'(T, Singles),
    fail.
'$warn_of_single_vars'(_, _, _).

'$warn_of_single_vars1'(_T, []) :- !.
'$warn_of_single_vars1'(T, Singles) :-
    '$warn_of_single_vars_aux'(Singles, Vars),
    write('Singleton variables warning: '),
    write(Vars),
    write(' in'),
    nl,
    writeTerm(T),
    nl,
    fail.

'$warn_of_single_vars_aux'([], []).
'$warn_of_single_vars_aux'([(_V = VN)|Singles], [VN|Vars]) :-
    %V = VN,
    '$warn_of_single_vars_aux'(Singles, Vars).

'$build_tuple'(['$tuple'(X)], '$tuple'(X)) :-
    !.
'$build_tuple'([X], X) :- !.
'$build_tuple'(X, '$tuple'(X)).

'$build_tuple_type'(['$tuple_type'(X)], '$tuple_type'(X)) :-
    !.
'$build_tuple_type'([X], X) :- !.
'$build_tuple_type'(X, '$tuple_type'(X)).

report_error(Msg, Rest) :-
    '$parser_info'(String, StartLine, EndLine, FileName),
    nl,
    add_here(String, Rest, HereString),
    (
     FileName = stdin
    ->
     write_term_list(stderr, [nl, 'Syntax Error: lines ', StartLine - EndLine,
			      nl, Msg, nl])
    ;
     write_term_list(stderr, [nl, 'Syntax Error: file ', FileName,
			      ' : lines ', StartLine - EndLine,
			      nl, Msg, nl])
     ),
    string_to_list(HereString, HereList),
    forall(member(Ch, HereList), put_code(stderr, Ch)),
    nl,nl,
    throw(syntax_error).

add_here(String, Rest, HereString) :-
    string_concat(S1, S2, String),
    string2tokens(S2, Rest, _VarList),
    !,
    string_concat(S1, " ###here### ", S3),
    (
      string_concat(S4, "\n", S2)
    ->
      true
    ;
      S4 = S2
    ),
    string_concat(S3, S4, HereString).
add_here(String, _Rest, String).
	    
						 
'$parser_list2tuple'([X], X) :- !.
'$parser_list2tuple'([X|Y], (X, Ys)) :-
    '$parser_list2tuple'(Y, Ys).

'$attach_minus'(T1, T2) :-
    integer(T1), !,
    T2 is -T1.
'$attach_minus'(T1, T2) :-
    number(T1), !,
    T2 is -T1.
'$attach_minus'(T1, T2) :-
    compound(T1), T1 = '$tuple'([T3]), !,
    T2 = -T3.
'$attach_minus'(T1, T2) :-
    T2 = -T1.

'$flip_apply'(F, A, T) :-
    list(A),!,
    T = F(A).
'$flip_apply'(F, A, T) :-
    \+compound(A),!,
    T = F(A).
'$flip_apply'(F, A, T) :-
    arith(A), !,
    T = F(A).
'$flip_apply'(F, A, T) :-
    A =.. [G|Args],
    (
      var(G)
    ->
      T =.. [F(G)|Args]
    ;
      '$parser_builtin_function'(G)
    ->
      T = F(A)
    ;
      T =.. [F(G)|Args]
    ).

'$qbuild_structure'(F, '$tuple'([]), Term) :-
    !,
    Term = F('$none_').
'$qbuild_structure'(F, '$tuple'(Tuple), Term) :-
    Term =.. [F|Tuple].

'$process_arith_tuples'(F, A1, B, Term) :-
    compound(A1), A1 = '$tuple'([A]),
    !,
    '$process_arith_tuples'(F, A, B, Term).
'$process_arith_tuples'(F, A, B1, Term) :-
    compound(B1), B1 = '$tuple'([B]),
    !,
    '$process_arith_tuples'(F, A, B, Term).
'$process_arith_tuples'(F, A, B, Term) :-
    Term = F(A,B).

'$process_set_tuples'(F, A1, B, Term) :-
    compound(A1), A1 = '$tuple'([A]),
    !,
    '$process_set_tuples'(F, A, B, Term).
'$process_set_tuples'(F, A, B1, Term) :-
    compound(B1), B1 = '$tuple'([B]),
    !,
    '$process_set_tuples'(F, A, B, Term).
'$process_set_tuples'(F, A, B, Term) :-
    Term = F(A,B).

allowed_term_as_predication(T) :-
    compound(T),
    functor(T, F, _),
    nonvar(F), !,
    F \= not, 
    F \= once,
    F \= '$tuple'.

allowed_term_as_predication(_).

allowed_term_as_string_predication(T) :-
    compound(T),!,
    functor(T, F, _),
    F \= not, 
    F \= once,
    F \= '$tuple',
    F \= '<>',
    F \= '++'.
allowed_term_as_string_predication(_).

'$parser_builtin_function'('<>').
'$parser_builtin_function'('++').
'$parser_builtin_function'('{}').
'$parser_builtin_function'('#').
'$parser_builtin_function'('+').
'$parser_builtin_function'('-').
'$parser_builtin_function'('*').
'$parser_builtin_function'('/').
'$parser_builtin_function'('::').
'$parser_builtin_function'('<=').
'$parser_builtin_function'('->').
'$parser_builtin_function'('~>').
'$parser_builtin_function'('~>>').
'$parser_builtin_function'('?').


%%%%%%%%%%%%%%%%%%%%%%
%% Parser
%%%%%%%%%%%%%%%%%%%%%%

error_handler(Msg) --> Rest, {report_error(Msg, Rest)}, consume_rest.
error_handler(Msg, Rest) --> {report_error(Msg, Rest)}, consume_rest.

cb_error_handler(cb(']')) -->
    Rest, {report_error('Expecting ]', Rest)}, consume_rest.
cb_error_handler(cb(')')) -->
    Rest, {report_error('Expecting )', Rest)}, consume_rest.
cb_error_handler(cb('}')) -->
    Rest, {report_error('Expecting }', Rest)}, consume_rest.

consume_rest --> [_], consume_rest.
consume_rest --> [].

has_token(T) --> Tokens, {member(T, Tokens)}.
next_token(T) --> Tokens, {Tokens = [T|_]}.

leftover -->
    Tokens, {Tokens \= []}, error_handler('Invalid syntax', Tokens).

no_leftover -->
    Tokens,
    ( {Tokens \= []} -> error_handler('Invalid syntax', Tokens) ; [] ).

at_end --> Tokens, {Tokens = []}.

parse_to_end_type(T) -->
    (
      '$type_rule'(T)
    ->
      no_leftover
    ;
      error_handler('Expecting type')
    ).            

parse_to_end_term(T) -->
    (
      term(T)
    ->
      no_leftover
    ;
      error_handler('Expecting term')
    ).
      


parse_to_end_goal(G) -->
    (
      goal(G)
    ->
      no_leftover
    ;
      error_handler('Expecting goal')
    ).

parse_to_end_action_body(G) -->
    (
      action_goal(G)
    ->
      no_leftover
    ;
      error_handler('Expecting action')
    ).

type_with_check(Type) -->
    (
      '$type_rule'(Type)
    ->
      []
    ;
      error_handler('Invalid type')
    ).

term_with_check(Term) -->
    (
      term(Term)
    ->
      []
    ;
      error_handler('Invalid term')
    ).

predication_with_check(G) -->
    (
      predication(G)
    ->
      []
    ;
      error_handler('Expecting predication')
    ).

predication_not_with_check(G) -->
    (
      predication_not(G)
    ->
      []
    ;
      error_handler('Expecting predication or negated predication')
    ).

predication_goals_with_check(G) -->
    (
      predication_goals(G)
    ->
      []
    ;
      error_handler('Expecting predication goals')
    ).

term_expected(Term) -->
    (
      term(Term)
    ->
      []
    ;
      error_handler('Expecting term')
    ).


atom_with_check(G) -->
    (
      atom(G)
    ->
      []
    ;
      error_handler('Expecting atom')
    ).

goal_with_check(G) -->
    (
      goal(G)
    ->
      []
    ;
      error_handler('Invalid goal')
    ).

var_with_check(Var) -->
    (
      var(Var)
    ->
      []
    ;
      error_handler('Expecting variable')
    ).

matching_bracket(CB) -->
    (
      [CB]
    ->
      []
    ;
      cb_error_handler(CB)
    ).

%% --------- EBNF for qulog ------------------
%% file_term = declaration | clause | type_definition
%% -------------------------------------------

file_term('?-'(G)) --> [op('?-')], term(G).
file_term(T) --> declaration(T), !.
file_term(T) --> has_token(op('::=')), !, type_definition(T).
file_term(T) -->
    basic_term(T1), 
    clause(T1, T).

%% --------- EBNF for queries ------------------
%% query = (term | goal) ['.']
%% -------------------------------------------

%query('$num_exists_vars_query'(N, Vars, T)) --> int(N), atom(exists), var_list(Vars), !, query_aux(T).
query('$num_vars_query'(N, Vars, T)) -->
    int(N), 
    var(V),
    (
     [punct(',')]
    ->
     var_list(Rest),
     {Vars = [V|Rest]}
    ;
     {Vars = [V]}
    ),
    [op('?')], !, query_aux(T).
query('$num_query'(N, T)) --> int(N), [op('?')], !, query_aux(T).
%query('$exists_vars_query'(Vars, T)) --> atom(exists), var_list(Vars), !, query_aux(T).
query('$vars_query'(Vars, T)) -->
    var(V),
    (
     [punct(',')]
    ->
     var_list(Rest),
     {Vars = [V|Rest]}
    ;
     {Vars = [V]}
    ),
    [op('?')], !, query_aux(T).
query(T) --> query_aux(T).
query_aux(prolog) --> atom(prolog).
query_aux(prolog(T)) --> atom(prolog), !, goal(T).
query_aux(types) --> atom(types).
query_aux(types(T)) --> atom(types), [ob('(')], term(T), [cb(')')], !.
query_aux(types(T)) --> atom(types), !, atom_tuple(TT), {'$tuple2list'(TT, T)}.
query_aux(stypes) --> atom(stypes).
query_aux(stypes(T)) --> atom(stypes), [ob('(')], term(T), [cb(')')], !.
query_aux(stypes(T)) --> atom(stypes), !, atom_tuple(TT),
                          {'$tuple2list'(TT, T)}.
%query_aux(types(T)) --> atom(type), [ob('(')], term(T), [cb(')')], !.
%query_aux(types(T)) --> atom(type), \+[ob('(')], term(T), !.
query_aux(answers(N)) --> atom(answers), !, int(N).
query_aux(show) --> atom(show).
query_aux(show(T)) --> atom(show), [ob('(')], atom(T), [cb(')')], !.
query_aux(show(T)) --> atom(show), !, atom_tuple(TT), {'$tuple2list'(TT, T)}.
query_aux(consult(T)) --> atom(consult), [ob('(')], term(T), [cb(')')], !.
query_aux(consult(T)) --> atom(consult), term(T), !.
query_aux(pconsult(T)) --> atom(pconsult), [ob('(')], term(T), [cb(')')], !.
query_aux(pconsult(T)) --> atom(pconsult), term(T), !.
query_aux(T) --> goal(T).
query_aux(T) --> action_goal(T).
query_aux('::'(A, B)) -->
    term(A), 
    [op('::')], !,
    goal(B).
query_aux(T) --> term(T).



%% --------- EBNF for declaration -------------
%% declaration = declaration_type , 
%%                  [ ( "->" , type ) | "<="  | "~>" | "->?" ];
%% declaration_type = atom_tuple , ":",
%%                   ( ( any_open_bracket , moded_type_list , ")" ) |
%%                    moded_type_list );
%% moded_type_list = moded_type , { "," , moded_type };
%% moded_type =   ( "^" , type ) | ( "?" , type ) | type;
%% -------------------------------------------
declaration(belief(B)) -->
    atom('belief'), !,
    belief_percept_decl(B).
declaration(percept(B)) -->
    atom('percept'), !,
    belief_percept_decl(B).

declaration(durative(D)) -->
    atom('durative'), !,
    belief_percept_decl(D).

declaration(discrete(D)) -->
    atom('discrete'), !,
    belief_percept_decl(D).

declaration(resources(D)) -->
    atom('resources'), !,
    dur_dis_type_list(D).
    
declaration(task_start(D)) -->
    atom('task_start'),
    (
     type_declaration(D)
    ;
     dur_dis_type_list(D)
    ), !.
   
declaration(task_atomic(D)) -->
    atom('task_atomic'), !,
    (
     type_declaration(D)
    ;
     dur_dis_type_list(D)
    ), !.
    
declaration(global_int_decl(A, N)) -->
    atom('int'), atom(A), [op(':=')], !,
    (
      int(N)
    ->
      []
    ;
      error_handler('Expecting int')
    ).
declaration(global_num_decl(A, N)) -->
    atom('num'), atom(A), [op(':=')], !,
    (
      ( num(N) ; int(N) )
    ->
      []
    ;
      error_handler('Expecting num')
    ).


declaration(D) -->
    type_declaration(D).
type_declaration((A:T)) -->
    atom_tuple(A), [op(':')], !,
    declaration_type_part(T1),
    { (T1 = [T] -> true ; T = alt_types(T1)) },
    no_leftover.

declaration_type_part(T1) -->
    (
      declaration_type(Head)
    ->
      (
        [op('->')],
        !,
        '$type_rule'(Type),
        { T = (Head -> Type) }
      ;
        [op('<=')],
        !,
        { T = (Head <=) }
      ;
       [op('~>>')],
        !,
        { T = (Head ~>>) }
      ;        
       [op('~>')],
        !,
        { T = (Head ~>) }
      ;        
        {T = Head}
      )
    ;
      error_handler('Invalid type declaration')
    ),
    (
      [op('|')]
    ->
      declaration_type_part(T2),
      { T1 = [T|T2] }
    ;
      { T1 = [T]}
    ).
    

declaration_type(Type) -->
    any_open_bracket,
    (
      [cb(')')], !,
      { Type = '$tuple_type'([])}
    ;
      declaration_type_list(Types),
      {'$build_tuple_type'(Types, Type)},
      [cb(')')], !
    ).

declaration_type(Type) -->
    declaration_type_list(Types),
    {  '$build_tuple_type'(Types, Type)}.


declaration_type_list(Result) -->
    moded_type(Type),
    (
      [punct(',')], !,
      Tokens,
      (
        [cb(_)]
      ->
        error_handler('Unexpected closing bracket', Tokens)
      ;
        declaration_type_list(Types),
        {Result = [Type|Types]}
      )
    ;
      {Result = [Type]}
    ).

%moded_type(T) -->  any_open_bracket, moded_type(T), [cb(')')], !.
%moded_type(^(T)) --> [op('^')], !, type_with_check(T).
%moded_type(? T) --> [op('?')], !, type_with_check(T).
%moded_type('!'(T)) --> [op('!')], !, type_with_check(T).
moded_type(T) -->  type_with_check(T).


%% --------- EBNF for clause -------------
%% clause = relation_clause | function_clause 
%% relation_clause =  atom_or_structure |
%%                    ( atom_or_structure ,
%%                        ( ( "::" , predication_goals , "<=" , goal ) |
%%                          ( "::" , predication_goals) |
%%                          "::" |
%%                          ("<=" , goal )
%%                        )
%%                    )
%% action_clause =  atom_or_structure |
%%                    ( atom_or_structure ,
%%                        ( ( "::" , predication_goals , "~>" , goal ) |
%%                          ( "::" , predication_goals) |
%%                          "::" |
%%                          ("~>" , goal )
%%                        )
%%                    )
%% function_clause =  ( atom_or_structure ,
%%                        ( ( "::" , predication_goals , "->" , term) |
%%                          ("->" , term, )
%%                        )
%%                    )            
%% -------------------------------------------

clause(Head, Clause) -->
    (
      [op('::')],
      !,
      (
        predication_goals(Test), !,
        (
          [op('->')], 
          !,
          parse_to_end_term(Body),
          {Clause = (Head :: Test -> Body)}
        ;
          [op('<=')],
          !,
          parse_to_end_goal(Body),
          {Clause = (Head :: Test <= Body)}
        ;
          [op('~>>')],
          !,
          parse_to_end_action_body(Body),
          {Clause = (Head :: Test ~>> Body)}
         ;
          leftover, !
        ;
          {Clause = (Head :: Test)}
        )
      ;
        at_end, !,
        {Clause = ::(Head)}
      ;
        error_handler('Expecting predication goals')
      )
    ;
      [op('->')],
      !,
      parse_to_end_term(Body),
      {Clause = (Head -> Body)}
     ;
      [op('<=')],
      !,
      parse_to_end_goal(Body),
      {Clause = (Head <= Body)}
     ;
      [op('~>>')],
      !,
      parse_to_end_action_body(Body),
      {Clause = (Head ~>> Body)}
    ;
     [ob('{')],
     !,
     tr_rules(Rules), 
     [cb('}')],
     {Clause = '$tr_prog'(Head, Rules)}
    ;
      no_leftover,
      {Clause = Head}
    ).

tr_rules(Rules) -->
    tr_rule(R),!,
    (
     \+[cb('}')]
    ->
     tr_rules(Rest),
     {Rules = [R|Rest]}
    ;
     {Rules = [R]}
    ).

tr_rule(R) -->
    tr_rule_LHS(G), 
    (
     [op('~>')]
    ->
     tr_rule_RHS(A),!,
     {R = '~>'(G, A)}
    ;
     error_handler('Expecting ~>')
    ).

tr_rule_LHS(G) -->
    predication_goals(G1),
    tr_rule_LHS_aux(G, G1).
tr_rule_LHS_aux(G, G1) -->
    [atom(min)], !,
    (
      term(T)
    ->
      { G = '$time'(min, G1, T) }
    ;
      error_handler('Expecting term')
    ).
tr_rule_LHS_aux(G, G1) -->
    [atom(for)], !, 
    (
      term(T)
    ->
      { G = '$time'(for, G1, T) }
    ;
      error_handler('Expecting term')
    ).
/*
  tr_rule_LHS_aux(G, G1) -->
    [op('|')], !, 
    (
      term(T)
    ->
      { G = '$time'('|', G1, T) }
    ;
      error_handler('Expecting term')
    ).
*/
tr_rule_LHS_aux(G, G1) -->
    [atom(while_until)], !, 
    while_goal_part(G2),
    { G = '$while_until'(while_until, G1, G2)}.
/*
tr_rule_LHS_aux(G, G1) -->
    [op('~^')], !, 
    while_goal_part(G2),
    { G = '$while_until'('~^', G1, G2)}.
*/
tr_rule_LHS_aux(G, G1) -->
    [atom(while)], !, 
    while_part(while, G1, G).
/*
tr_rule_LHS_aux(G, G1) -->
    [op('~~')], !, 
    while_part('~~', G1, G).
*/
tr_rule_LHS_aux(G, G1) -->
    [atom(until)], !, 
    until_part(U),
    {G = '$until'(until, G1, U)}.
/*
  tr_rule_LHS_aux(G, G1) -->
    [op('^^')], !, 
    until_part(U),
    {G = '$until'('^^', G1, U)}.
*/
tr_rule_LHS_aux(G, G1) -->
    {G = G1}.
      
while_part(Op, G, W) -->
    while_goal_part(WG), 
    (
      [atom(until)]
    ->
     until_part(U),
     {W = '$until'(until, '$while'(Op, G, WG), U)}
/*
  ;
      [op('^^')]
    ->
     until_part(U),
     {W = '$until'('^^', '$while'(Op, G, WG), U)}
*/
    ;
      {W = '$while'(Op, G, WG)}
    ).

while_goal_part(WG) -->
    [atom(min)], !,
    term(T),
    {WG = '$time'(min, T)}.
/*
  while_goal_part(WG) -->
    [op('|')], !,
    term(T),
    {WG = '$time'('|', T)}. 
*/
while_goal_part(WG) -->
    single_goal(G), 
    (
      [atom(min)]
    ->
      term(T),
      {WG = '$time'(min, G, T)}
/*
  ;
      [op('|')]
    ->
      term(T),
      {WG = '$time'('|', G, T)}
*/
    ;
      {WG = G}
    ).
        
until_part(U) -->
    until_goal_part(UG),
    (
      [atom(until)]
    ->
      until_part(U),
      {U = '$until'(until, UG, U)}
/*
  ;
      [op('^^')]
    ->
      until_part(U),
      {U = '$until'('^^', UG, U)}
*/
    ;
      {U = UG}
    ).

/*
until_goal_part(UG) -->
    [op('|')], !,
    basic_term(T),
    {UG = '$time'('|', T)}.
*/
until_goal_part(UG) -->
    [atom(min)], !,
    term(T),
    {UG = '$time'(min, T)}.

until_goal_part(UG) -->
    single_goal(G),
    (
      [atom(min)]
    ->
      basic_term(T),
      {UG = '$time'(min, G, T)}

    /*;
      [op('|')]
    ->
      basic_term(T),
      {UG = '$time'('|', G, T)}
*/
    ;
      {UG = G}
    ).

tr_rule_RHS(Action) -->
    atom(wait), !,
    {ip_set(parse_pp, no)}, term(T), {ip_set(parse_pp, yes)},
    {Action = '$wait_repeat'(['$nil'], T)}.
/*
tr_rule_RHS('$nil') -->
    any_open_bracket,
    [cb(')')], !.
*/
tr_rule_RHS(Action) -->
    (
      any_open_bracket
    ->
      {HasBracket = true}
    ;
      {HasBracket = false}
    ),
   tr_action(A, HasBracket, FoundClose),
    (
      {HasBracket = true, FoundClose = false}
    ->
      (
        [cb(')')]
      ->
        true
      ;
        error_handler('Expecting closing bracket')
      )
    ;
      []
    ),
     (
       atom(wait)
     ->
       wait_repeat(WR),
       { Action = '$wait_repeat'(A, WR) }
     ;
      [op('++')]
    ->
       single_goal(QA),
       (
         \+[op(';')]
       ->
         { Action = '++'(A, QA) }
       ;
         error_handler('Unexpected ;')
       )
    ;
      { Action = A }
    ).

tr_action(Action, HasBracket, FoundClose) -->
    tr_action_basic(Action1), !,
    (
      [cb(')')]
    ->
      (
        {HasBracket = false, Action1L = Action1}
      ->
        error_handler('Unexpected closing bracket')
      ;
        {FoundClose = true,
         (   Action1 = '$tuple'(Action1L)
         ->
             true
         ;
             Action1L = Action1
         )
         }
      )
    ;
      {Action1L = Action1}
    ),
    (
      [atom(for)]
    ->
      (
        \+[op(';')]
      ->
        term(T)
      ;
        error_handler('Expecting time')
      ),
      (
        [op(';')]
      ->
        tr_timed_seq(Action2),
        {Action = timed_seq(['$time'(for, Action1L, T)|Action2])}
      ;
        {Action = timed_seq(['$time'(for, Action1L, T)])}
      ),
      (
        [cb(')')]
      ->
        (
          {HasBracket = false}
        ->
          error_handler('Unexpected closing bracket')
        ;
          {FoundClose = true}
        )
      ;
        []
      )
     /*
    ;
      atom(time)
    ->
      (
        \+[op(';')]
      ->
        term(T)
      ;
        error_handler('Expecting time')
      ),
      (
        [op(';')]
      ->
        tr_timed_seq(Action2),
        {Action = timed_seq(['$time'(time, Action1L, T)|Action2])}
      ;
        {Action = timed_seq(['$time'(time, Action1L, T)])}
      ),
      (
        [cb(')')]
      ->
        (
          {HasBracket = false}
        ->
          error_handler('Unexpected closing bracket')
        ;
          {FoundClose = true}
        )
      ;
        []
      )
       */
    ;
      {Action = Action1L}
    ).

tr_action_basic(Action) -->
    TK, {TK = [cb(')')|_]}, !,
    {Action = ['$nil']}.
tr_action_basic(Action) -->
    {ip_set(parse_pp, no)},
    single_goal(G0),  %{G0 \= '++'(_,_)},
    {
     (
       G0 = '$tuple'(G)
     ->
       true
     ;
       G = [G0]
     )
    },
    (
      [punct(',')]
    ->
      tr_action_basic(Action1),
      {append(G,Action1,Action) }
    ;
      {Action = G}
    ).

wait_repeat(WR) -->
    {ip_set(parse_pp, no)}, term(T),  {ip_set(parse_pp, yes)},
    (
      atom(repeat)
    ->
      {ip_set(parse_pp, no)}, term(R),  {ip_set(parse_pp, yes)},
      { WR = '$repeat'(repeat, T, R) }
    ;
      [op('^')]
    ->
      {ip_set(parse_pp, no)}, term(R),  {ip_set(parse_pp, yes)},
      { WR = '$repeat'('^', T, R) }
    ;
      { WR = T }
    ).
      
tr_timed_seq(Action) -->
    tr_action_basic(G), 
    (
     /*
      [op('|')]
    ->
      {ip_set(parse_pp, no)}, term(T),  {ip_set(parse_pp, yes)},
     (
        [op(';')]
      ->
        tr_timed_seq(Action2),
        {Action = ['$time'('|', G, T)|Action2]}
      ;
        {Action = ['$time'('|', G, T)]}
      )
    ;
       */
      atom(for)
    ->
      {ip_set(parse_pp, no)}, term(T),  {ip_set(parse_pp, yes)},
     (
        [op(';')]
      ->
        tr_timed_seq(Action2),
        {Action = ['$time'(for, G, T)|Action2]}
      ;
        {Action = ['$time'(for, G, T)]}
      )
    ;
      {Action = [G]}
    ).

%% --------- EBNF for type definitions -------------
%% type_definition = definition_term , ":=" , definition_type
%% definition_type = ( type | constructor_type )
%% definition_term = atom , [ "(" , var , { "," , var } , ")" ]
%% constructor_type = structure , { "|" , structure }
%% -------------------------------------------------

type_definition((T ::= Type)) -->
    definition_term(T), 
    [op('::=')], !,
    (
      definition_type(Type)
    ->
      no_leftover
    ;
       error_handler('Expecting type or type constructor')
    ).

definition_type(T) --> constructor_type(T), !.
%definition_type(T) --> struct(T1), TK, {TK = []}, !, {T = '$union_type'([T1])}.
definition_type(T) --> atom(T1), TK, {TK = []}, !, {T = T1}.
definition_type('$enum_type'([T | Rest])) -->
    atom(T), [op('|')], !, rest_enum_atom_type(Rest).
definition_type('$enum_type'([T | Rest])) -->
    string(T), [op('|')], !, rest_enum_string_type(Rest).
definition_type('$enum_type'([T | Rest])) -->
    anynum(T), [op('|')], !, rest_enum_num_type(Rest).
definition_type((T1 .. T2)) -->
    any_open_bracket, int(T1),
    [op('..')], !,
    int(T2),
    matching_bracket(cb(')')).
definition_type((T1 .. T2)) -->
    int(T1),
    [op('..')], !,
    int(T2).

definition_type(T) -->
    '$type_rule'(T).

constructor_type('$constr_enum_type'(T)) -->
   type_struct(T1),
    [op('|')], !,
    constructor_type_aux(T1, T).

constructor_type_aux(T1, T) -->
    (
      type_struct(T2)
    ->
      (
        [op('|')]
      ->
        constructor_type_aux(T2, T3),
        {T = [T1 | T3]}
      ;
        at_end, !,
        {T = [T1, T2]}
      ;
        error_handler('Expecting \'|\'')
      )
    ;
      error_handler('Expecting type constructor')
    ).

definition_term(Struct) -->
    atom(R),
    (
      [ob('(')], !,
      var_list(Args),
      {Struct =.. [R|Args]},
      (
        [cb(')')]
      ->
        []
      ;
        error_handler('Expecting \')\' or \',\'')
      )
    ;
      {Struct = R}
    ).

var_list(Result) -->
    var_with_check(V),
    (
      [punct(',')], !,
      Tokens,
      (
        [cb(_)]
      ->
        error_handler('Unexpected closing bracket', Tokens)
      ;
        []
      ),
     var_list(Vars),
     {Result = [V|Vars]}
    ;
     {Result = [V]}
    ).
  
%% --------- EBNF for goals ------------------------
%% goal = ( single_goal , "&" , goal ) | single_goal
%% predication = atom | ( atom , "(" , term , { "," , term } , ")" )
%% predication_not = predication | ("not" , predication ) |
%%                   ("not" , "(" , predication , ")" )
%% predication_goals = predication_not |
%%                     ( predication_not , "&" , predication_goals)
%% 
%% single_goal =  predication_not |
%%               ( "once" ,  predication_not) |
%%               ( "once" ,  "(" , predication_goals , ")") |
%%               ( for_all , at_var_list , "::" , "(" , 
%%                    predication_goals , "=>" , predication_goals , ")" ) |
%%               ( "forall" , "(", ( var |
%%                                 ( "(" , var , { "," , var }  , ")" ) ) ,
%%                 "," , predication_goals , "," ,
%%                 predication_goals , ")" ) |
%%               ( term ,
%%                 ( ">" | ">=" | "<" | "=<" |
%%                   "=" | "=!" | "in" ),
%%                 term )
%% at_var_list = ( "(" , var , { "," , var } ")" ) |
%%               ( var , { "," , var } )
%% for_all = "@" | "forall"
%% -------------------------------------------------
    
goal(G) -->
    single_goal(Body1),
    (
      [op('&')]
    ->
      goal(Body2),
      { G = (Body1 & Body2) }
    ;
      {G = Body1}
    ).
action_goal(G) -->
    single_goal(Body1), 
    (
      \+ \+ [op('&')]
    ->
      error_handler('Expecting ; not &')
    ;
      [op(';')]
    ->
      action_goal(Body2),
      { G = ';'(Body1, Body2) }
    ;
      {G = Body1}
    ).

single_goal(type(A, T)) --> atom(type), 
    [ob('(')], term1(A), [punct(',')], '$type_rule'(T),
    matching_bracket(cb(')')).
single_goal(type(A, T, V)) --> atom(type), !,
    [ob('(')], term1(A), [punct(',')], '$type_rule'(T),
    [punct(',')], term1(V), matching_bracket(cb(')')).
single_goal(':='(A, I)) --> atom(A), [op(':=')], !, term1(I).
single_goal('+:='(A, I)) --> atom(A), [op('+:=')], !, term1(I).
single_goal('-:='(A, I)) --> atom(A), [op('-:=')], !, term1(I).

single_goal(not(Goal)) --> atom(not), !,
    (
      any_open_bracket
    ->
      predication_with_check(Goal),
      matching_bracket(cb(')'))
    ;
      predication_with_check(Goal)
    ).
single_goal(once(Goal)) --> atom(once), !,
    (
      any_open_bracket
    ->
      predication_goals_with_check(Goal),
      matching_bracket(cb(')'))
    ;
      predication_not_with_check(Goal)
    ).
single_goal(watch(P)) --> atom(watch), !,
    (
      any_open_bracket
    ->
      atom(P),
      matching_bracket(cb(')'))
    ;
      atom(P)
    ).
single_goal(set_num_answers(P)) --> atom(set_num_answers), !,
    (
      any_open_bracket
    ->
      int(P),
      matching_bracket(cb(')'))
    ;
      int(P)
    ).
single_goal(unwatch(P)) --> atom(unwatch), !,
    (
      any_open_bracket
    ->
      atom(P),
      matching_bracket(cb(')'))
    ;
      atom(P)
    ).
single_goal(watchC(P)) --> atom(watchC), !,
    (
      any_open_bracket
    ->
      atom(P),
      matching_bracket(cb(')'))
    ;
      atom(P)
    ).
single_goal(unwatchC(P)) --> atom(unwatchC), !,
    (
      any_open_bracket
    ->
      atom(P),
      matching_bracket(cb(')'))
    ;
      atom(P)
    ).
single_goal(G) --> atom(remember), !,
    (
      any_open_bracket
    ->
      predication_with_check(Goal),
      matching_bracket(cb(')'))
    ;
      predication_with_check(Goal)
    ),
    (
     atom(for)
    ->
     term1(T), {G = remember_for(Goal, T)}
    ;
     {G = remember(Goal)}
    ).
single_goal(G) --> atom(rememberA), !,
    (
      any_open_bracket
    ->
      predication_with_check(Goal),
      matching_bracket(cb(')'))
    ;
      predication_with_check(Goal)
    ),
    (
     atom(for)
    ->
     term1(T), {G = rememberA_for(Goal, T)}
    ;
     {G = rememberA(Goal)}
    ).
single_goal(G) --> atom(forget), !,
    (
      any_open_bracket
    ->
      predication_with_check(Goal),
      matching_bracket(cb(')'))
    ;
      predication_with_check(Goal)
    ),
    (
     atom(after)
    ->
     term1(T), {G = forget_after(Goal, T)}
    ;
     {G = forget(Goal)}
    ).

single_goal(replace_by(Goal1, Goal2)) --> atom(replace), !,
      predication_with_check(Goal1),
      (
        atom(by)
      ->
        []
      ;
        error_handler('Expecting \'by\'')
      ),
      predication_with_check(Goal2).
single_goal(replaceA_by(Goal1, Goal2)) --> atom(replaceA), !,
      predication_with_check(Goal1),
      (
        atom(by)
      ->
        []
      ;
        error_handler('Expecting \'by\'')
      ),
      predication_with_check(Goal2).
/*
single_goal(teleor(Handle, EnvHandle)) -->
    atom(teleor),[ob('(')],!,
    (
      atom(Handle)
    ->
      []
    ;
      error_handler('Expecting name of agent (atom)')
    ),
    (
      [punct(',')]
    ->
      (
        start_agent_handle_term(EnvHandle)
      ->
        []
      ;
        error_handler('Expecting handle')
      )
    ;
      error_handler('Expecting comma')
    ),
    matching_bracket(cb(')')).
 */
single_goal(start_task(Name, TR)) -->
    atom(start_task), !,
    (
      atom(Name)
    ->
      []
    ;
     var(Name)
    ->
     []
    ;
      error_handler('Expecting name of task (atom or variable)')
    ),
    term1(TR).
single_goal(kill_task(Name)) -->
    atom(kill_task), !,
    (
      atom(Name)
    ->
      []
    ;
     var(Name)
    ->
     []
    ;
      error_handler('Expecting name of task (atom or variable)')
    ).


single_goal(actions(Actions)) -->
    atom(actions), !,
    (
      list(Actions)
    ->
      []
    ;
      atom([])
    ->
      {Actions = []}
    ;
      error_handler('Expecting list of actions')
    ).
single_goal(start_agent(Handle, EnvHandle, Interface)) -->
    atom(start_agent), !,
    (
      atom(Handle)
    ->
      []
    ;
     var(Handle)
    ->
     []
    ;
      error_handler('Expecting name of agent (atom)')
    ),
    (
      start_agent_handle_term(EnvHandle)
    ->
      []
    ;
      error_handler('Expecting handle')
    ),
    (
      atom(all)
    ->
     {Interface = all}
    ;
     atom(updates)
    ->
     {Interface = updates}
    ;
     atom(user)
    ->
     {Interface = user}
    ;
     error_handler('Expecting all, updates or user')
    ).
single_goal(start_drive(Handle, EnvHandle)) -->
    atom(start_drive), !,
    (
      atom(Handle)
    ->
      []
    ;
     var(Handle)
    ->
     []
    ;
      error_handler('Expecting name of agent (atom)')
    ),
    (
      start_agent_handle_term(EnvHandle)
    ->
      []
    ;
      error_handler('Expecting handle')
    ).
single_goal(log(Handle)) -->
    atom(log),[ob('(')],!,
    (
      start_agent_handle_term(Handle)
    ->
      []
    ;
      error_handler('Expecting handle of logger')
    ),
    matching_bracket(cb(')')).
single_goal(log(Handle)) -->
    atom(log),!,
    (
      start_agent_handle_term(Handle)
    ->
      []
    ;
      error_handler('Expecting handle of logger')
    ).
                    
single_goal(mode_correct(Goal)) -->
    atom(mode_correct), !,
    (
     any_open_bracket
    ->
     single_goal(Goal),
     matching_bracket(cb(')'))
    ;
    single_goal(Goal)
    ).
single_goal(call(Goal)) -->
    atom(call), !,
    (
     any_open_bracket
    ->
     single_goal(Goal),
     matching_bracket(cb(')'))
    ;
    single_goal(Goal)
    ).
/*
single_goal(do(Goal)) -->
    atom(do), !, 
    (
     any_open_bracket
    ->
     single_goal(Goal),
     matching_bracket(cb(')'))
    ;
    single_goal(Goal)
    ).
*/
single_goal(to(Msg, Addr)) -->
    term1(Msg), atom(to), !, term1(Addr).
single_goal(from(Msg, Addr)) -->
    term1(Msg), atom(from), !,
    term1(Addr).
single_goal(fork_as(Goal1, Name)) -->
    atom(fork), !,
      predication_with_check(Goal1),
      (
        atom(as)
      ->
        []
      ;
        error_handler('Expecting \'as\'')
      ),
      term1(Name).
/*
single_goal(FA(Vars, G1, G2)) --> 
    [op('@')],!,
    at_var_list(Vars),
    (
      any_open_bracket
    ->
      []
    ;
      error_handler('Expecting \'(\'')
    ),
    predication_goals(G1),
    (
      [op('=>')]
    ->
      {FA = '$at_forall'},
      (
        predication_goals(G2)
      ->
        (
          [cb(')')]
        ->
          []
        ;
          error_handler('Expecting closing bracket')
        )
      ;
        error_handler('Expecting goal')
      )
    ;
      [op('~>>')]
    ->
      {FA = '$at_forall_actions'},
      (
        action_goal(G2)
      ->
        (
          [cb(')')]
        ->
          []
        ;
          error_handler('Expecting closing bracket')
        )
      ;
        error_handler('Expecting goal')
      )
    ;
      error_handler('Expecting => or ~>>')
    ).
  */
single_goal(FA(Vars, G1, G2)) --> 
    [op('forall')],!,
    at_var_list(Vars),
    (
      any_open_bracket
    ->
      []
    ;
      error_handler('Expecting \'(\'')
    ),
    predication_goals(G1),
    (
      [op('=>')], !,
      {FA = '$forall'},
      (
        goal(G2)
      ->
        (
          [cb(')')]
        ->
          []
        ;
          error_handler('Expecting closing bracket')
        )
      ;
        error_handler('Expecting goal')
      )
    ;
      [op('~>>')]
    ->
      {FA = '$forall_actions'},
      (
        action_goal(G2)
      ->
        (
          [cb(')')]
        ->
          []
        ;
          error_handler('Expecting closing bracket')
        )
      ;
        error_handler('Expecting goal')
      )
    ;
      error_handler('Expecting => or ~>>')
    ).
single_goal('$exists'(Vars, G)) -->
    [op('exists')],!,
    at_var_list(Vars),
    (
      any_open_bracket
    ->
      predication_goals(G),
      matching_bracket(cb(')'))
    ;
      predication_goals(G)
    ).
single_goal(G) --> predication(G).

predication(G) -->
    term1(T),!,
    (
      predication_rest(T, G)
    ;
      {G = T}
    ).

predication_rest(_, _) --> [op('~>')], !, {fail}.
predication_rest(A, A = B)       --> [op('=')], !, term_expected(B).
predication_rest(A, A > B)       --> [op('>')], !, term_expected(B).
predication_rest(A, A >= B)      --> [op('>=')], !, term_expected(B).
predication_rest(A, A < B)       --> [op('<')], !, term_expected(B).
predication_rest(A, A =< B)      --> [op('=<')], !, term_expected(B).
predication_rest(T1, (T1 = T2))  --> [op('=')], !, term_expected(T2).
predication_rest(T1, (T1 \= T2))  --> [op('\\=')], !, term_expected(T2).
predication_rest(A, A @> B)       --> [op('@>')], !, term_expected(B).
predication_rest(A, A @>= B)      --> [op('@>=')], !, term_expected(B).
predication_rest(A, A @< B)       --> [op('@<')], !, term_expected(B).
predication_rest(A, A @=< B)      --> [op('@=<')], !, term_expected(B).
predication_rest(T1, (T1 @= T2))  --> [op('@=')], !, term_expected(T2).
predication_rest(T1, (T1 $= T2)) --> [op('$=')], !, term_expected(T2).
predication_rest(T1, (T1 in T2)) --> [op('in')], !, term_expected(T2).
predication_rest(T1, (T1 '=?' T2)) --> [op('=?')], !, eqqrhs(T2).

predication_rest(A, A) --> {allowed_term_as_predication(A)}.

string_predication(A) --> term1(A), {allowed_term_as_string_predication(A)}.

predication_not(not(Goal)) -->
    atom(not), any_open_bracket, !,
    (
      predication(Goal)
    ->
      []
    ;
      error_handler('Expecting predication')
    ),
    matching_bracket(cb(')')).
predication_not(not(Goal)) -->
    atom(not), !,
    (
      predication(Goal)
    ->
      []
    ;
      error_handler('Expecting predication')
    ).
predication_not(Goal) -->
    (
      predication(Goal)
    ->
      []
    ;
      error_handler('Expecting predication')
    ).

predication_goals(G) -->
    predication_not(Body1),
    (
      [op('&')]
    ->
      predication_goals(Body2),
      { G = (Body1 & Body2) }
    ;
      {G = Body1}
    ).

%% RHS of T =? pattern

eqqrhs('@..'(T1,T2)) -->
    (var(T1);(basic_term(T1), {not_a_concat(T1)})), [op('@..')],
    !, basic_term(T2).
eqqrhs(T) -->
    eqqrhs_string(T), {nonvar(T), T = (_ ++ _)}, !.
eqqrhs(T) -->
    eqqrhs_append(T), {nonvar(T), T = (_ <> _)}, !.
eqqrhs(_T) -->
    error_handler('Expecting  @.. <> or ++ term').

eqqrhs_string(T) -->
    any_open_bracket, !,
    eqqrhs_string(T1),
    (
      [cb(')')]
    ->
      []
    ;
      error_handler('Expecting \')\'')
    ),
    (
      [op('++')]
    ->
      eqqrhs_string(T2),
      { T = (T1 ++ T2) }
    ;
      {T = T1}
    ).
eqqrhs_string(T) -->
     eqq_string_term(T1), [op('++')], !,
     eqqrhs_string(T2),
     { T = (T1 ++ T2) }.
eqqrhs_string(T) -->     
     eqq_string_term(T), !.

eqqrhs_append(T) -->
    any_open_bracket, !,
    eqqrhs_append(T1),
    (
      [cb(')')]
    ->
      []
    ;
      error_handler('Expecting \')\'')
    ),
    (
      [op('<>')]
    ->
      eqqrhs_append(T2),
      { T = (T1 <> T2) }
    ;
      {T = T1}
    ).
eqqrhs_append(T) -->     
     eqq_list_term(T1), [op('<>')], !,
     eqqrhs_append(T2),
     { T = (T1 <> T2) }.
eqqrhs_append(T) -->     
     eqq_list_term(T), !.

/*
  eqq_term_aux(T1, '@..'(T1,T)) --> [op('@..')], !, basic_term(T).
  eqq_term_aux(T1, T) --> [op('++')], !, eqq_term_concat(T1, T).
  eqq_term_aux(T1, T) --> [op('<>')], !, eqq_term_append(T1, T).
  eqq_term_aux(_T1, _T) --> error_handler('Expecting @.. <> or ++').

eqq_term_concat(T1, T)  -->
    (
      eqq_string_term(A1)
    ->
      (
        [op('++')], !,
        eqq_term_concat(A1, B),
        { T = (T1 ++ B) }
      ;
        { T = (T1 ++ A1)}
      )
    ;
      error_handler('Expecting term')
    ).
eqq_term_append(T1, T)  -->
    (
      eqq_list_term(A1)
    ->
      (
        [op('<>')], !,
        eqq_term_append(A1, B),
        { T = (T1 <> B) }
      ;
        { T = (T1 <> A1)}
      )
    ;
      error_handler('Expecting term')
    ).
*/

eqq_string_term((T?B)) --> var(T), [op('?')], !, eqq_goal(B).
eqq_string_term(T) --> basic_term(T), {not_a_concat(T)}.

eqq_list_term((T?B)) --> (var(T);list(T)), [op('?')],  !, eqq_goal(B).
eqq_list_term(T) --> basic_term(T), {not_a_concat(T)}.


eqq_goal(G) --> any_open_bracket, !, predication(G), matching_bracket(cb(')')).
eqq_goal(G) --> string_predication(G).

not_a_concat(T) :- compound(T), !, functor(T, F, _), F \== '<>', F \== '++'.
not_a_concat(_).

at_var_list(Vars) -->
    (
      any_open_bracket
    ->
      var_list(Vars),
      (
        [cb(')')]
      ->
        []
      ;
        error_handler('Expecting \')\' or \',\'')
      )      
    ;
      var_list(Vars)
    ).


%% --------- EBNF for terms ------------------------
%% term = append_term |
%%        concat_term |
%%        arith_term |
%%        basic_term
%% basic_term = ( "(" , term , ")" ) |
%%              list_constructor |
%%              ( "{" , term , "|" , predication_goals , "}" ) |
%%              tuple_constructor |
%%              apply_term |
%%              ( "{" , term , "}" ) |
%%              ( var , "?" , predication ) |
%%              ( "#" , basic_term ) |
%%              ( "-" , basic_term )
%%              list | atom | int | num | var | string
%% apply_term = basic_term , bracketed_tuple 
%% bracketed_tuple = "()" | ( "(" , term ,  { "," , term } , ")" )
%% list_constructor = "(" , term , "|" , predication_goals , ")"
%% tuple_constructor = "(" , term , "," term , [ ", " , term] , ")"
%% append_term = basic_term , "<>" , basic_term , [ "<>" , basic_term ] 
%% concat_term = basic_term , "++" , basic_term , [ "++" , basic_term ]
%% arith_term = basic_term

%% arith_term = ( neg_term , "**" , neg_term ) |  arith_mult_div_term
%% arith_mult_div_term = ( neg_term , "*" , neg_term ) |
%%                       ( neg_term , "-" , neg_term ) |
%%                       arith_plus_minus_term
%% arith_plus_minus_term = ( neg_term , "+" , arith_term ) |
%%                         ( neg_term , "-" , arith_term ) |
%%                         neg_term
%% 
%% -------------------------------------------------

/*
term(type(A,T)) -->
    atom(type),
    [ob('(')],
    term(A),    [punct(',')],
    moded_type(T1),
    (
      [op('<=')],
      { T = '<='(T1) }
    ;
      { T = T1 }
    ),
    [cb(')')].
term(type(A,T,V)) -->
    atom(type),
    [ob('(')],!,
    term(A),
    [punct(',')],
    moded_type(T1),
    (
      [op('<=')],
      { T = '<='(T1) }
    ;
      { T = T1 }
    ),
    [punct(',')],
    term(V),
    [cb(')')].
*/
%term(T) --> term1(T), !.
term(T) --> single_goal(T), !.
%term(T) --> predication(T).

term1(type(A, T)) --> atom(type), 
    [ob('(')], term(A), [punct(',')], '$type_rule'(T), [cb(')')], !.
term1(T) -->
    [ob('(')],
    term(T),
    [cb(')')], !.
term1('@..'(T1, T2)) -->
    basic_term(T1), [op('@..')], !, basic_term(T2).
term1(T) -->
    basic_term(T1),
     (
      [op('<>')], append_term(T1, T)
     ;
      [op('++')], {ip_lookup(parse_pp, yes)}, concat_term(T1, T)
     ;
      ( \+ \+ [op(union)] ; \+ \+ [op(diff)]; \+ \+ [op(inter)] ), !,
      set_expression(T1, T)
     ;
      arith_expression(T1, T), !
     ;
      {T = T1}
     ).


apply_term(T1, T) -->
    (
      [ob('(')], !,
      bracketed_tuple(T2),
      {'$qbuild_structure'(T1, T2, T3)},
      apply_term(T3, T)
    ;
      {T1 = T}
    ).

bracketed_tuple('$tuple'(Tuple)) -->
    (
      [cb(')')]
    ->
      {Tuple = []}
    ;
      term(H),
      bracketed_tuple_aux(T),
      {Tuple = [H|T]}
    ).

bracketed_tuple_aux([]) -->
    [cb(')')],!.
bracketed_tuple_aux([H|T]) -->
    (
      [punct(',')]
    ->
      (
        term(H)
      ->
        bracketed_tuple_aux(T)
      ;
        error_handler('Expecting a term')
      )
    ;
      error_handler('Expecting \')\' or \',\'')
    ).

append_term(A, App) -->
    (
      basic_term(A1)
    ->
      (
        [op('<>')], !,
        append_term(A1, B),
        { App = (A <> B) }
      ;
        {App = (A <> A1)}
      )
    ;
      error_handler('Expecting term')
    ).
concat_term(A, App) -->
    (
      basic_term(A1)
    ->
      (
        [op('++')], !,
        concat_term(A1, B),
        { App = (A ++ B) }
      ;
        {App = (A ++ A1)}
      )
    ;
      error_handler('Expecting term')
    ).
    
arith_expression(A, Result) -->
    power_term(A, Power),
    mult_div_term(Power, MultDiv),
    add_minus_term(MultDiv, Result).

set_expression(A, Result) -->
    inter_term(A, Inter),
    union_diff_term(Inter, Result).

power_term(A, Result) -->
    [op('**')], !,
    basic_term(A1),
    power_term(A1, R1),
    { '$process_arith_tuples'('**',  A, R1, Result) }.
power_term(A, A) --> [].

mult_div_term(A, Result) -->
    [op('*')], !,
    basic_term(A1),
    power_term(A1, A2),
    { '$process_arith_tuples'('*',  A, A2, A3) }, 
    mult_div_term(A3, Result).
mult_div_term(A, Result) -->
    [op('/')], !,
    basic_term(A1),
    power_term(A1, A2),
    { '$process_arith_tuples'('/',  A, A2, A3) }, 
    mult_div_term(A3, Result).
mult_div_term(A, Result) -->
    [op('//')], !,
    basic_term(A1),
    power_term(A1, A2),
    { '$process_arith_tuples'('//',  A, A2, A3) }, 
    mult_div_term(A3, Result).
mult_div_term(A, Result) -->
    [op('>>')], !,
    basic_term(A1),
    power_term(A1, A2),
    { '$process_arith_tuples'('>>',  A, A2, A3) }, 
    mult_div_term(A3, Result).
mult_div_term(A, Result) -->
    [op('<<')], !,
    basic_term(A1),
    power_term(A1, A2),
    { '$process_arith_tuples'('<<',  A, A2, A3) }, 
    mult_div_term(A3, Result).
mult_div_term(A, Result) -->
    [op('mod')], !,
    basic_term(A1),
    power_term(A1, A2),
    { '$process_arith_tuples'('mod',  A, A2, A3) }, 
    mult_div_term(A3, Result).
mult_div_term(A, Result) -->
    [op('rem')], !,
    basic_term(A1),
    power_term(A1, A2),
    { '$process_arith_tuples'('rem',  A, A2, A3) }, 
    mult_div_term(A3, Result).
mult_div_term(A, A) --> [].

add_minus_term(A, Result) -->
    [op('+')], !,
    basic_term(A1),
    power_term(A1, A2),
    mult_div_term(A2, A3),
    { '$process_arith_tuples'('+',  A, A3, A4) }, 
    add_minus_term(A4, Result).
add_minus_term(A, Result) -->
    [op('-')], !,
    basic_term(A1),
    power_term(A1, A2),
    mult_div_term(A2, A3),
    { '$process_arith_tuples'('-',  A, A3, A4) }, 
    add_minus_term(A4, Result).
add_minus_term(A, Result) -->
    [op('\\/')], !,
    basic_term(A1),
    power_term(A1, A2),
    mult_div_term(A2, A3),
    { '$process_arith_tuples'('\\/',  A, A3, A4) }, 
    add_minus_term(A4, Result).
add_minus_term(A, Result) -->
    [op('/\\')], !,
    basic_term(A1),
    power_term(A1, A2),
    mult_div_term(A2, A3),
    { '$process_arith_tuples'('/\\',  A, A3, A4) }, 
    add_minus_term(A4, Result).
add_minus_term(A,A) --> [].



apply_minus(A, B) :-
    integer(A), !, B is -A.
apply_minus(A, B) :-
    number(A), !, B is -A.
apply_minus(A, -A).


inter_term(A, Result) -->
    [op('inter')], !,
    basic_term(A1),
    inter_term(A1, A2),
    { '$process_set_tuples'('inter',  A, A2, A3) }, 
    union_diff_term(A3, Result).
inter_term(A, A) --> [].

union_diff_term(A, Result) -->
    [op('union')], !,
    basic_term(A1),
    inter_term(A1, A2),
    { '$process_set_tuples'('union',  A, A2, A3)}, 
    union_diff_term(A3, Result).
union_diff_term(A, Result) -->
    [op('diff')], !,
    basic_term(A1),
    inter_term(A1, A2),
    { '$process_set_tuples'('diff',  A, A2, A3) }, 
    union_diff_term(A3, Result).
union_diff_term(A, A) --> [].


basic_term('$set2list'(A)) --> atom([]), [ob('(')], !,
    term(A), matching_bracket(cb(')')).
basic_term('$'(A)) -->
    [op('$')],!, ( atom(A) -> [] ; error_handler('Expecting atom')).
basic_term(Result) -->
    any_open_bracket,
    [cb(')')], !,
    { Result = '$tuple'([]) }.
basic_term(Result) -->
    any_open_bracket,
    term(A),
    tuple_constructor(A, Tuple), [cb(')')], !,
    {  Tuple = [X] -> Result = X ; Result = '$tuple'(Tuple)}.
basic_term(List) --> set_constructor(List), !.
basic_term(List) --> list_constructor(List), !.

/*
  basic_term({Term}) -->
    [ob('{')],
    term(A),
    tuple_constructor(A, Tuple), [cb('}')], !, 
    {  Tuple = [X] -> Term = X ; Term = Tuple}.
basic_term({Term}) -->
    [ob('{')], !, 
    term(Term),
    matching_bracket(cb('}')).
basic_term((A ? B)) -->
    var(A),
    [op('?')], !,
    string_predication(B).
*/    

basic_term(A) --> [op('-')], basic_term(A1), !, {'$attach_minus'(A1, A)}.
basic_term('\\'(A)) --> [op('\\')], basic_term(A), !.
basic_term(#(A)) --> [op('#')], basic_term(A), !.
basic_term(A) --> handle_term(A).
basic_term('$list2set'(A)) --> [ob('{'), cb('}')], [ob('(')], !,
    term(A), matching_bracket(cb(')')).
basic_term(A) --> basic_term0(A1),apply_term(A1, A).

handle_term(T) -->
    ( atom(A) ; var(A) ),
    \+[ob('(')],
    handle_term_aux(A, T).

handle_term_aux(A, T) -->
    [op(':')], !,
    ( atom(A1) ; var(A1) ),
    handle_term_aux_colon(A, A1, T).
handle_term_aux(A, T) -->
    [op('@')], !,
    ( atom(A1) ; var(A1) ),
    {T = '@'(A, A1)}.
handle_term_aux(A, T) -->
    {T = A}.

handle_term_aux_colon(A, A1, T) -->
    [op('@')], !,
    ( atom(A2) ; var(A1) ),
    {T = '@'(':'(A, A1), A2)}.
handle_term_aux_colon(A, A1, T) -->
    {T = ':'(A, A1)}.

 start_agent_handle_term('@'(Process, Machine)) -->
     (
       atom(Process)
     ->
       []
     ;
       error_handler('Expecting process name (atom)')
     ),
     (
       [op('@')]
     ->
       (
         atom(Machine)
       ->
         []
       ;
         error_handler('Expecting machine name (atom)')
       )
     ;
       {Machine = localhost}
     ).
       

/*
basic_term0({Term}) -->
    [ob('{')], 
    term(Term),
    [cb('}')], !.
*/
basic_term0('$set_enum'([])) -->
    [ob('{'), cb('}')], !.
basic_term0('$set_enum'(List)) -->
    [ob('{')], !,
    term(Term),
    tuple_constructor(Term, List),
    matching_bracket(cb('}')).


basic_term0(A) --> list(A), !.
basic_term0(A) --> atom(A), !.
basic_term0(A) --> int(A), !.
basic_term0(A) --> num(A), !.
basic_term0(A) --> var(A), !.
basic_term0(A) --> string(A), !.

apply_basic_term(A) -->
    basic_term(A1), 
    apply_term(A1, A).

atom_or_structure(T) --> struct(T), !.
atom_or_structure(T) --> atom(T).

set_constructor('$set_constr'(A, B)) -->
    [ob('{')],
    term(A),
    [op('::')], !,
    goal(B),
    [cb('}')].
   
list_constructor(Result) -->
    [ob('[')],
    term(A), !,
    [op('::')],
    goal(B),
    {Result = '$list_constr'(A, B)},
    [cb(']')].


tuple_constructor(A1, Args) -->
    (
      [punct(',')]
    ->
     term(A2),
     tuple_constructor(A2, Args2),
     {Args = [A1|Args2]}
    ;
      {Args = [A1]}
    ).

dur_dis_type_list(D) -->
    atom_or_structure(D1),
    (
     [punct(',')]
    ->
     dur_dis_type_list(D2),
     {D = [D1|D2]}
    ;
     {D = [D1]}
    ).

belief_percept_decl(D) -->
    belief_percept_decl_one(D1),
    (
     [punct(',')]
    ->
     belief_percept_decl(D2),
     {D = [D1|D2]}
    ;
     {D = [D1]}
    ).

belief_percept_decl_one(B) -->
    atom(A), 
    (
      [op(':')]
    ->
      (
        any_open_bracket
      ->
        (
          [cb(')')]
        ->
          {B = A}
        ;
          declaration_type_list(Type),
          (
            [cb(')')]
          ->
            {B =.. [A|Type]}
          ;
            error_handler('Expecting )')
          )
        )
      ;
        error_handler('Expecting (')
      )
    ;
      error_handler('Expecting :')
    ).
    
task_atomic_list(D) -->
    atom(A),
    atom_or_plus(A, D1),
    (
     [punct(',')]
    ->
     task_atomic_list(D2),
     {D = [D1|D2]}
    ;
     {D = [D1]}
    ).

atom_or_plus(A, D1) -->
    [op('+')], !,
    atom(B),
    {D1 = (A+B)}.
atom_or_plus(A, A) -->
    [].


type_struct(T) -->
    atom(F),
    (
     any_open_bracket, [cb(')')], !,
     {T = F('$none_')}
    ;
      any_open_bracket, !,
      type_struct_apply(F, T)
    ).
type_struct_apply(F, Term) -->
    type_struct_args(Args),
    [cb(')')],
    {F1 =.. [F|Args]},
    type_struct_apply_aux(F1, Term).
type_struct_apply_aux(F, Term) -->
    [ob('(')],!,
    type_struct_args(Args),
    [cb(')')],
    {F1 =.. [F|Args]},
    type_struct_apply_aux(F1, Term).
type_struct_apply_aux(F, F) --> [].

type_struct_args(AllArgs) -->
    '$type_rule'(A), 
    (
      [punct(',')], !, type_struct_args(Args),
      {AllArgs = [A|Args]}
    ;
      {AllArgs = [A]}
    ).
    
struct(T) -->
    atom(F), !,
    (
     any_open_bracket, [cb(')')], !,
     {T = F('$none_')}
    ;
      any_open_bracket, !,
      struct_apply(F, T)
    ).
struct(T) -->
    var(F),
    (
      term(A), !,
      {'$flip_apply'(F, A, T)}
    ;
      [ob('(')],
      struct_apply(F, T)
    ).

struct_apply(F, Term) -->
    struct_args(Args),
    [cb(')')],
    {F1 =.. [F|Args]},
    struct_apply_aux(F1, Term).
struct_apply_aux(F, Term) -->
    [ob('(')],!,
    struct_args(Args),
    [cb(')')],
    {F1 =.. [F|Args]},
    struct_apply_aux(F1, Term).
struct_apply_aux(F, F) --> [].

struct_args(AllArgs) -->
    term(A),
    (
      [punct(',')], !, struct_args(Args),
      {AllArgs = [A|Args]}
    ;
      {AllArgs = [A]}
    ).

list([H|T]) --> [ob('[')], !, term(H), list_tail(T).

list_tail(T) --> [op('|')], !, term(T), matching_bracket(cb(']')).
list_tail(T) --> [punct(','), op('..')], term(T),  matching_bracket(cb(']')), !.
list_tail(_) --> [punct(','), op('..')], matching_bracket(cb(']')), !.
list_tail([H|T]) --> [punct(',')], !, term(H), list_tail(T).
list_tail([]) --> matching_bracket(cb(']')), !.

simple_struct_or_atom(T) -->
    atom(F),
    (
      [ob('(')], !, struct_args(Args), [cb(')')], {T =.. [F|Args]}
    ;
      {T = F}
    ).


%% --------- EBNF for types ------------------
%% type =  ( basic_type , "->" , type ) |
%%         ( basic_type , "<=" ) |
%%         ( basic_type , "~>" ) |
%%      xxx   ( basic_type , "->" ) |
%%         ( "(" , type_tuple , ")" ) |
%%         basic_type
%% basic_type = var | enum_type | range_type | atom | structure |
%%              disj_type | list_type | set_type
%%              ( "(" , type , ")" )
%% type_tuple = type , { "," , type }
%% enum_type = atom , "|" , atom , [ "|" , atom ]
%% range_type = int , ".." , int
%% disj_type =  basic_type , "||", basic_type , { "||" , basic_type }
%% list_type = "[" , basic_type , "]"
%% set_type = "{" , basic_type , "}"
%% -------------------------------------------


'$type_rule'(T) -->
    basic_type(T1),
    (
      [op('?')]
    ->
      {T11 = '+?'(T1)}
    ;
      {T11 = T1}
    ),
   (
      [op('->')], !, '$type_rule'(T2),
      {T = (T11 -> T2)}
    ;
      [op('<=')], !,
      { (
	 compound(T1), functor(T1, F, _), F ==  '$tuple_type'
	->
	 T = (T1 <=)
	 ;
	 T = ('$tuple_type'([T11]) <=)
	)
      }
    ;
      [op('~>>')], !,
      { (
	 compound(T1), functor(T1, F, _), F ==  '$tuple_type'
	->
	 T = (T1 ~>>)
	 ;
	 T = ('$tuple_type'([T11]) ~>>)
	)
      }
    ;
      [op('~>')], !,
      { (
	 compound(T1), functor(T1, F, _), F ==  '$tuple_type'
	->
	 T = ('~>'(T1))
	 ;
	 T = '~>'('$tuple_type'([T11]))
	)
      }
    ;
     [op('||')], !,
     union_type_tuple(T2),
     { T = '$union_type'([T11|T2])}
   ;
      {T = T11}
    ).
    
%%% KLC 
basic_type(T) --> var(T), !.
basic_type('?'(T)) --> [op('?')], basic_type(T), !,
    (
      next_token(op('??'))
    ->
      error_handler('Unexpected ??')
    ;
      next_token(op('?'))
    ->
      error_handler('Unexpected ?')
    ;
      next_token(op('!'))
    ->
      error_handler('Unexpected !')
    ;
      []
    ).
basic_type('??'(T)) --> [op('??')], basic_type(T), !,
    (
      next_token(op('??'))
    ->
      error_handler('Unexpected ??')
    ;
      next_token(op('?'))
    ->
      error_handler('Unexpected ?')
    ;
      next_token(op('!'))
    ->
      error_handler('Unexpected !')
    ;
      []
    ).
basic_type('!'(T)) --> [op('!')], basic_type(T), !,
        (
      next_token(op('??'))
    ->
      error_handler('Unexpected ??')
    ;
      next_token(op('?'))
    ->
      error_handler('Unexpected ?')
    ;
      next_token(op('!'))
    ->
      error_handler('Unexpected !')
    ;
      []
    ).
basic_type('!'(term)) --> [op('!')], !.
basic_type('?'(term)) --> [op('?')], !.
basic_type('??'(term)) --> [op('??')], !.
basic_type(MT) -->
    basic_type_unannotated(T),
    (
      next_token(op('??'))
    ->
      error_handler('Unexpected ??')
    ;
      [op('?')]
    ->
      {MT = '+?'(T)}
    ;
      next_token(op('!'))
    ->
      error_handler('Unexpected !')
    ;
      {MT = T}
    ).

basic_type_unannotated(T) -->
    any_open_bracket,
    '$type_rule'(T),
    [cb(')')], !.
/*
basic_type_unannotated('$enum_type'([T | Rest])) -->
    atom(T), [op('|')], !, rest_enum_atom_type(Rest).
basic_type_unannotated('$enum_type'([T | Rest])) -->
    string(T), [op('|')], !, rest_enum_string_type(Rest).
basic_type_unannotated('$enum_type'([T | Rest])) -->
    anynum(T), [op('|')], !, rest_enum_num_type(Rest).
basic_type_unannotated((T1 .. T2)) -->
    any_open_bracket, int(T1),
    [op('..')], !,
    int(T2),
    matching_bracket(cb(')')).
basic_type_unannotated((T1 .. T2)) -->
    int(T1),
    [op('..')], !,
    int(T2).
*/
basic_type_unannotated('$set_type'(T)) --> [ob('{')], !, '$type_rule'(T), matching_bracket(cb('}')).
basic_type_unannotated('$list_type'(T)) --> [ob('[')],!, '$type_rule'(T),matching_bracket(cb(']')).
basic_type_unannotated(TT) --> any_open_bracket, !, type_tuple(T),
    matching_bracket(cb(')')),
    {'$build_tuple_type'(T, TT)}.
basic_type_unannotated(A(T)) --> atom(A), [ob('(')], basic_type(T),
    [cb(')')], !.
basic_type_unannotated(T) --> type_struct(T), !.
basic_type_unannotated(T) --> term(T).


rest_enum_atom_type([T | Rest]) -->
    (
     atom(T)
    ->
     []
    ;
     error_handler('Expecting atom')
    ),
    [op('|')], !,
    rest_enum_atom_type(Rest).
rest_enum_atom_type([T]) -->
    (
     atom(T)
    ->
     []
    ;
     error_handler('Expecting atom')
    ), !.
rest_enum_atom_type([]) --> [].

rest_enum_string_type([T | Rest]) -->
    (
     string(T)
    ->
     []
    ;
     error_handler('Expecting string')
    ),
    [op('|')], !,
    rest_enum_string_type(Rest).
rest_enum_string_type([T]) -->
    (
     string(T)
    ->
     []
    ;
     error_handler('Expecting string')
    ),!.
rest_enum_string_type([]) --> []. 

rest_enum_num_type([T | Rest]) -->
    (
     anynum(T)
    ->
     []
    ;
     error_handler('Expecting number')
    ),
    [op('|')], !,
    rest_enum_num_type(Rest).
rest_enum_num_type([T]) -->
    (
     anynum(T)
    ->
     []
    ;
     error_handler('Expecting number')
    ), !.
rest_enum_num_type([]) --> []. 




type_tuple([T1|T2]) --> '$type_rule'(T1), [punct(',')], !, type_tuple(T2).
type_tuple([T]) --> '$type_rule'(T).

union_type_tuple(Result) -->
    basic_type(T1),
    (
     [op('||')]
    ->
     union_type_tuple(Rest),
     {Result = [T1|Rest]}
    ;
     {Result = [T1]}
     ).

atom_tuple(Tuple) -->
    atom(T1),
    (      
      [punct(',')]
    ->
      atom_tuple(T2),
      { Tuple = (T1, T2) }
    ;
      { Tuple = T1 }
    ).
    

atom(T) --> [atom(T)].
atom('`'(T)) --> [op('`'), atom(T)], !.
atom('`'(T)) --> [op('`'), op(T)].
atom(T) --> [op(T)].
int(T) --> [int(T)].
num(T) --> [num(T)].
anynum(N) --> int(N).
anynum(N) --> num(N).
string(T) --> [string(T)].
var(V) --> [var(V, _)].



any_open_bracket --> [ob('(')].
any_open_bracket --> [ob(' (')].

						  
