%% Copyright 2014 Keith Clark, Peter Robinson
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.


?- compile_time_only(consult(op_decls)).

'$check_tr_prog'(Clause, VarNames) :-
    Clause = '$tr_prog'(Head, Rules),
    Head =.. [_|Args],
    '$annotated_type_info'(ATypes, Head),
    manage_errors1,
    '$check_tr_head_args'(Args, ATypes, VMT, Head, VarNames),
    manage_errors2,
    freeze_term(Clause),
    forall(member('~>'(RuleLHS, RuleRHS), Rules),
	   once((
            '$check_tr_rule'(RuleLHS, RuleRHS,
                             VMT, Head, '~>'(RuleLHS, RuleRHS), VarNames)
           ;
             true
           ))
          ).


'$check_tr_head_args'([], [], [], _Head, _VarNames).
'$check_tr_head_args'([A|Args], [T|Types], [(A:T)|VMT], Head, VarNames) :-
    (
     ( nonvar(A) ; frozen_var(A) )
    ->
     thaw_term(A),
     '$display_error'(tr_head_nonvar(A, Head, VarNames)),fail
    ;
     freeze_term(A)
    ),
    %'$update_varmap'(A, T),
    '$check_tr_head_args'(Args, Types, VMT, Head, VarNames).

    
%%'$check_tr_rule'(LHS, RHS, VMT, Head, Rule, VarNames) :- errornl('$check_tr_rule'(LHS, RHS, VMT, Head, Rule, VarNames)), fail.

'$check_tr_rule'(LHS, RHS, VMT, Head, Rule, VarNames) :-
    Clause = '$tr'(Head, Rule),
    '$check_tr_rule_LHS'(LHS, Clause, VarNames),
    '$check_tr_rule_RHS'(RHS, Clause, VarNames),
    '$initialize_copy_and_save_context'(Clause, VarNames, (LHS, RHS),
                                        (LHSCopy, RHSCopy), tr_prog),
    '$check_tr_rule_LHS_types'(LHS, LHSCopy, VMT, OutVMT),
    '$check_tr_rule_RHS_types'(RHS, RHSCopy, OutVMT).
    

'$check_tr_rule_LHS'('$time'(_, G, T), Clause, VarNames) :-
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_expressions'(T, Clause, VarNames).
'$check_tr_rule_LHS'('$while_until'(_, G, '$time'(_, T)), Clause, VarNames) :-
      !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_expressions'(T, Clause, VarNames).
'$check_tr_rule_LHS'('$while_until'(_, G, '$time'(_, WU, T)),
                     Clause, VarNames) :-
      !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_goal_expressions'(WU, Clause, VarNames, rel),
    '$check_expressions'(T, Clause, VarNames).
'$check_tr_rule_LHS'('$while_until'(_, G, WU), Clause,  VarNames) :-
      !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_goal_expressions'(WU, Clause, VarNames, rel).
'$check_tr_rule_LHS'('$while'(_, G, '$time'(_, T)), Clause,  VarNames) :-
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_expressions'(T, Clause, VarNames).
'$check_tr_rule_LHS'('$while'(_, G, '$time'(_, W, T)), Clause, VarNames) :-
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_goal_expressions'(W, Clause, VarNames, rel),
    '$check_expressions'(T, Clause, VarNames).
'$check_tr_rule_LHS'('$while'(_, G, W), Clause, VarNames) :-
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_goal_expressions'(W, Clause, VarNames, rel).
'$check_tr_rule_LHS'('$until'(_, '$while'(_, G, '$time'(_, W, TW)),
                              '$time'(_, U, TU)), Clause, VarNames) :-
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_goal_expressions'(W, Clause, VarNames, rel),
    '$check_expressions'(TW, Clause, VarNames),
    '$check_goal_expressions'(U, Clause, VarNames, rel),
    '$check_expressions'(TU, Clause, VarNames).
'$check_tr_rule_LHS'('$until'(_, '$while'(_, G, '$time'(_, W, TW)), U),
                     Clause, VarNames) :-
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_goal_expressions'(W, Clause, VarNames, rel),
    '$check_expressions'(TW, Clause, VarNames),
    '$check_goal_expressions'(U, Clause, VarNames, rel).
'$check_tr_rule_LHS'('$until'(_, '$while'(_, G, W), '$time'(_, U, TU)),
                     Clause, VarNames) :-
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_goal_expressions'(W, Clause, VarNames, rel),
    '$check_goal_expressions'(U, Clause, VarNames, rel),
    '$check_expressions'(TU, Clause, VarNames).
'$check_tr_rule_LHS'('$until'(_, '$while'(_, G, W), U), Clause, VarNames) :-
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_goal_expressions'(W, Clause, VarNames, rel),
    '$check_goal_expressions'(U, Clause, VarNames, rel).
'$check_tr_rule_LHS'('$until'(_, G, '$time'(_, T)), Clause, VarNames) :-
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_expressions'(T, Clause, VarNames).
'$check_tr_rule_LHS'('$until'(_, G, '$time'(_, U, T)), Clause, VarNames) :-
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_goal_expressions'(U, Clause, VarNames, rel),
    '$check_expressions'(T, Clause, VarNames).
'$check_tr_rule_LHS'('$until'(_, G, U), Clause, VarNames) :-
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel),
    '$check_goal_expressions'(U, Clause, VarNames, rel).
'$check_tr_rule_LHS'(G, Clause, VarNames) :-
    '$check_goal_expressions'(G, Clause, VarNames, rel).


'$check_tr_rule_LHS_types'('$time'(_, G, T), '$time'(_, GC, TC),
                           InVMT, OutVMT) :-
    !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_time_TR_expression'(T, TC, OutVMT).
'$check_tr_rule_LHS_types'('$while_until'(_, G, '$time'(_, T)),
                           '$while_until'(_, GC, '$time'(_, TC)),
                           InVMT, OutVMT) :-
      !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_time_TR_expression'(T, TC, OutVMT).
'$check_tr_rule_LHS_types'('$while_until'(_, G, '$time'(_, WU, T)),
                           '$while_until'(_, GC, '$time'(_, WUC, TC)),
                           InVMT, OutVMT) :-
      !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_if_ground_TR_call'(WU, WUC, OutVMT),
    '$check_types_tr_rule'(WU, WUC, OutVMT, _),
    '$check_time_TR_expression'(T, TC, OutVMT).
'$check_tr_rule_LHS_types'('$while_until'(_, G, WU),
                           '$while_until'(_, GC, WUC),
                           InVMT, OutVMT) :-
      !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_if_ground_TR_call'(WU, WUC, OutVMT),
    '$check_types_tr_rule'(WU, WUC, OutVMT, _).
'$check_tr_rule_LHS_types'('$while'(_, G, '$time'(_, T)),
                           '$while'(_, GC, '$time'(_, TC)),
                           InVMT, OutVMT) :-
    !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_time_TR_expression'(T, TC, OutVMT).
'$check_tr_rule_LHS_types'('$while'(_, G, '$time'(_, W, T)),
                           '$while'(_, GC, '$time'(_, WC, TC)),
                           InVMT, OutVMT) :-
    !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_if_ground_TR_call'(W, WC, OutVMT),
    '$check_types_tr_rule'(W, WC, OutVMT, _),
    '$check_time_TR_expression'(T, TC, OutVMT).
'$check_tr_rule_LHS_types'('$while'(_, G, W),
                           '$while'(_, GC, WC),
                           InVMT, OutVMT) :-
    !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_if_ground_TR_call'(W, WC, OutVMT),
    '$check_types_tr_rule'(W, WC, OutVMT, _).
'$check_tr_rule_LHS_types'('$until'(_, '$while'(_, G, '$time'(_, W, TW)),
                                    '$time'(_, U, TU)),
                           '$until'(_, '$while'(_, GC, '$time'(_, WC, TWC)),
                                    '$time'(_, UC, TUC)),
                           InVMT, OutVMT) :-
    !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_if_ground_TR_call'(W, WC, OutVMT),
    '$check_types_tr_rule'(W, WC, OutVMT, _),
    '$check_time_TR_expression'(TW, TWC, OutVMT),
    '$check_if_ground_TR_call'(U, UC, OutVMT),
    '$check_types_tr_rule'(U, UC, OutVMT, _),
    '$check_time_TR_expression'(TU, TUC, OutVMT).
'$check_tr_rule_LHS_types'('$until'(_, '$while'(_, G, '$time'(_, W, TW)), U),
                           '$until'(_, '$while'(_, GC, '$time'(_, WC, TWC)), UC),
                           InVMT, OutVMT) :-
    !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_if_ground_TR_call'(W, WC, OutVMT),
    '$check_types_tr_rule'(W, WC, OutVMT, _),
    '$check_time_TR_expression'(TW, TWC, OutVMT),
    '$check_if_ground_TR_call'(U, UC, OutVMT),
    '$check_types_tr_rule'(U, UC, OutVMT, _).
'$check_tr_rule_LHS_types'('$until'(_, '$while'(_, G, W), '$time'(_, U, TU)),
                           '$until'(_, '$while'(_, GC, WC), '$time'(_, UC, TUC)),
                           InVMT, OutVMT) :-
    !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_if_ground_TR_call'(W, WC, OutVMT),
    '$check_types_tr_rule'(W, WC, OutVMT, _),
    '$check_if_ground_TR_call'(U, UC, OutVMT),
    '$check_types_tr_rule'(U, UC, OutVMT, _),
    '$check_time_TR_expression'(TU, TUC, OutVMT).
'$check_tr_rule_LHS_types'('$until'(_, '$while'(_, G, W), U),
                           '$until'(_, '$while'(_, GC, WC), UC),
                           InVMT, OutVMT) :-
    !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_if_ground_TR_call'(W, WC, OutVMT),
    '$check_types_tr_rule'(W, WC, OutVMT, _),
    '$check_if_ground_TR_call'(U, UC, OutVMT),
    '$check_types_tr_rule'(U, UC, OutVMT, _).
'$check_tr_rule_LHS_types'('$until'(_, G, '$time'(_, T)),
                           '$until'(_, GC, '$time'(_, TC)),
                           InVMT, OutVMT) :-
    !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_time_TR_expression'(T, TC, OutVMT).
'$check_tr_rule_LHS_types'('$until'(_, G, '$time'(_, U, T)),
                           '$until'(_, GC, '$time'(_, UC, TC)),
                           InVMT, OutVMT) :-
    !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_if_ground_TR_call'(U, UC, OutVMT),
    '$check_types_tr_rule'(U, UC, OutVMT, _),
    '$check_time_TR_expression'(T, TC, OutVMT).
'$check_tr_rule_LHS_types'('$until'(_, G, U),
                           '$until'(_, GC, UC),
                           InVMT, OutVMT) :-
    !,
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT),
    '$check_if_ground_TR_call'(U, UC, OutVMT),
    '$check_types_tr_rule'(U, UC, OutVMT, _).
'$check_tr_rule_LHS_types'(G, GC, InVMT, OutVMT) :-
    '$check_types_tr_rule'(G, GC, InVMT, OutVMT).




'$check_if_ground_TR_call'(G, GC, VMT) :-
    ip_set(current_predicate, G),
    ip_set(current_predicate_copy, GC),
    manage_errors1,
    check_if_ground_call(G, VMT),
    manage_errors2, !.

'$check_time_TR_expression'(T, TCopy, VMT) :-
     ip_set(current_predicate, '$time'(T)),
     ip_set(current_predicate_copy, '$time'(TCopy)),
    manage_errors1,
    get_mt_constraints(T, '!', num, VMT, VMT, _, [], _),
    manage_errors2, !.
'$check_repeat_TR_expression'(T, TCopy, VMT) :-
    ip_set(current_predicate, '$repeat'(T)),
    ip_set(current_predicate_copy, '$repeat'(TCopy)),
    manage_errors1,
    get_mt_constraints(T, '!', int, VMT, VMT, _, [], _),
    manage_errors2, !.
    
'$check_types_tr_rule'(G, GCopy, InVMT, OutVMT) :-
    ip_set(current_predicate, G),
    ip_set(current_predicate_copy, GCopy),
    '$goals2list'(G, GL),
    '$goals2list'(GCopy, GLCopy),
    manage_errors1,
    process_body_calls(GL, GLCopy, InVMT, OutVMT, [], _),
    manage_errors2, !.

'$check_types_tr_rule_TR_action'(TR, TRCopy, InVMT) :-
    ip_set(current_predicate, TR),
    ip_set(current_predicate_copy, TRCopy),
    manage_errors1,
    '$annotated_type_info'(Types, TR),
    ip_set(call_type, '~>'('$tuple'(Types))),
    phase_one(TR, Types, InVMT),
    manage_errors2, !.


'$check_tr_rule_RHS'('++'(A, G), Clause, VarNames) :-
    !,
    (
     '$type_info'(_, action, _, G, _)
    ->
      '$check_goal_expressions'(G, Clause, VarNames, any)
    ;
      '$display_error'(non_action_in_tr_goal(G, Clause, VarNames))
    ),
    '$check_tr_rule_RHS'(A, Clause, VarNames).
'$check_tr_rule_RHS'('$wait_repeat'(A, '$repeat'(_, W,R)), Clause, VarNames) :-
    !,
    '$check_expressions'(W, Clause, VarNames),
    '$check_expressions'(R, Clause, VarNames),
    '$check_tr_rule_RHS'(A, Clause, VarNames).
'$check_tr_rule_RHS'('$wait_repeat'(A, WR), Clause, VarNames) :-
    !,
    '$check_expressions'(WR, Clause, VarNames),
     '$check_tr_rule_RHS'(A, Clause, VarNames).
 '$check_tr_rule_RHS'(timed_seq(Seq), Clause, VarNames) :-
    !,
    '$check_tr_rule_RHS_ts'(Seq, Clause, VarNames).
'$check_tr_rule_RHS'([A], _Clause, _VarNames) :-
    '$type_info'(_, action, _, A, _), !.
'$check_tr_rule_RHS'(Actions, Clause, VarNames) :-
    '$check_tr_rule_RHS_tr_or_actions'(Actions, Clause, VarNames).

'$check_tr_rule_RHS_tr_or_actions'([TR], _Clause, _VarNames) :-
    '$type_info'(_, tr_prog, _, TR, _), !.
'$check_tr_rule_RHS_tr_or_actions'(Actions, Clause, VarNames) :-
    '$check_tr_rule_RHS_actions'(Actions, Clause, VarNames).

'$check_tr_rule_RHS_actions'([], _, _).
'$check_tr_rule_RHS_actions'(['$nil'], _, _) :- !.
'$check_tr_rule_RHS_actions'([A|Actions], Clause, VarNames) :-
    (
      '$type_info'(_, Kind, _, A, _), ( Kind = discrete ; Kind = durative )
    ->
      true
    ;
      '$display_error'(non_dis_dur_in_tr_actions(A, Clause,
                                                 VarNames))
    ),
    '$check_tr_rule_RHS_actions'(Actions, Clause, VarNames).
    

%%'$check_tr_rule_RHS_ts'(A, _, _) :- errornl('$check_tr_rule_RHS_ts'(A)),fail.
'$check_tr_rule_RHS_ts'(['$time'(_, A, T)], Clause, VarNames) :-
    !,
    '$check_expressions'(T, Clause, VarNames),
    '$check_tr_rule_RHS_tr_or_actions'(A, Clause, VarNames).
'$check_tr_rule_RHS_ts'([A], Clause, VarNames) :-
    !,
    '$check_tr_rule_RHS_tr_or_actions'(A, Clause, VarNames).
'$check_tr_rule_RHS_ts'(['$time'(_, A, T)|Seq], Clause, VarNames) :-
    !,
    '$check_expressions'(T, Clause, VarNames),
    '$check_tr_rule_RHS_tr_or_actions'(A, Clause, VarNames),
    '$check_tr_rule_RHS_ts'(Seq, Clause, VarNames).

%%'$check_tr_rule_RHS_types'(A, _, _) :- errornl('$check_tr_rule_RHS_types'(A)),fail.
'$check_tr_rule_RHS_types'('++'(A, G), '++'(AC, GC), VMT) :-
    !,
    '$check_if_ground_TR_call'(G, GC, VMT),
    '$check_types_tr_rule'(G, GC, VMT, _),
    '$check_tr_rule_RHS_types'(A, AC, VMT).
'$check_tr_rule_RHS_types'('$wait_repeat'(A, '$repeat'(_, W,R)),
                     '$wait_repeat'(AC, '$repeat'(_, WC,RC)), VMT) :-
    !,
    '$check_time_TR_expression'(W, WC, VMT),
    '$check_repeat_TR_expression'(R, RC, VMT),
    '$check_tr_rule_RHS_types'(A, AC, VMT).
'$check_tr_rule_RHS_types'('$wait_repeat'(A, WR),
                           '$wait_repeat'(AC, WRC), VMT) :-
    !,
    '$check_time_TR_expression'(WR, WRC, VMT),
    '$check_tr_rule_RHS_types'(A, AC, VMT).
'$check_tr_rule_RHS_types'(timed_seq(Seq), timed_seq(SeqC), VMT) :-
    !,
    '$check_tr_rule_RHS_ts_types'(Seq, SeqC, VMT).
'$check_tr_rule_RHS_types'([A], [AC], VMT) :-
    '$type_info'(_, action, _, A, _), !,
    '$check_if_ground_TR_call'(A, AC, VMT),
    '$check_types_tr_rule'(A, AC, VMT, _).
'$check_tr_rule_RHS_types'(Actions, ActionsC, VMT) :-
    '$check_tr_rule_RHS_tr_or_actions_types'(Actions, ActionsC, VMT).

'$check_tr_rule_RHS_tr_or_actions_types'([TR], [TRC], VMT) :-
    '$type_info'(_, tr_prog, _, TR, _), !,
    '$check_if_ground_TR_call'(TR, TRC, VMT),
    '$check_types_tr_rule_TR_action'(TR, TRC, VMT).
'$check_tr_rule_RHS_tr_or_actions_types'(Actions, ActionsC, VMT) :-
    '$check_tr_rule_RHS_actions_types'(Actions, ActionsC, VMT).

'$check_tr_rule_RHS_actions_types'([], _, _).
'$check_tr_rule_RHS_actions_types'(['$nil'], _, _) :- !.
'$check_tr_rule_RHS_actions_types'([A|Actions], [AC|ActionsC], VMT) :-
    '$check_if_ground_TR_call'(A, AC, VMT),
    '$check_types_tr_rule'(A, AC, VMT, _),
    '$check_tr_rule_RHS_actions_types'(Actions, ActionsC, VMT).
    
'$check_tr_rule_RHS_ts_types'(['$time'(_, A, T)], ['$time'(_, AC, TC)], VMT) :-
    !,
    '$check_time_TR_expression'(T, TC, VMT),
    '$check_tr_rule_RHS_tr_or_actions_types'(A, AC, VMT).
'$check_tr_rule_RHS_ts_types'([A], [AC], VMT) :-
    !,
    '$check_tr_rule_RHS_tr_or_actions_types'(A, AC, VMT).
'$check_tr_rule_RHS_ts_types'(['$time'(_, A, T)|Seq],
                               ['$time'(_, AC, TC)|SeqC], VMT) :-
    !,
    '$check_time_TR_expression'(T, TC, VMT),
    '$check_tr_rule_RHS_tr_or_actions_types'(A, AC, VMT),
    '$check_tr_rule_RHS_ts_types'(Seq, SeqC, VMT).


