%% Copyright 2014 Keith Clark, Peter Robinson
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.

?- compile_time_only(consult(op_decls)).

%% TypeMap is a list of V:T where V is a type variable and T is a type
%% V:T means that V is a supertype of T
'$make_union_types'([]) :- !.

'$make_union_types'(TypeMap) :-
    '$strip_out_nonvars_or_equals'(TypeMap, StrippedMap),
    sort(StrippedMap, SortedTypeMap),
    (
      SortedTypeMap = []
    ->
      true
    ;
      SortedTypeMap = [(V:T)|RestMap],
      '$group_union_types'(RestMap, V, T, [], Groups),
      '$bind_type_vars'(Groups)
    ).


'$bind_ground_types'([], []).
'$bind_ground_types'([(T1 < T2)|Rest], Out) :-
    '$bind_ground_types'(Rest, Out1),
    (
     includes(T2, T1)
    ->
     Out = Out1
    ;
     ground(T2)
    ->
      T1 = T2,
      Out = Out1
    ;
      Out = [(T1 < T2)|Out1]
    ).

%/*
'$strip_out_nonvars_or_equals'([], []).
'$strip_out_nonvars_or_equals'([(T:T1)|Tail], R) :-
    (
      T == T1
    ->
      R = R1
    ;
      nonvar(T), var(T1)
    ->
      T = T1,
      R = R1
    ;
      nonvar(T)
    ->
      includes(T, T1),
      R = R1
    ;
      R = [(T:T1)|R1]
    ),
    '$strip_out_nonvars_or_equals'(Tail, R1).
%*/

'$bind_type_vars'(Groups) :-
    '$bind_simple_type_vars'(Groups, Rest),
    '$bind_type_vars_aux'(Rest).

'$bind_simple_type_vars'([], []).
'$bind_simple_type_vars'([(V:T)|Groups], Rest) :-
    (
      compound(T), T = '$constr_union_type'(_)
    ->
      Rest = [(V:T)|Rest1]
    ;
      V = T,
      Rest = Rest1
    ),
    '$bind_simple_type_vars'(Groups, Rest1).

'$bind_type_vars_aux'([]).
'$bind_type_vars_aux'([(V:T)|Groups]) :-
    '$flatten_list'([T], FT),
    '$simplify_types'(FT, ST),
    V = ST,
    '$bind_type_vars_aux'(Groups).

'$group_union_types'([], V, T, Grp, Groups) :-
    '$simplify_types'([T|Grp], Simp),
    Groups = [(V:Simp)].
'$group_union_types'([(V1:T1)|RestMap], V, T, Grp, Groups) :-
    V1 == V,  !,
    (
      V == T
    ->
      '$group_union_types'(RestMap, V, T1, Grp, Groups)
    ;
    '$group_union_types'(RestMap, V, T1, [T|Grp], Groups)
    ).
'$group_union_types'([(V1:T1)|RestMap], V, T, Grp, OutGroups) :-
    (
      V == T
    ->
      OutGroups = [(V:Grp)|Groups]
    ;
      '$simplify_types'([T|Grp], Simp),
      OutGroups = [(V:Simp)|Groups]
    ),
    '$group_union_types'(RestMap, V1, T1, [], Groups).

simplify_unify_vars([], []).
simplify_unify_vars([X, Y|SDGroup], SimpGroup) :-
    var(X), var(Y), 
    !,
    X = Y,
    simplify_unify_vars([X|SDGroup], SimpGroup).
simplify_unify_vars([X|SDGroup], [X|SimpGroup]) :-
    !,
    simplify_unify_vars(SDGroup, SimpGroup).

'$simplify_types'(Group, T) :-
    freeze_term((Group, T), FVars),
    '$flatten_list'(Group, FGroup),
    sort(FGroup, SGroup),
    '$remove_includes'(SGroup, SGroup, IGroup), 
    '$lookForDefinedType'(IGroup, DGroup),
    thaw_term(FVars),
    sort(DGroup, SDGroup),
    %simplify_unify_vars(SDGroup, SimpGroup),
    SimpGroup = SDGroup,
    (
     SimpGroup = [T]
    ->
     true
    ;
     %'$list2tuple'(DGroup, Disj),
     T = '$union_type'(SimpGroup)
    ).

'$remove_includes'([], _SGroup, []).
'$remove_includes'([T|G], SGroup, IGroup) :-
    '$remove_includes'(G, SGroup, IGroup1),
    (
      member(T1, SGroup), T \== T1, T = T1
    ->
      IGroup = IGroup1
    ;
      member(T1, SGroup), T \== T1, includes(T1, T)
    ->
      IGroup = IGroup1
    ;
      IGroup = [T|IGroup1]
    ).

%%
%% Delays is a list whose elements are of the form T1 < T2
%%

'$bind_all_types'(Delays) :-
    thaw_term(Delays, FVars),
    freeze_vars(FVars),
    '$bind_ground_types'(Delays, Delays1),
    map(map_constraint, Delays1, TypeMap),
   (
      '$make_union_types'(TypeMap)
    ->
      true
    ;
      process_no_type_error
   ), !.
'$bind_all_types'(Delays) :- errornl(failed(Delays)), fail.

type(Term, Type, VarMapOut) :-
    nonvar(Type),
    moded(Type),
    Type = '!'(T), !,
    ground(Term),
    ip_set(check_index_, 1),
    global_state_set(gs_check_index_, 1),
    get_mt_constraints_top1(Term, unify('!'), T, [], [], OutVMT, [], OutLE),
    bind_type_constraints(OutLE),
    '$strip_annotations'(OutVMT, VarMapOut),
    ip_set('$interpreter_vmt', OutVMT).
type(Term, Type, VarMapOut) :-
    nonvar(Type),
    moded(Type), !,
    Type = M(T),
    ip_set(check_index_, 1),
    global_state_set(gs_check_index_, 1),
    get_mt_constraints_top1(Term, unify(M), T, [], [], OutVMT, [], OutLE),
    bind_type_constraints(OutLE),
    '$strip_annotations'(OutVMT, VarMapOut),
    ip_set('$interpreter_vmt', OutVMT).
type(Term, Type, VarMapOut) :-
    Query = type(Term, Type, VarMapOut),
    copy_term(Query, QueryCopy),
    ip_set(type_checker_context, relation_query(Query, [])),
    ip_set(type_checker_copy_context, relation_query(QueryCopy, [])),
    has_type_top(Term, Type, [], OutVMT),
    '$strip_annotations'(OutVMT, VarMapOut),
    ip_set('$interpreter_vmt', OutVMT).

/*
  type(Term, Type) :-
    fail,
    nonvar(Type),
    moded(Type), !,
    Type = M(T),
    (
      M = '!'
    ->
      ground(Term),
      type(Term, T)
    ;
      type(Term, T)
    ).
*/
type(Term, Type) :- 
    Query = type(Term, Type),
    copy_term(Query, QueryCopy),
    ip_set(type_checker_context, relation_query(Query, [])),
    ip_set(type_checker_copy_context, relation_query(QueryCopy, [])),
    has_type_top(Term, Type, [], _OutVMT).

%qtype(A,T,V) :- type(A,T,V).


/*
qtype(A, T, VarMapOut, VarNames) :-
    collect_vars(A, Vars),
    map('$add_type_vars', Vars, VarMap),
    ip_set(varmap, VarMap),
    %global_state_set('$error_level', 1),
    ip_set(type_order, type_le),
    ip_set(type_checker_context, relation(type(A, T), VarNames)),
    manage_errors1,
    (
      '$qtype_aux'(A, T, OutDelays)
    ;
      \+ '$qtype_aux'(A, T, _),
      ip_lookup(type_checker_context, Context),
      process_error_context(type_type_query_error(A, T, type(A, T), Context))
    ),
    '$bind_all_types'(OutDelays),
    manage_errors2,
    ip_lookup(varmap, VarMapOut).
*/

'$qtype_aux'(A, T, OutDelays) :-
    '$hasTypes_aux'([A], [T1], [], Delays),
    '$bind_all_types'(Delays),
    %transform_subterms('$remove_bottom', T1, T1B),
    '$check_lt'(T1, T, [], OutDelays).

get_type(A, T) :-
    '$get_constraints'(Constraints),
    '$has_type_interface'(A, T, Constraints, Delays),
    '$bind_all_types'(Delays).



wrap_var_functor(A, _(A)).

'$update_dynamic_annotations'([], [], _).
'$update_dynamic_annotations'([_(T)|Types], [Ann(AT)|ATypes], Ann) :-
    '$add_annotations_to_type'(T, AT),
    '$update_dynamic_annotations'(Types, ATypes, Ann).


%% '$add_annotations'(InTypes, OutTypes, Annotation)
%% adds Annotation to elements of InTypes to get OutTypes
'$add_annotations'([], [], _).
'$add_annotations'([T|Types], [AT|ATypes], Ann) :-
    (
     var(T)
    ->
     AT = Ann(T)
    ;
     '$has_annotation'(T)
    ->
      '$add_annotations_to_type'(T, AT)
    ;
      '$add_annotations_to_type'(T, AT1), AT = Ann(AT1)
    ),
    '$add_annotations'(Types, ATypes, Ann).

'$has_annotation'('??'(_)).
'$has_annotation'('?'(_)).
'$has_annotation'('+?'(_)).
'$has_annotation'('!'(_)).


'$strip_annotations'([], []).
'$strip_annotations'([V:_(SH)|T], [V:SH|ST]) :-
    !,
    '$strip_annotations'(T, ST).
'$strip_annotations'([V:SH|T], [V:SH|ST]) :-
    '$strip_annotations'(T, ST).



'$unify_all'([], _).
'$unify_all'([X|R], X) :-
    '$unify_all'(R, X).



/*
%% Precondition is that this is only called when A is not frozen - and therefore
%% contains an unbound var at call time.
%% It should fail if function evaluation happens - i.e. a variable functor
%% or an application of a defined function.
%% IgnoreJoin is true if top-level function is <> or ++
'$check_eval_modes'(A, _IgnoreJoin, _Kind, _Pred) :-
    '$get_ultimate_functor'(A, F),
    var(F), !,
    F == A.
'$check_eval_modes'(A, IgnoreJoin, Kind, Pred) :-
    IgnoreJoin = true,
    '$get_ultimate_functor'(A, F),
    ( F = '<>' ; F = '++' ), !,
    '$get_all_join_args'(A, F, Args),
    forall(member(Arg, Args), '$check_eval_modes'(Arg, fail, Kind, Pred)).
'$check_eval_modes'(A, _IgnoreJoin, Kind, Pred) :-
    '$get_ultimate_functor'(A, F),
    nonvar(F),
    '$type_info'(F, function, _, _, _), !,
    (
      '$frozen_term'(A)
    ->
      true
    ;
      freeze_term(A, FVars),
      process_mode_error(function_call(A), Kind, FVars, Pred)
    ).

'$check_eval_modes'([H|T], _IgnoreJoin, Kind, Pred) :-
    !,
    '$check_eval_modes'(H, fail, Kind, Pred),
    '$check_eval_modes'(T, fail, Kind, Pred).
'$check_eval_modes'(Constr, _, Kind, Pred) :-
    compound(Constr), !,
    Constr =.. [_|Args],
    forall(member(Arg, Args), '$check_eval_modes'(Arg, fail, Kind, Pred)).
'$check_eval_modes'(_, _, _, _).
*/

'$get_all_join_args'(A, _F, Args) :-
    \+compound(A), !, Args = [A].
'$get_all_join_args'(A, F, Args) :-
    A = F(X, Y), !,
    '$get_all_join_args'(X, F, A1),
    '$get_all_join_args'(Y, F, A2),
    append(A1, A2, Args).
'$get_all_join_args'(A, _F, Args) :-
    Args = [A].

/*
'$check_special_modes_types'(Term, _, _) :-
     \+ compound(Term), !.
'$check_special_modes_types'('$list_constr'(T, Pred), VMT, OutVMT) :-
    !,
    '$goals2list'(Pred, Goals),
    '$check_mode_type_list'(Goals, VMT, OutVMT),
    unbound_vars(T, OutVMT, UB),
    (
      UB = []
    ->
      true
    ;
      process_mode_error('!', T)
    ).
'$check_special_modes_types'('$set'(T, Pred), Kind) :-
    !,
    '$check_mode_type_goals'(Pred, Kind),
    (
     '$frozen_term'(T)
    ->
     true
    ;
      freeze_term(T, FVars),
      process_mode_error(output(T), Kind, FVars, Pred)
    ).
'$check_special_modes_types'((T ? Pred), Kind) :-
    !,
    '$check_mode_type_goals'(Pred, Kind),
    (
     '$frozen_term'(T)
    ->
     true
    ;
      freeze_term(T, FVars),
      process_mode_error(Kind, FVars, Pred)
    ).
'$check_special_modes_types'(Term, Kind) :-
    Term =.. [_|Args],
    '$check_special_modes_types_aux'(Args, Kind). 

'$check_special_modes_types_aux'([], _Kind). 
'$check_special_modes_types_aux'([A|Args], Kind) :-
    '$check_special_modes_types'(A, Kind),
    '$check_special_modes_types_aux'(Args, Kind). 

  */

/*
check_valid_query(Query, Vars) :-
    '$goals2list'(Query, QL),
    (
      QL = [G|_], '$is_goal_term'(G)
    ->
      forall(member(G, QL), check_valid_query_one(G, Vars))
    ;
      check_valid_expression(Query, Vars)
    ).

check_valid_query_one(G, _Vars) :-
    '$check_valid_terms_aux'(G), !.
check_valid_query_one(G, Vars) :-
    '$display_error'(valid_rel_query(G, Vars)).

check_valid_expression(Expr, _Vars) :-
    '$get_function_type_info'(Expr, Args, Types, _RT),
    '$check_valid_expression_aux'(Args, Types),
    !.
check_valid_expression(Expr, Vars) :-
    '$display_error'(valid_expression_query(Expr, Vars)).

'$check_valid_expression_aux'([], []).
'$check_valid_expression_aux'([A|Args], [T|Types]) :-
    '$tuple_len'(A, AL),
    '$tuple_len'(T, AL),
    '$check_valid_expression_aux'(Args, Types).
*/

check_query_mode_type(Query, Vars, Kind) :-
    ip_set(check_index_, 1),
    global_state_set(gs_check_index_, 1),
    %global_state_set('$error_level', 1),
    global_state_set(num_errors, 0),
    %freeze_term(Query),
    '$goals2list'(Query, QueryList),
    QueryList = [G|_],
    %thaw_term(Query),
    (
     '$has_action_query'(QueryList)
   ->
      \+ \+ check_query_action_mode_type(Query, Vars, OutVMT), 
     Kind = action(OutVMT)
    ;     
      ('$is_goal_term'(G) ; Query = (_ & _) )
    ->
      \+ \+  check_query_relation_mode_type(Query, Vars, OutVMT),
     Kind = relation(OutVMT)
    ;
      \+ \+ check_query_expression_mode_type(Query, Vars, Type),
     Kind = expression(Type)
    ),
    global_state_lookup(num_errors, 0).

check_query_builtin_mode_type(remember(Query), Vars) :-
    %collect_vars(Query, QVars),
    %map('$add_type_vars', QVars, VarMap),
    %ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'(Query, relation_query(remember(Query), Vars)),
    manage_errors2.
check_query_builtin_mode_type(remember_for(Query,T), Vars) :-
    %collect_vars((Query,T), QVars),
    %map('$add_type_vars', QVars, VarMap),
    %ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'(remember_for(Query,T),
			     relation_query(remember_for(Query,T), Vars)),
    manage_errors2.
check_query_builtin_mode_type(rememberA(Query), Vars) :-
    %collect_vars(Query, QVars),
    %map('$add_type_vars', QVars, VarMap),
    %ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'(Query, relation_query(rememberA(Query), Vars)),
    manage_errors2.
check_query_builtin_mode_type(rememberA_for(Query,T), Vars) :-
    %collect_vars((Query,T), QVars),
    %map('$add_type_vars', QVars, VarMap),
    %ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'(remember_for(Query,T),
			     relation_query(rememberA_for(Query,T), Vars)),
    manage_errors2.
check_query_builtin_mode_type(forget(Query), Vars) :-
    %collect_vars(Query, QVars),
    %map('$add_type_vars', QVars, VarMap),
    %ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'(Query, relation_query(Query, Vars)),
    manage_errors2.
check_query_builtin_mode_type(forget_after(Query, T), Vars) :-
    %collect_vars((Query,T), QVars),
    %map('$add_type_vars', QVars, VarMap),
    %ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'(forget_after(Query, T),
			     relation_query(forget_after(Query, T), Vars)),
    manage_errors2.
check_query_builtin_mode_type(replace_by(Query1, Query2), Vars) :-
    %collect_vars((Query1, Query2), QVars),
    %map('$add_type_vars', QVars, VarMap),
    %ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'((Query1 & Query2), relation_query((Query1 & Query2), Vars)),
    manage_errors2.
/*
check_query_builtin_mode_type('$at_forall'(V, Query1, Query2), Vars) :-
    collect_vars((Query1, Query2), QVars),
    map('$add_type_vars', QVars, VarMap),
    ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'((Query1 & Query2),
                             relation_query('$at_forall'(V, Query1, Query2),
                                            Vars)),
    manage_errors2.
check_query_builtin_mode_type('$at_forall_actions'(V, Query1, Query2), Vars) :-
    collect_vars((Query1, Query2), QVars),
    map('$add_type_vars', QVars, VarMap),
    ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'((Query1 & Query2),
                             relation_query('$at_forall_actions'(V, Query1, Query2),
                                            Vars)),
    manage_errors2.
  */
check_query_builtin_mode_type('$exists'(V, Query1), Vars) :-
    %collect_vars(Query1, QVars),
    %map('$add_type_vars', QVars, VarMap),
    %ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'(Query1,
                             relation_query('$exists'(V, Query1),
                                            Vars)),
    manage_errors2.
check_query_builtin_mode_type('$forall'(V, Query1, Query2), Vars) :-
    %collect_vars((Query1, Query2), QVars),
    %map('$add_type_vars', QVars, VarMap),
    %ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'((Query1 & Query2),
                             relation_query('$forall'(V, Query1, Query2),
                                            Vars)),
    manage_errors2.
check_query_builtin_mode_type('$forall_actions'(V, Query1, Query2), Vars) :-
    %collect_vars((Query1, Query2), QVars),
    %map('$add_type_vars', QVars, VarMap),
    %ip_set(varmap, VarMap),
    manage_errors1,
    '$check_mode_type_goals'((Query1 & Query2),
                             relation_query('$forall_actions'(V, Query1, Query2),
                                            Vars)),
    manage_errors2.
   
check_query_relation_mode_type(Query, _VarNames, _OutVMT) :-
    freeze_term(Query),
    '$goals2list'(Query, QL),
    length(QL, N), N > 1,
    ( member(type(_, _), QL) ; member(typeC(_, _, _), QL) ), !,
    writeL(stderr, ["A type query cannot be part of a conjunction.", nl_]),
    fail.
    
check_query_relation_mode_type(Query, VarNames, OutVMT) :-
    once((
           copy_term((Query, VarNames), (QueryCopy, VarNamesCopy)),
           ip_set(type_checker_context, relation_query(Query, VarNames)),
           ip_set(type_checker_copy_context, relation_query(QueryCopy, VarNamesCopy)),
           freeze_term(Query),
           '$goals2list'(Query, QueryL),
           %thaw_term(FVars),
           freeze_term(QueryCopy),
           '$goals2list'(QueryCopy, QueryLCopy)
           %thaw_term(FVarsCopy)
         )),
    manage_errors1,
    process_body_calls(QueryL, QueryLCopy, [], OutVMT, [], _),
     manage_errors2.

check_query_action_mode_type(Query, VarNames, OutVMT) :-
    once((
           copy_term((Query, VarNames), (QueryCopy, VarNamesCopy)),
           ip_set(type_checker_context, action_query(Query, VarNames)),
           ip_set(type_checker_copy_context, action_query(QueryCopy, VarNamesCopy)),
           freeze_term(Query),
           '$goals2list'(Query, QueryL),
           %thaw_term(FVars),
           freeze_term(QueryCopy),
           '$goals2list'(QueryCopy, QueryLCopy)
           %thaw_term(FVarsCopy)
         )),
    manage_errors1,
    process_body_calls(QueryL, QueryLCopy, [], OutVMT, [], _),
     manage_errors2.


check_query_expression_mode_type('::'(A, Query), _Vars, T) :-
    !,
    %collect_vars((A, Query), QVars),
    %map('$add_type_vars', QVars, VarMap),
    %ip_set(varmap, VarMap),
    %manage_errors1,
    %'$check_mode_type_goals'(Query, expression_query(Query, Vars)),
    %freeze_term(A),
    %manage_errors2.
    once((
           copy_term((Query, VarNames), (QueryCopy, VarNamesCopy)),
           ip_set(type_checker_context, relation_query(Query, VarNames)),
           ip_set(type_checker_copy_context, relation_query(QueryCopy, VarNamesCopy)),
           freeze_term(Query),
           '$goals2list'(Query, QueryL),
           %thaw_term(FVars),
           freeze_term(QueryCopy),
           '$goals2list'(QueryCopy, QueryLCopy)
           %thaw_term(FVarsCopy)
         )),
    manage_errors1,
    process_body_calls(QueryL, QueryLCopy, [], OutVMT, [], _),
     manage_errors2,
     has_type2(A, T, OutVMT, _).

check_query_expression_mode_type(Query, VarNames, T) :-
    (
     var(Query)
    ->
     writeL(stderr, ["Error: cannot use a variable as a query", nl_]),
     fail
    ;
     true
    ),
    (
     '$get_ultimate_functor'(Query, QF), var(QF)
    ->
     writeL(stderr, ["Error: cannot have a query with a variable functor", nl_]),
     fail
    ;
     true
    ),
    once((
           copy_term((Query, VarNames), (QueryCopy, VarNamesCopy)),
           ip_set(type_checker_context, expression_query(Query, VarNames)),
           ip_set(type_checker_copy_context, expression_query(QueryCopy, VarNamesCopy))
         )),
    manage_errors1,
    ip_set(current_predicate, Query),
    get_mt_constraints_top(Query, '!', T, [], [], _, [], OutLE),
    bind_type_constraints(OutLE),
    manage_errors2.


manage_errors1 :-
    retractall('$error'(_)),
    global_state_set('$error_level', 1).
manage_errors1 :-
    %global_state_lookup('$error_level',EL),
    %%errornl(bbbbb(EL)),
    %%forall('$error'(EL1, Error), errornl('$error'(EL1, Error))),
   % EL1 is EL-1,
    forall(retract('$error'(Error)), '$process_saved_error'(Error)),
    fail.

/*
manage_errors2 :-
    global_state_lookup('$error_level',EL),
    '$error'(EL, _), !, 
    global_state_increment('$error_level', _),
    fail.
*/
manage_errors2 :-
    retractall('$error'(_)).

pcovers(top, _) :- !.
pcovers(term, atomic) :- !.
pcovers(term, relcall) :- !.
pcovers(term, actcall) :- !.
pcovers(term, trcall) :- !.
pcovers(atomic,int) :- !.
pcovers(atomic,atom) :- !.
pcovers(atomic,num) :- !.
pcovers(atomic,string) :- !.
pcovers(num,int) :- !.
pcovers(relcall, belief) :- !.
pcovers(relcall, percept) :- !.
pcovers(_, bottom) : !.

%%covers(T1, T2) :- errornl(covers(T1, T2)),fail.
covers(T1, T2) :- T1 == T2, !.
covers(T1,_T2) :- var(T1), !, fail.
covers(_T1,T2) :- var(T2), !, fail.
covers(T1,T2) :- pcovers(T1,T2), !.

covers('$enum_type'(T1), '$enum_type'(T2)) :-
    !,
    forall(member(X, T2), member(X, T1)).
covers(atom, '$enum_type'(List)):-
    List = [A|_],
    atom(A),
    !,
    forall( member(A,List), atom(A)).
covers(string, '$enum_type'(List)):-
    List = [A|_],
    string(A),
    !,
    forall( member(A,List), string(A)).
covers(num, '$enum_type'(List)):-
    List = [A|_],
    number(A),
    !,
    forall( member(A,List), number(A)).
covers(int, '$enum_type'(List)):-
    List = [A|_],
    int(A),
    !,
    forall( member(A,List), int(A)).
covers(nat, '$enum_type'(List)):-
    List = [A|_],
    is_nat(A),
    !,
    forall( member(A,List), is_nat(A)).

covers(A, '$union_type'(List)):-
    is_primitive_or_defined_type(A), %% change to is_primitive, no ok as is   
    !,
    forall( member(E,List), includes(A,E) ).
covers('$union_type'(Types),atomic) :-
    !,
    forall( member(T,[atom,int,num,string]),
             (member(ST,Types),includes(ST,T))).
covers('$union_type'(List),A):-
    member(E,List), includes(E,A). 
covers('$union_type'(List1),'$union_type'(List2)):-
    forall( member(E2,List2), 
            (member(E1,List1), once(includes(E1,E2)) )
	  ).
covers('$tuple_type'(List1), '$tuple_type'(List2)) :-
    covers_list(List1, List2).
covers((T1A -> T1B), (T2A -> T2B)) :-
    !,
    covers(T1B, T2B),
    covers(T2A, T1A).
covers(num,(_ .. _)).
covers(nat,(A .. _)) :- A >= 0.
covers(int,(_ .. _)).
covers(atomic,(_ .. _)).
covers((N1 .. M1), (N2 .. M2)):-
	N1=<N2,
	M1>=M2.

%covers([T1], '$set_type'(T2)) :- covers(T1, T2).
covers('$list_type'(T1), '$list_type'(T2)) :- covers(T1, T2).
covers('$set_type'(T1), '$set_type'(T2)) :- covers(T1, T2).

/*
'$smallest_range_type_for'(X, Range) :-
    errornl(xxx),fail,
    '$defined_type'(Range, (N..M)),between(N, M,  X),
    \+ ('$defined_type'(Range1, (N1..M1)),between(N1, M1,  X),
          Range1 \= Range, covers((N..M), (N1..M1))),!.
*/
'$smallest_enum_type_for'(X, EnumType) :-
    findall(f(N, Type), '$is_in_enum_type'(X, Type, N), Pairs),
    sort(Pairs, Sorted),
    Sorted = [f(_, EnumType)|_].

'$is_in_enum_type'(X, Type, Size) :-
    integer(X), '$defined_type'(Type, (N..M)),
    between(N, M, X),
    Size is M - N + 1.
'$is_in_enum_type'(X, Type, Size) :-
    '$defined_type'(Type,'$enum_type'(E)),
    member(X, E),
    length(E, Size).
    

% not sure about next four clauses, maybe should not be there
covers_val((N .. M),K) :- integer(K), !, K>=N,K=<M.
covers_val(List,A):-
        closed_list(List),
        atom(A),
	member(A,list). 
covers_val(atom,A) :- atom(A). 
covers_val(int,N) :- integer(N).
covers_val(num,N) :- number(N).

covers_list([], []).
covers_list([T1|List1], [T2|List2]) :-
    includes(T1, T2),
    covers_list(List1, List2).

%includes(T,_) :- var(T), !. 
%%includes(T1,T2) :- errornl(includes(T1,T2)),fail.
includes(T1,T2) :-
    T1==T2, !.			% identical types. 

includes(T1,T2) :-  % when T1 and T2 are two instances of same poly. type
    nonvar(T1), nonvar(T2), ( compound(T1) ; compound(T2) ),
    '$defined_type'(T1, _),
    '$defined_type'(T2, _),
    !,
    T1=..[F | ArgTs1],
    T2=..[F | ArgTs2],
     map(includes, ArgTs1, ArgTs2).

includes(T1,T2) :-  % when T1 and T2 both primitive
    nonvar(T1), nonvar(T2),
    pcovers(T1,T2),
    !. 

includes(T1,T2) :-  
% not same types and not diff instances of same poly. type
% get their descriptions if defined types, else use type name
        get_def(T1,DefT1),
        get_def(T2,DefT2),
         covers(DefT1,DefT2).



get_def(T,DefT) :-
    var(T), !, DefT = T.
get_def(T,DefT) :-
    '$defined_type'(T, '$type_macro'(DefT)),
    !.
get_def(T,DefT) :-
    '$defined_type'(T, DefT),
    !.
get_def(T,DefT):-
    T=DefT.




%%%%%%%%%%%%%% Type Inference %%%%%%%%%%%%%%%%%%%%

'$strip_functor'(_(X), X) :- !.
'$strip_functor'(X, X).

'$simplify_union'(A, B, _) :- errornl(simp__(A,B)), fail.
'$simplify_union'([], Seen, Seen).
'$simplify_union'([A|R], Seen, Out) :-
    nonvar(A), A = '$union_type'(X),
    append(X, R, XR),
    '$simplify_union'(XR, Seen, Out).
'$simplify_union'([A,B|R], Seen, Out) :-
    A = B, !,
    '$simplify_union'([B|R], Seen, Out).
'$simplify_union'([A|R], Seen, Out) :-
    Seen = [B|_],
    A = B, !,
    '$simplify_union'(R, Seen, Out).
'$simplify_union'([A|R], Seen, Out) :-
    member(B, R),
    type_le(A, B), !,
    '$simplify_union'(R, Seen, Out).
'$simplify_union'([A|R], Seen, Out) :-
    member(B, Seen),
    type_le(A, B), !,
    '$simplify_union'(R, Seen, Out).
'$simplify_union'([A|R], Seen, Out) :-
    '$simplify_union'(R, [A|Seen], Out).

'$match_constructor_types'([], [], []).
    
'$match_constructor_types'([AT|ArgTypes], [CT|ConstrTypes], OutPolyTypes) :-
    copy_term(CT, CCT),
    collect_vars(CCT, Vars),
    type_le(AT, CCT),
    '$match_constructor_types'(ArgTypes, ConstrTypes, PolyTypes),
    (
      Vars = [VT]
    ->
      OutPolyTypes = [VT|PolyTypes]
    ;
      OutPolyTypes = PolyTypes
    ).


has_type(Term, Type, InVMT, OutVMT) :-
    ip_set(check_index_, 1),
    global_state_set(gs_check_index_, 1),
    manage_errors1,
    get_mt_constraints_top1(Term, '?', Type, InVMT, InVMT, OutVMT, [], OutLE),
    manage_errors2,
    bind_type_constraints(OutLE).

has_type_top(Term, Type, InVMT, OutVMT) :-
    ip_set(check_index_, 1),
    global_state_set(gs_check_index_, 1),
    (
      moded(Type)
    ->
      Type = M(T)
    ;
      M = '?',
      T = Type
    ),
    get_mt_constraints_top1(Term, M, T, InVMT, InVMT, OutVMT, [], OutLE),
    bind_type_constraints(OutLE).
has_type_top(_Term, _Type, _InVMT, _OutVMT) :-
    manage_errors2, fail.

has_type2(Term, SType, InVMT, OutVMT) :-
    check_assert(var(SType)),
    ip_set(check_index_, 1),
    global_state_set(gs_check_index_, 1),
    %manage_errors1,
    get_mt_constraints_top1(Term, '?', Type, InVMT, InVMT, OutVMT, [], OutLE),
    manage_errors2,
    bind_type_constraints(OutLE),
    '$remove_repeats_from_union'(Type, SType).
has_type2(_Term, _Type, _InVMT, _OutVMT) :-
    manage_errors2, fail.

'$remove_repeats_from_union'(T, T) :-
    var(T), !.
'$remove_repeats_from_union'(T, T) :-
    atomic(T), !.
'$remove_repeats_from_union'('$enum_type'(T), T) :-
    !.
'$remove_repeats_from_union'('$constr_enum_type'(T), T) :-
    !.
'$remove_repeats_from_union'('$union_type'(T), ST) :-
    !,
    map('$remove_repeats_from_union', T, RT),
    sort(RT, SRT),
    (
      SRT = [T1]
    ->
      ST = T1
    ;
      ST = '$union_type'(SRT)
    ).
'$remove_repeats_from_union'(C, T) :-
    C =.. [F|Args],
    map('$remove_repeats_from_union', Args, RArgs),
    T =.. [F|RArgs].

    
%%type_le_check(_Term, T2, Type, MidLE, OutLE) :- errornl(type_le_check(_Term, T2, Type, MidLE, OutLE)),fail.
type_le_check(_Term, T2, Type, MidLE, OutLE) :-
    type_le(T2, Type, MidLE, OutLE), !.
type_le_check(Term, T2, Type, MidLE, _OutLE) :-
    find_a_chain(Type, MidLE, _Chain, Tn), !,
    process_type_error(Term, Tn, T2).
type_le_check(Term, T2, Type, _MidLE, _OutLE) :-
    process_type_error(Term, Type, T2).

'$is_an_atom'('') :- !.
'$is_an_atom'(Term) :-
    \+atom(Term), !, fail.
'$is_an_atom'(Term) :-
    '$type_info'(Term, Kind, _, _, _),
    \+member(Kind, [relation, action, function]), !, fail.
'$is_an_atom'(Term) :-
    atom_concat('$', Term, DTerm),
    '$type_info'(DTerm, _, _, _, _), !, fail.
'$is_an_atom'(_).

%%DEBUG
%%get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE):- errornl(get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE)), fail.

%% get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE)
%% Check if Term has type Type and mode Mode
%% CurrVMT - VMT from head and previous calls
%% InVMT - VMT while processing argument
%% OutVMT - update of InVMT after processing this Term
%% InLE - type le constraints before processing this term
%% OutLE - update to InLE after processing this term

%% First check groundedness
get_mt_constraints(Term, '!', _Type, CurrVMT, _InVMT, _OutVMT, _InLE, _OutLE) :-
    var(Term),
    ip_lookup(check_phase, Phase),
    Phase \= in_head,
    \+type_member_of(Term:_, CurrVMT), !,
    update_check_global_vars(Term),
    process_mode_error('!', Term).
get_mt_constraints(Term, '!', Type, CurrVMT, _InVMT, _OutVMT, _InLE, _OutLE) :-
    var(Term), type_member_of(Term:M(T), CurrVMT),
    \+matched_groundedness('!'(Type), M(T)),
    !,
    process_mode_error('!', Term).
%% If the mode is ?? then nothing changes
get_mt_constraints(_Term, '??', _, _CurrVMT, OutVMT, OutVMT, OutLE, OutLE) :-
    !.
get_mt_constraints(_Term, _, Type, _CurrVMT, OutVMT, OutVMT, OutLE, OutLE) :-
    moded(Type), Type = '??'(_),
    !.

%% Second check matched constructor types
get_mt_constraints(Term, Mode, Type, CurrVMT, _InVMT, _OutVMT, _InLE, _OutLE) :-
    var(Term), type_member_of(Term:M(T), CurrVMT),
    \+check_matching_constructor_types(Type, T),
    !,
    get_inner_mode_and_type(Mode(Type), _, IType),
    get_inner_mode_and_type(M(T), _, IT),
    process_type_error(Term, IType, IT).
   
%% Term is a variable in ! mode so should already appear in CurrVMT in
%% ! mode with a more constrainted type
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    all_bang_mode(Mode(Type)),
    var(Term), type_member_of(Term:M(T), CurrVMT), !,
    update_check_global_vars(Term),
    get_inner_mode_and_type(Mode(Type), _IMode, IType),
    get_inner_mode_and_type(M(T), IM, IT),
    (
      IM \= '!'
    ->
      process_mode_error('!', Term)
    ;
      type_le_check(Term, IT, IType, InLE, OutLE)
    ),
    OutVMT = InVMT.
%% Special case for processing unification problems
get_mt_constraints(Term, unify(Mode), Type, CurrVMT, InVMT, OutVMT,
                   InLE, OutLE) :-
    var(Term), type_member_of(Term:_M(_T), CurrVMT), !,
    update_check_global_vars(Term),
    get_inner_mode_and_type(Mode(Type), IMode, IType),
    %get_inner_mode_and_type(M(T), IM, IT),
    InLE = OutLE,
    (
      type_remove_from(Term:M1(T1), InVMT, MidVMT)
    ->
     get_inner_mode_and_type(M1(T1), IM1, IT1),
     check_assert((IM1 = IMode)),
      (
        type_le(IType, IT1)
      ->
        OutVMT = [Term : Mode(Type)|MidVMT]
      ;
        type_le(IT1, IType)
      ->
        OutVMT = [Term : Mode(T1)|MidVMT]
      ;
        process_inconsistent_type_error(Term, IType, IT1)
      )
     ;
      OutVMT = [(Term:Mode(Type))|InVMT]
     ).
    
%% The case when the variable is already known to be in ! mode    
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    var(Term), type_member_of(Term:M(T), CurrVMT),
    all_bang_mode(M(T)), !,
    get_inner_mode_and_type(Mode(Type), _IMode, IType),
    get_inner_mode_and_type(M(T), _IM, IT),
    update_check_global_vars(Term),
    %% Mode is other than '!'
    %% If we know that Term is already ground by a previous call
    %% then we simply check if T =< Type
    OutVMT = InVMT,
    type_le_check(Term, IT, IType, InLE, OutLE).

%% The variable already appears before but neither the old or new modes is !
%% and we are in phase one - i.e. checking type consistency for input
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    ip_lookup(check_phase, phase_one),
    var(Term), type_member_of(Term:M(T), CurrVMT), !,
    get_inner_mode_and_type(Mode(Type), _IMode, IType),
    get_inner_mode_and_type(M(T), _IM, IT),
    update_check_global_vars(Term),
    %% IMode is not '!'
    %% If we are in phase one then we are checking 
    %% if IT =< IType in case Term is already given a value of type IT
    OutVMT = InVMT,
    type_le_check(Term, IT, IType, InLE, OutLE).
%% The variable already appears before but neither the old or new modes is !
%% and we are not in phase one 
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    var(Term), type_member_of(Term:_M(_T), CurrVMT), !,
    OutLE = InLE,
    update_check_global_vars(Term),
    get_inner_mode_and_type(Mode(Type), IMode, IType),
    %get_inner_mode_and_type(M(T), IM, IT),
    %% IMode and IM are not '!'
    (
     type_remove_from(Term:M1(T1), InVMT, MidVMT)
    ->
     %% The modes should be the same
     get_inner_mode_and_type(M1(T1), IM1, IT1),
     check_assert((IM1 = IMode)),
     (
        type_le(IType, IT1)
      ->
        OutVMT = [Term : Mode(Type)|MidVMT]
      ;
        type_le(IT1, IType)
      ->
        OutVMT = [Term : Mode(T1)|MidVMT]
      ;
        process_inconsistent_type_error(Term, Type, T1)
      )
     ;
      OutLE = InLE,
      OutVMT = [(Term:Mode(Type))|InVMT]
     ).
    
get_mt_constraints(Term, unify(Mode), Type, _CurrVMT, InVMT, OutVMT,
		   InLE, OutLE) :-
    var(Term), !,
    update_check_global_vars(Term),
    get_inner_mode_and_type(Mode(Type), IMode, IType),
    OutLE = InLE,
    (
      type_remove_from(Term:M(T), InVMT, MidVMT)
    ->
     get_inner_mode_and_type(M(T), IM, IT),
     check_assert((IM = IMode)),
      (
        type_le(IType, IT)
      ->
        OutVMT = [Term : Mode(Type)|MidVMT]
      ;
        type_le(IT, IType)
      ->
        OutVMT = [Term : Mode(T)|MidVMT]
      ;
        process_inconsistent_type_error(Term, Type, T)
      )
     ;
      OutVMT = [(Term:Mode(Type))|InVMT]
     ).
get_mt_constraints(Term, Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    var(Term), !,
    update_check_global_vars(Term),
    %OutLE = InLE,
    ip_lookup(check_phase, Phase),
    get_inner_mode_and_type(Mode(Type), IMode, IType),
    (
     Mode = '!', Phase \= in_head
    ->
      process_mode_error(Mode, Term)
    ;
      (
        type_remove_from(Term:M(T), InVMT, MidVMT)
      ->
        get_inner_mode_and_type(M(T), IM, IT),
        check_assert((IM = IMode)),
        (
	 var(IType), var(IT)
	->
	 type_le(IType, IT, InLE, OutLE),
	 OutVMT = [Term : Mode(Type)|MidVMT]
	;
          type_le(IType, IT, InLE, OutLE)
       % ->
          ,OutVMT = [Term : Mode(Type)|MidVMT]
        ;
          type_le(IT, IType, InLE, OutLE)
        %->
         , OutVMT = [Term : Mode(T)|MidVMT]
        ;
          process_inconsistent_type_error(Term, Type, T)
        )
      ;
       (
	var(Type)
       ->
        %OutLE = InLE,
        OutVMT = [(Term:Mode(Type1))|InVMT],
       type_le_check(Term, Type1, Type, InLE, OutLE)
       ;
	OutLE = InLE,
        OutVMT = [(Term:Mode(Type))|InVMT]
       )
      )
    ).
/* 
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    Term = '$none_', !,
    Type = '$none_', 
    InVMT = OutVMT,
    InLE = OutLE.        
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    '$is_a_typeXXX'(Term), !,
    Type = term,
    InVMT = OutVMT,
    InLE = OutLE.        
*/
get_mt_constraints(_Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    Type == top, !,
    InVMT = OutVMT,
    InLE = OutLE.        
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    Type == code, !, '$is_code'(Term),
    InVMT = OutVMT,
    InLE = OutLE.        
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    Type == term, !, \+'$is_code'(Term),
    InVMT = OutVMT,
    InLE = OutLE.        
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    Term = '`'(_), !,
    Type = atom,
    InVMT = OutVMT,
    InLE = OutLE.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    Term \= '$none_', '$is_a_type'(Term), !,
    (
     Type = typeE(Term)
    ->
     InVMT = OutVMT,
     InLE = OutLE
    ;
     process_type_error(Term, Type, typeE(Term))
    ).
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    nonvar(Type),
    Type = (N .. M), 
    integer(Term), 
    between(N, M, Term), !, 
    InVMT = OutVMT,
    InLE = OutLE.
get_mt_constraints(Term, _Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term), functor(Term, TermF, _), atom(TermF),
    Term = '$list_constr'(T, Pred), !,
    ip_lookup_init_empty('$global_vars', GVars),
    '$goals2list'(Pred, PredList),
    copy_term(Pred, PredCopy),
    freeze_term(PredCopy),
    '$goals2list'(PredCopy, PredListCopy),
    ip_lookup(current_predicate, OG),
    ip_lookup(current_predicate_copy, OGCopy),
    once(process_body_calls(PredList, PredListCopy, CurrVMT, LocalVMT, [], _)),
    ip_lookup_init_empty('$global_vars', GVarsPred),
    ip_set(current_predicate, OG),
    ip_set(current_predicate_copy, OGCopy),
    once(( Type = '$list_type'(TypeE); Type = '$set_type'(TypeE) )),
    get_mt_constraints_top1(T, '!', TypeE, LocalVMT, [], _, [], TLE),
    bind_type_constraints(TLE),
    %ip_lookup('$global_vars',GVarsT),
    ip_lookup_init_empty('$local_vars',Locals),
    diff_list(GVarsPred, GVars, PredLocalVars),
    collect_vars(T, TVars),
    diff_list(PredLocalVars, TVars, PossUnders),
    ip_lookup_init_empty('$underscore_vars', Unders),
    diff_list(PossUnders, Unders, NonUnders),
    (
      NonUnders = []
    ->
      true
    ;
      process_not_underscore_mode_error(NonUnders)
    ),
    union_list(Locals, PredLocalVars, NewLocals),
    ip_set('$local_vars',NewLocals),
    OutLE = InLE,
    OutVMT = InVMT.
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term), functor(Term, TermF, _), atom(TermF),
    Term = '$set_constr'(T, Pred), !,
    Type = '$set_type'(_),
    get_mt_constraints('$list_constr'(T, Pred), Mode, Type, CurrVMT, InVMT,
		       OutVMT, InLE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    nonvar(Type),
    '$defined_type'(Type, DefType),
    DefType='$union_type'(TypeList), !,
    member(T, TypeList),
    get_mt_constraints(Term, Mode, T, CurrVMT, InVMT, OutVMT, InLE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    nonvar(Type),
    Type='$union_type'(TypeList), !,
    member(T, TypeList),
    get_mt_constraints(Term, Mode, T, CurrVMT, InVMT, OutVMT, InLE, OutLE).
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    Term = '$set'([]),
    Type = '$set_type'(_),
    InVMT = OutVMT,
    OutLE = InLE.
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    Term = '$set'(List),
    List = [H|T],
    nonvar(Type),!,
    (
      Type = '$set_type'(_), ground(List), sorted(List)
    ->
      Type = '$set_type'(EType),
      %ground(List),
      %sorted(List), 
      get_mt_constraints(H, Mode, EType, CurrVMT, InVMT, MidVMT, InLE, MidLE),
      get_mt_constraints('$set'(T), Mode, Type, CurrVMT, MidVMT, OutVMT,
                         MidLE, OutLE)
    ;
      process_type_error(Term, Type,'$set_type'(_))
    ).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    Term = '$set'(List),
    List = [H|T], !,
    sorted(List),
    %Type = '$set_type'(EType),
    (
      ground(List)
    ->
      true
    ;
      process_mode_error('!', Term)
    ),
    get_mt_constraints(H, Mode, EType, CurrVMT, InVMT, MidVMT, InLE, MidLE),
    get_mt_constraints('$set'(T), Mode, '$set_type'(EType), CurrVMT,
                       MidVMT, OutVMT, MidLE, Mid1LE),
    type_le_check(Term, '$set_type'(EType), Type, Mid1LE, OutLE).
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    Term = [], !, Type = '$list_type'(_),
    InVMT = OutVMT,
    OutLE = InLE.
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    list(Term), nonvar(Type), 
    Term = [H|T], !,
    (
      Type = '$list_type'(_)
    ->
      get_inner_mode_and_type(Mode(Type), IMode, IType),
      IType = '$list_type'(EType),
      get_mt_constraints(H, IMode, EType, CurrVMT, InVMT, MidVMT, InLE, MidLE),
      get_mt_constraints(T, Mode, Type, CurrVMT, MidVMT, OutVMT,
                         MidLE, OutLE)
    ;
      process_type_error(Term, Type, '$list_type'(_))
    ).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    list(Term), (var(Type) ; Type \= '$set_type'(_) ),
    Term = [H|T], !,
    get_inner_mode_and_type(Mode(Type), IMode, IType),
    %IType = [EType],
    get_mt_constraints(H, IMode, EType, CurrVMT, InVMT, MidVMT, InLE, MidLE),
    get_mt_constraints(T, Mode, '$list_type'(EType), CurrVMT, MidVMT, OutVMT,
		       MidLE, Mid1LE),
    type_le_check(Term, '$list_type'(EType), IType, Mid1LE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term), functor(Term, F, _), atom(F),
    Term = '$tuple'(As), !, Type = '$tuple_type'(AsT),
    get_mt_constraints_iterate(As, Mode, AsT, CurrVMT, InVMT, OutVMT,
                               InLE, OutLE).
/*get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    %errornl(xxxxxxxxxxxxxxxxx),
    integer(Term), %%atom(Type),
    '$smallest_range_type_for'(Term, Range), !,
    InVMT = OutVMT,
    type_le_check(Term, Range, Type, InLE, OutLE).
*/
%% Type could be e.g. gender
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    ( atomic(Term) ; string(Term) ), %var(Type),
    '$smallest_enum_type_for'(Term, EnumType), !,
    InVMT = OutVMT,
    type_le_check(Term, EnumType, Type, InLE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    %compound(Term), functor(Term, TermF, _), atom(TermF),
    '$percept_type_info'(_, PType, Term), !,
    compound_arg_mode(Mode, AMode),
    '$update_dynamic_annotations'(PType, APType, AMode),
    Term =.. [_|Args],
    iterate_get_mt_constraints(Args, APType, CurrVMT, InVMT, OutVMT,
			       InLE, MidLE),
    type_le_check(Term, percept, Type, MidLE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term), functor(Term, TermF, _), atom(TermF),
    '$belief_type_info'(_, PType, Term), !,
    compound_arg_mode(Mode, AMode),
    '$update_dynamic_annotations'(PType, APType, AMode),
    Term =.. [_|Args],
    iterate_get_mt_constraints(Args, APType, CurrVMT, InVMT, OutVMT,
			       InLE, MidLE),
    type_le_check(Term, belief, Type, MidLE, OutLE).

get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    nonvar(Term),
    '$type_info'(_, durative, DType, Term, _), !,
    '$add_annotations'(DType, ADType, Mode),
    Term =.. [_|Args],
    iterate_get_mt_constraints(Args, ADType, CurrVMT, InVMT, OutVMT,
			       InLE, MidLE),
    type_le_check(Term, durative, Type, MidLE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    nonvar(Term),
    '$type_info'(_, discrete, DType, Term, _), !,
    '$add_annotations'(DType, ADType, Mode),
    Term =.. [_|Args],
    iterate_get_mt_constraints(Args, ADType, CurrVMT, InVMT, OutVMT,
			       InLE, MidLE),
    type_le_check(Term, discrete, Type, MidLE, OutLE).

get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term), functor(Term, TermF, _), atom(TermF),
    '$relation_type_info'(_, PType, Term), !,
    compound_arg_mode(Mode, AMode),
    '$update_dynamic_annotations'(PType, APType, AMode),
    Term =.. [_|Args],
    iterate_get_mt_constraints(Args, APType, CurrVMT, InVMT, OutVMT,
			       InLE, MidLE),
    type_le_check(Term, relcall, Type, MidLE, OutLE).
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term),
    '$relation_type_info'(_, _, Term), !,
    OutVMT = InVMT,
    type_le_check(Term, relcall, Type, InLE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term), functor(Term, TermF, _), atom(TermF),
    '$action_type_info'(_, PType, Term), !,
    compound_arg_mode(Mode, AMode),
    '$update_dynamic_annotations'(PType, APType, AMode),
    Term =.. [_|Args],
    iterate_get_mt_constraints(Args, APType, CurrVMT, InVMT, OutVMT,
			       InLE, MidLE),
    type_le_check(Term, actcall, Type, MidLE, OutLE).
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term),
    '$action_type_info'(_, _, Term), !,
    OutVMT = InVMT,
    type_le_check(Term, actcall, Type, InLE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term), functor(Term, TermF, _), atom(TermF),
    '$type_info'(TermF, tr_prog, PType, _, _), !,
    compound_arg_mode(Mode, AMode),
    '$add_annotations'(PType, APType, AMode),
    Term =.. [_|Args],
    iterate_get_mt_constraints(Args, APType, CurrVMT, InVMT, OutVMT,
			       InLE, MidLE),
    type_le_check(Term, trcall, Type, MidLE, OutLE).
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term),
    '$type_info'(_, tr_prog, _, Term, _), !,
    OutVMT = InVMT,
    type_le_check(Term, trcall, Type, InLE, OutLE).
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term), %nonvar(Type),
    '$type_info'(Term, function, _, _, _), !,
    '$type_info'(Term, function, Type1, _, _),
    once(type_le_check(Term, Type1, Type, InLE, OutLE)), 
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term), compound(Type), functor(Type, F, _), F == '<=', !,
    '$type_info'(Term, relation, Type1, _, _),
    type_le_check(Term, '<='('$tuple_type'(Type1)), Type, InLE, OutLE),
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term), %nonvar(Type),
    '$type_info'(Term, relation, _, _, _), !, 
    '$type_info'(Term, relation, Type1, _, _), 
    type_le_check(Term, '<='('$tuple_type'(Type1)), Type, InLE, OutLE),
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term), %nonvar(Type),
    atom_concat('$', Term, DTerm),
    '$type_info'(DTerm, relation, Type1, _, _), !,
    type_le_check(Term, '<='('$tuple_type'(Type1)), Type, InLE, OutLE),
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term), compound(Type), functor(Type, F, _), F == '~>>', !,
    '$type_info'(Term, action, Type1, _, _),
    type_le_check(Term, '~>>'('$tuple_type'(Type1)), Type, InLE, OutLE), 
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term), %nonvar(Type),
    '$type_info'(Term, action, Type1, _, _), !,
    type_le_check(Term, '~>>'('$tuple_type'(Type1)), Type, InLE, OutLE), 
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term), compound(Type), functor(Type, F, _), F == '~>', !,
    '$type_info'(Term, tr_prog, Type1, _, _),
    type_le_check(Term, '~>'('$tuple_type'(Type1)), Type, InLE, OutLE), 
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term), %nonvar(Type),
    '$type_info'(Term, tr_prog, Type1, _, _), !,
    type_le_check(Term, '~>'('$tuple_type'(Type1)), Type, InLE, OutLE), 
    InVMT = OutVMT.
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    var(Type),
    compound(Term),
    functor(Term, F, _N),
    atom(F),
    '$type_info'(F, function, (_ -> _), _, _), !,
    '$type_info'(F, function, (T1 -> T2), _, _),
    '$add_annotations_to_type'(T1, AT1),
    '$get_function_args'(Term, Args),
    '$tuple_type_to_list'(AT1, T1L),
    get_mt_constraints_iterate(Args, Mode, T1L, CurrVMT, InVMT, OutVMT,
                               InLE, MidLE),
    type_le_check(Term, T2, Type, MidLE, OutLE).
%% Type is e.g. tree(T)
%% So check Term against empty('$none') or tr(tree(T), T, tree(T))
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term),
    functor(Term, F, N),
    nonvar(F),
    nonvar(Type),
    '$defined_type'(Type, '$constr_enum_type'(Constructors)),    
    member(T1, Constructors),
    functor(T1, F, N), !,
    get_inner_mode_and_type(Mode(Type), IMode, _IType),
    Term =.. [_|Args],
    T1 =.. [_|ArgTypes],
    get_mt_constraints_iterate_constr(Args, IMode, ArgTypes, CurrVMT,
                                      InVMT, OutVMT, InLE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term),
    '$real_functor'(Term, F, N),
    nonvar(F),
    '$defined_type'(_, '$constr_enum_type'(Constructors0)),
    member(T0, Constructors0),
    '$real_functor'(T0, F, N), !,
    '$defined_type'(Type1, '$constr_enum_type'(Constructors)),
    member(T1, Constructors),
    '$real_functor'(T1, F, N),
    get_inner_mode_and_type(Mode(Type), IMode, _IType),
    Term =.. [_|Args],
    T1 =.. [_|ArgTypes],
    get_mt_constraints_iterate_constr(Args, IMode, ArgTypes, CurrVMT,
                                      InVMT, OutVMT, InLE, MidLE),
    type_le_check(Term, Type1, Type, MidLE, OutLE).
get_mt_constraints(Term, _Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term),
    functor(Term, F, _N),
    atom(F),
    '$type_info'(F, function, (_ -> _), _, _), !,
    '$type_info'(F, function, (T1 -> T2), _, _),
    '$add_annotations_to_type'(T1, AT1),
    '$get_function_args'(Term, Args),
    '$tuple_type_to_list'(AT1, T1L),
    get_mt_constraints_iterate(Args, '!', T1L, CurrVMT, InVMT, OutVMT,
                               InLE, MidLE),
    type_le_check(Term, T2, Type, MidLE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term),
    functor(Term, F, _N),
    var(F), 
    %FType = (T1 -> T2),
    FType = (T1 -> T2),
    get_mt_constraints_top1(F, Mode, FType, CurrVMT, InVMT, MidVMT, InLE, MidLE),
    '$add_annotations_to_type'(T1, AT1),
    '$get_function_args'(Term, Args),
    '$tuple_type_to_list'(AT1, T1L),
    get_mt_constraints_iterate(Args, Mode, T1L, CurrVMT, MidVMT, OutVMT,
                               MidLE, Mid1LE),
    type_le_check(Term, T2, Type, Mid1LE, OutLE).
get_mt_constraints(Term, _Mode, Type, _CurrVMT, _InVMT, _OutVMT, _InLE, _OutLE) :-
    nonvar(Type), Type = '$set_type'(_T),
    Term \= '$set'(_), !,
    process_type_error(Term, Type, '$unknown').
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term),
    functor(Term, F, _N),
    \+atom(F),
    FType = (T1 -> T2),
    get_mt_constraints_top1(F, Mode, (_ -> _), CurrVMT,
                            InVMT, _MidVMT, InLE, _MidLE),
    !,
    get_mt_constraints_top1(F, Mode, FType, CurrVMT,
                            InVMT, MidVMT, InLE, MidLE),
    %!,
    '$add_annotations_to_type'(T1, AT1),
    '$get_function_args'(Term, Args),
    '$tuple_type_to_list'(AT1, T1L),
    get_mt_constraints_iterate(Args, Mode, T1L, CurrVMT, MidVMT, OutVMT,
                               MidLE, Mid1LE),
    type_le_check(Term, T2, Type, Mid1LE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term),
    functor(Term, F, N),
    '$defined_type'(Type, '$enum_type'(Constructors)),
    member(T1, Constructors),
    functor(T1, F, N), !,
    %% F is a var
    (
     ( Mode = '!?' ; Mode = '!??' )
    ->
     process_mode_error(Mode, Term)
    ;
     true
    ),
    compound_arg_mode(Mode, AMode),
    Term =.. [_|Args],
    T1 =.. [_|ArgTypes],
    get_mt_constraints_iterate(Args, AMode, ArgTypes, CurrVMT, InVMT, OutVMT,
                               InLE, OutLE).
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term),
    functor(Term, F, N),
    nonvar(F),
    '$defined_type'(Type, '$type_macro'(Constructor)),
    functor(Constructor, F, N), %!,
    compound_arg_mode(Mode, AMode),
    Term =.. [_|Args],
    Constructor =.. [_|ArgTypes],
    get_mt_constraints_iterate(Args, AMode, ArgTypes, CurrVMT, InVMT, OutVMT,
			       InLE, OutLE).
%% Type could be e.g. church(T) - so replace by test against T -> T -> T
%% Type could be e.g. int_tree - so replace by test against tree(int)
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    ( compound(Term) ; atom(Term) ), nonvar(Type),
    functor(Term, F, _),
    nonvar(F),
    '$type_info'(_, defined_type, Type, '$type_macro'(Type1), _),
    get_mt_constraints(Term, Mode, Type1, CurrVMT, InVMT, OutVMT, InLE, OutLE).
%% special case - see above clause
get_mt_constraints(Term, _Mode, _Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    Term = '$none_', !, 
    InVMT = OutVMT,
    OutLE = InLE.
get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    compound(Term),
    !,
    (
      Type = term
    ->
      Term =.. [_|Args],
      length(Args, N),
      length(ArgTypes, N),
      get_mt_constraints_iterate(Args, Mode, ArgTypes, CurrVMT, InVMT, OutVMT,
			       InLE, OutLE)
    ;
      process_type_error(Term, atom, '$unknown')
    ).
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    integer(Term), Term >= 0, type_le_check(Term, nat, Type, InLE, OutLE), !, 
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    integer(Term), type_le_check(Term, int, Type, InLE, OutLE), !, 
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    string(Term), type_le_check(Term, string, Type, InLE, OutLE), !, 
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    number(Term), type_le_check(Term, num, Type, InLE, OutLE), !, 
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term), Term = bottom_, type_le_check(Term, bottom, Type, InLE, OutLE), !, 
    InVMT = OutVMT.
get_mt_constraints(Term, _Mode, Type, _CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    atom(Term), !,
    (
      '$is_an_atom'(Term)
    ->
      type_le_check(Term, atom, Type, InLE, OutLE), 
      InVMT = OutVMT
    ;
      process_type_error(Term, atom, '$unknown')
    ).
get_mt_constraints(Term, Mode, Type, CurrVMT, _InVMT, _OutVMT,
                       _InLE, _OutLE) :-
    get_inner_mode_and_type(Mode(Type), _, IType),
    (
     get_mt_constraints(Term, '!', T, CurrVMT, [], _, [], OutLE)
    ->
    get_inner_mode_and_type('!'(T), _, IT),
     bind_type_constraints(OutLE),
     process_type_error(Term, IType, IT)
    ;
     process_type_error(Term, IType, '$unknown')
    ).


%%get_mt_constraints_iterate_constr(Args, Mode, APType, CurrVMT, InVMT, OutVMT, InLE, MidLE) :- errornl(get_mt_constraints_iterate_constr(Args, Mode, APType, CurrVMT, InVMT, OutVMT, InLE, MidLE)),fail.
get_mt_constraints_iterate_constr([], _, [], _, InVMT, InVMT, InLE, InLE).
get_mt_constraints_iterate_constr([A|Args], Mode, [Type|ArgTypes], CurrVMT,
			   InVMT, OutVMT, InLE, OutLE) :-
    compound(A), A = _(X), X == '$none_', !,
    (
      '$defined_type'(Type, '$constr_enum_type'(Constructors)),
      member(A, Constructors)
    ->
      true
    ;
      process_type_error(A, Type, '$unknown')
    ),
    get_mt_constraints_iterate_constr(Args, Mode, ArgTypes, CurrVMT,
                                      InVMT, OutVMT, InLE, OutLE).
get_mt_constraints_iterate_constr([A|Args], Mode, [Type|ArgTypes], CurrVMT,
			   InVMT, OutVMT, InLE, OutLE) :-
    get_mt_constraints(A, Mode, Type, CurrVMT, InVMT, MidVMT, InLE, MidLE),
    get_mt_constraints_iterate_constr(Args, Mode, ArgTypes, CurrVMT,
                                      MidVMT, OutVMT, MidLE, OutLE).
%%get_mt_constraints_iterate(Args, Mode, APType, CurrVMT, InVMT, OutVMT, InLE, MidLE) :- errornl(get_mt_constraints_iterate(Args, Mode, APType, CurrVMT, InVMT, OutVMT, InLE, MidLE)),fail.
get_mt_constraints_iterate([], _, [], _, InVMT, InVMT, InLE, InLE).
get_mt_constraints_iterate([A|Args], Mode, [Type|ArgTypes], CurrVMT,
			   InVMT, OutVMT, InLE, OutLE) :-
    get_mt_constraints(A, Mode, Type, CurrVMT, InVMT, MidVMT, InLE, MidLE),
    get_mt_constraints_iterate(Args, Mode, ArgTypes, CurrVMT, MidVMT, OutVMT,
                               MidLE, OutLE).

iterate_get_mt_constraints([], [], _, InVMT, InVMT, InLE, InLE).
iterate_get_mt_constraints([A|Args], [Mode(Type)|ArgTypes], CurrVMT,
			   InVMT, OutVMT, InLE, OutLE) :-
    get_mt_constraints(A, Mode, Type, CurrVMT, InVMT, MidVMT, InLE, MidLE),
    iterate_get_mt_constraints(Args, ArgTypes, CurrVMT, MidVMT, OutVMT,
                               MidLE, OutLE).


get_mt_constraints_top1(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    ip_lookup(check_index_, N),
    N1 is N+1,
    ip_set(check_index_, N1),
    get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE).


get_mt_constraints_top(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE) :-
    ip_lookup(check_index_, N),
    N1 is N+1,
    ip_set(check_index_, N1),
    get_mt_constraints(Term, Mode, Type, CurrVMT, InVMT, OutVMT, InLE, OutLE),
    !.
/*
get_mt_constraints_top(Term, _Mode, Type, CurrVMT, _InVMT, _OutVMT,
                       _InLE, _OutLE) :-
    (
     get_mt_constraints(Term, '!', T, CurrVMT, [], _, [], OutLE)
    ->
        bind_type_constraints(OutLE),

        process_type_error(Term, Type, T)
     
    ;
     process_type_error(Term, Type, '$unknown')
    ).
  */
%% update_VMT(InHead, VMT, Var, Mode, Type, OutVMT)
%% InHead is true if we are processing the head of a rule
%% InHead is false if process body and
%% InHead is type if doing a type check
%% VMT is the current VMT to be updated with given var, mode type
%% to produce OutVMT
%%update_VMT(Kind, VMT, Var, M, Type, OutVMT) :- errornl(update_VMT(Kind, VMT, Var, M, Type, OutVMT) ),fail.

%% In the head and the mode is ! and we have Var before in ! mode
%% In that case update the type with the min of the 2 types
update_VMT(true, VMT, Var, '!', Type, OutVMT) :-
    type_remove_from(Var:'!'(T), VMT, MidVMT), !,
    get_inner_mode_and_type('!'(Type), _IMode, IType),
    get_inner_mode_and_type('!'(T), _IM, IT),
    (
     type_le(IT, IType)
    ->
     OutVMT = [Var:'!'(T)|MidVMT]
    ;
     type_le(IType, IT)
    ->
     OutVMT = [Var:'!'(Type)|MidVMT]
    ;
     process_inconsistent_type_error_head(Var, T, Type)
    ). 
%% In the head and the mode is ! and we have Var before not in ! mode
%% In that case the new type must be =< old type
%% e.g. p : ?int, !num
%% p(X, X) <= ...  would be a type error
update_VMT(true, VMT, Var, '!', Type, OutVMT) :-
    type_remove_from(Var:M(T), VMT, MidVMT), !,
    get_inner_mode_and_type('!'(Type), _IMode, IType),
    get_inner_mode_and_type(M(T), _IM, IT),
    (
     type_le(IType, IT)
    ->
     OutVMT = [Var:'!'(Type)|MidVMT] 
    ;
     type_le(IT, IType)
    ->
     process_head_output_type_error(Var, IType, IT, M)
    ;
     process_inconsistent_type_error_head(Var, IT, IType)
    ).
update_VMT(true, VMT, Var, Mode, Type, OutVMT) :-
    type_remove_from(Var:'!'(T), VMT, MidVMT), !,
    get_inner_mode_and_type(Mode(Type), _IMode, IType),
    get_inner_mode_and_type('!'(T), _IM, IT),
    (
     type_le(IT, IType)
    ->
     OutVMT = [Var:'!'(T)|MidVMT]
    ;
     type_le(IType, IT)
    ->
      ip_lookup('$body', B),
      (
        followed_by_type_constraint(B, Var, Type, '+?')
      ->
        OutVMT = [Var:'!'(T)|MidVMT]
      ;
        process_head_output_type_error(Var, IT, IType, Mode)
      )
    ;
     process_inconsistent_type_error_head(Var, IT, IType)
    ).
update_VMT(true, VMT, Var, Mode, Type, OutVMT) :-
    type_remove_from(Var:M(T), VMT, MidVMT), !,
    min_mode(M, Mode, MinM),
    min_mode_constructor_types(MinM, T, Type, MinT),
    get_inner_mode_and_type(Mode(Type), _, IType),
    get_inner_mode_and_type(M(T), _, IT),
    (
     IT == IType
    ->
     OutVMT = [Var:MinT|MidVMT]
    ;
     type_le(IT, IType)
    ->
     process_type_error(Var, IT, IType)
    ;
     process_type_error(Var, IType, IT)
    ).
%% not head processing
update_VMT(InHead, VMT, Var, Mode, Type, OutVMT) :-
    type_remove_from(Var:M(T), VMT, MidVMT), !,
    min_mode(M, Mode, MinM),
    get_inner_mode_and_type(Mode(Type), _, IType),
    get_inner_mode_and_type(M(T), _, IT),
    (
     type_le(IT, IType)
    ->
     min_mode_constructor_types(MinM, T, Type, MinT),
     OutVMT = [Var:MinT|MidVMT],
     (
       InHead = type
     ->
       ip_lookup(current_predicate_copy, OGCopy),
       (
         OGCopy = isa(_, _), M \= '!'
       ->
         true
       ;
	OGCopy = type(_, _), Mode \= M
       ->
	true
       ;
         process_type_check_succeeds_warning
       )
     ;
      true
     )
    ;
     type_le(IType, IT)
    ->
     min_mode_constructor_types(MinM, Type, T, MinT),
     OutVMT = [Var:MinT|MidVMT]
    ;
     InHead = true
    ->
     %% Can we get here?
     check_assert(fail),
     process_inconsistent_type_error_head(Var, T, Type)
    ;
     InHead = false
    ->
     process_inconsistent_type_error(Var, T, Type)
    ;
     %% InHead = type
     process_type_check_failure_warning,
     OutVMT = [Var:M(T)|MidVMT]
    ).
update_VMT(_, VMT, Var, Mode, Type, OutVMT) :-
    !,
    OutVMT = [Var:Mode(Type)|VMT].

type_le(A, B) :-
    type_le_aux(A, B, [], []).

compatible(T1, T2) :-
    type_le_aux(T1, T2, [], []), !.
compatible(T1, T2) :-
    type_le_aux(T2, T1, [], []), !.

check_clause(Head, B, BCopy) :-
    ip_set(current_predicate, '$head'(Head)),
    freeze_term(B, FVars),
    '$goals2list'(B, BL),
    thaw_term(FVars),    
    ip_set('$body', BL),
    process_head(Head, VMT, OVMT),
    collect_vars(Head, HeadVars),
    ip_set('$global_vars', HeadVars),
    check_head_output(OVMT, VMT, _, B),
    freeze_term(BCopy, FVarsCopy),
    '$goals2list'(BCopy, BLCopy),
    thaw_term(FVarsCopy),
    ip_set(check_phase, phase_one),
    process_body_calls(BL, BLCopy, VMT, OutVMT, OVMT, OutOVMT),
    check_head_output_end(OutOVMT, OutVMT, Unbound, B),
    (
      Unbound = []
    ->
      true
    ;
      process_mode_error(unbound_by_body, Unbound)
    ),
    ip_lookup('$delayed_unify', Unifys),
    '$end_unifys_check'(Unifys, OutVMT, OutOVMT).

check_function_clause(Head, function(B, Result), function(BCopy, ResultCopy)) :-
    ip_set(current_predicate, '$head'(Head)),
    collect_vars(Head, HeadVars),
    ip_set('$global_vars', HeadVars),
    process_function_head(Head, VMT, OVMT, ResultType),
    freeze_term(B, FVars),
    '$goals2list'(B, BL),
    thaw_term(FVars),
    freeze_term(BCopy, FVarsCopy),
    '$goals2list'(BCopy, BLCopy),
    thaw_term(FVarsCopy),
    process_body_calls(BL, BLCopy, VMT, MidVMT, OVMT, OutOVMT),
    ip_set(current_predicate, Result),
    ip_set(current_predicate_copy, ResultCopy),
    phase_one_check([Result], ['!'(ResultType)], MidVMT, [], _AConstr),
    ip_lookup('$delayed_unify', Unifys),
    '$end_unifys_check'(Unifys, MidVMT, OutOVMT).
  
process_head(Head, VMT, OVMT) :-
    '$annotated_type_info'(ATypes, Head),
    !,
    Head =.. [_|Args],
    freeze_term(ATypes),
    ip_set(check_phase, in_head),
    process_head_args(Args, ATypes, [], VMT, [], OVMT),
    ip_set(check_phase, phase_one).
%% Head is a higher_order function
process_head(Head, VMT, OVMT) :-
    '$get_ultimate_functor'(Head, R),
    '$type_info'(R, Kind, TType, G, _), !,
    (
     G \= '$function', \+'$belief'(G), \+'$percept'(G),
     '$get_function_type_info'(Head,  Args, Types, _ResultType)
    ->
     Args = [A1|RArgs],
     Types = [T1|RTypes],
     map('$add_input', RTypes, IRTypes),      
     '$tuple_to_list'(A1, A1L),
     '$tuple_type_to_list'(T1, T1L),
     append(RArgs, A1L, AllArgs),
     append(IRTypes, T1L, AllTypes),
     freeze_term(AllTypes),
     ip_set(check_phase, in_head),
     process_head_args(AllArgs, AllTypes, [], VMT, [], OVMT),
     ip_set(check_phase, phase_one)
    ;
     '$belief'(G),
     Head =.. [_|Args]
    ->
     freeze_term(TType),
     TType = '$tuple_type'(TTypes),
     ip_set(check_phase, in_head),
     process_head_args(Args, TTypes, [], VMT, [], OVMT),
     ip_set(check_phase, phase_one)
     ;
     '$percept'(G),
     Head =.. [_|Args]
    ->
     freeze_term(TType),
     TType = '$tuple_type'(TTypes),
     ip_set(check_phase, in_head),
     process_head_args(Args, TTypes, [], VMT, [], OVMT),
     ip_set(check_phase, phase_one)
    ;
     functor(Head, R, N),
     functor(G, R, NG),
     writeL(stderr, [nl_, "Error: the definition of ", R/N, 
		     " does not match the declaration of the ",
		     Kind, " ", R/NG, nl_, nl_]),
     fail
    ).

process_function_head(Head, VMT, OVMT, ResultType) :-
    '$get_function_type_info'(Head,  Args, Types, ResultType),
    '$add_annotations'(Types, ATypes, '!'),
    freeze_term(ATypes),
    ip_set(check_phase, in_head),
    process_head_args(Args, ATypes, [], VMT, [], OVMT),
    ip_set(check_phase, phase_one).

process_head_args([], [], VMT, VMT, OVMT, OVMT).
process_head_args([A|Args], [M(T)|Types], InVMT, OutVMT, InOVMT, OutOVMT) :-
    check_head_mode_type(A, M, T, InVMT, MidVMT, InOVMT, MidOVMT),
    process_head_args(Args, Types, MidVMT, OutVMT, MidOVMT, OutOVMT).

check_head_mode_type(Term, Mode, Type, InVMT, OutVMT, InOVMT, OutOVMT) :-
    get_mt_constraints_top(Term, Mode, Type, [], [], VMT, [], OutLE),
    bind_type_constraints(OutLE),
    %%errornl(check_head_mode_type_aux___XXX___(VMT, InVMT, OutVMT, InOVMT, OutOVMT)),
    check_head_mode_type_aux(VMT, InVMT, OutVMT, InOVMT, OutOVMT).

check_head_mode_type_aux([], InVMT, InVMT, InOVMT, InOVMT).
check_head_mode_type_aux([Term : Mode(Type)|VMT], InVMT, OutVMT,
                         InOVMT, OutOVMT) :-
    (
      all_bang_mode(Mode(Type))
    ->
      update_VMT(true, InVMT, Term, Mode, Type, MidVMT),
      MidOVMT = InOVMT
    ;
      update_VMT(true, InVMT, Term, Mode, Type, MidVMT),
      MidOVMT = [(Term:Mode(Type))|InOVMT]
    ),
    check_head_mode_type_aux(VMT, MidVMT, OutVMT, MidOVMT, OutOVMT).

%%followed_by_type_constraint(A,B,C,D) :- errornl(followed_by_type_constraint(A,B,C,D)),fail.
followed_by_type_constraint((A&B), V, T, Mode) :-
    !,
    '$goals2list'((A&B), BL),
    followed_by_type_constraint(BL, V, T, Mode). 
followed_by_type_constraint([type(V1, T1)|_Body], V, T, _Mode) :-
    V == V1, !,
    type_le(T1, T).
followed_by_type_constraint([typeG(V1, T1)|_Body], V, T, _Mode) :-
    V == V1, !,
    type_le(T1, T).
followed_by_type_constraint([isa(V1, T1)|_Body], V, T, _Mode) :-
    V == V1, !,
    type_le(T1, T).
followed_by_type_constraint([TypeCheck|_Body], V, T, _Mode) :-
    compound(TypeCheck), functor(TypeCheck, F, _), nonvar(F),
    '$is_a_type_test'(TypeCheck, V1, T1),
    V == V1, !,
    type_le(T1, T).
followed_by_type_constraint([type(_V1, _T1)|Body], V, T, Mode) :-
    !,
    followed_by_type_constraint(Body, V, T, Mode).
followed_by_type_constraint([typeG(_V1, _T1)|Body], V, T, Mode) :-
    !,
    followed_by_type_constraint(Body, V, T, Mode).
followed_by_type_constraint([TypeCheck|Body], V, T, Mode) :-
    compound(TypeCheck), functor(TypeCheck, F, _), nonvar(F),
    '$is_a_type_test'(TypeCheck, _V1, _T1), !,
    followed_by_type_constraint(Body, V, T, Mode).
followed_by_type_constraint([_|Body], V, T, Mode) :-
    Mode = '+?', !,
    followed_by_type_constraint(Body, V, T, Mode).

%check_head_output(A,B,_, _) :- errornl(check_head_output(A, B)), fail.
check_head_output(_, _, _, _).
%check_head_output(A,B,C,D) :- check_head_output_end(A,B,C,D).

check_head_output_end([], _VMT, [], _).
check_head_output_end([V:Mode(Type)|OVMT], VMT, Unbound, Body) :-    
    type_member_of(V : M(T), VMT), !,
    get_inner_mode_and_type(Mode(Type), IMode, IType),
    get_inner_mode_and_type(M(T), IM, IT),
    (
      (type_le(IT, IType) ; followed_by_type_constraint(Body, V, IType, IMode))
    ->
      true
    ;
     type_le(IType, IT)
    ->
      process_head_output_type_error(V, IT, IType, IMode)
    ;
      process_inconsistent_type_error_head(V, IT, IType)
    ),
    (
       IM \= '!', IMode \= '+?'
    ->
      Unbound = [V|Unbound1]
    ;
      Unbound = Unbound1
    ),
    check_head_output_end(OVMT, VMT, Unbound1, Body).
check_head_output_end([V:Mode(_Type)|OVMT], VMT, Unbound, Body) :-
    (
     Mode \= '!'
    ->
      Unbound = [V|Unbound1]
    ;
      Unbound = Unbound1
    ),
    check_head_output_end(OVMT, VMT, Unbound1, Body).

check_head_output_unify(A, B, OVMT, VMT, Body) :-
    type_member_of(A : MA(TA), OVMT),
    type_member_of(B : MB(TB), VMT), !,
    get_inner_mode_and_type(MA(TA), IMA, ITA),
    get_inner_mode_and_type( MB(TB), IMB, ITB),
    (
      ( type_le(ITB, ITA) ;
          followed_by_type_constraint(Body, A, ITA, IMA) ;
          followed_by_type_constraint(Body, B, ITA, IMB))
    ->
      true
    ;
      type_le(ITA, ITB)
    ->
      process_head_output_type_error(A, ITB, ITA, MA)
    ;
      process_inconsistent_type_error(A, ITA, ITB)
    ).
check_head_output_unify(B, A, OVMT, VMT, Body) :-
    type_member_of(A : MA(TA), VMT),
    type_member_of(B : MB(TB), OVMT), !,
    get_inner_mode_and_type(MA(TA), IMA, ITA),
    get_inner_mode_and_type( MB(TB), IMB, ITB),
    (
      ( type_le(ITB, ITA) ;
          followed_by_type_constraint(Body, A, ITA, IMA) ;
          followed_by_type_constraint(Body, B, ITA, IMB))
    ->
      true
    ;
      type_le(ITA, ITB)
    ->
      process_head_output_type_error(A, ITB, ITA, MA)
    ;
      process_inconsistent_type_error(A, ITA, ITB)
    ).
check_head_output_unify(_B, _A, _OVMT, _VMT, _Body).

merge_VMT([], LocalVMT, LocalVMT).
merge_VMT([V : Mode(Type)|VMT], LocalVMT, [V:MT|OutVMT]) :-
    type_remove_from(V : M(T), LocalVMT, OutLocalVMT),
    get_inner_mode_and_type(Mode(Type), _IMode, IType),
    get_inner_mode_and_type(M(T), _IM, IT),
    !,
    (
      compatible(IT, IType) %%XXX can this ever fail?
    ->
      true
    ;
      process_inconsistent_type_error(V, IT, IType)
    ),
    update_mode_on_call_exit(M(T), Mode(Type), MT),
    merge_VMT(VMT, OutLocalVMT, OutVMT).
merge_VMT([V : Mode(Type)|VMT], LocalVMT, [V:Mode(Type)|OutVMT]) :-
    merge_VMT(VMT, LocalVMT, OutVMT).

merge_unify_VMT([], LocalVMT, LocalVMT).
merge_unify_VMT([V : Mode(Type)|VMT], LocalVMT, [V:UM(T)|OutVMT]) :-
    type_remove_from(V : M(T), LocalVMT, OutLocalVMT),
    !,   
    get_inner_mode_and_type(Mode(Type), _IMode, IType),
    get_inner_mode_and_type(M(T), _IM, IT),
    (
      compatible(IT, IType) %%XXX can this ever fail?
    ->
      true
    ;
      process_inconsistent_type_error(V, IT, IType)
    ),
    min_mode(Mode, M, UM),
    merge_unify_VMT(VMT, OutLocalVMT, OutVMT).
merge_unify_VMT([V : Mode(Type)|VMT], LocalVMT, [V:Mode(Type)|OutVMT]) :-
    merge_unify_VMT(VMT, LocalVMT, OutVMT).

%%process_body_calls(A,B,C,D,E,F) :- errornl(process_body_calls(A,B,C,D,E,F)),fail.
process_body_calls([], [], InVMT, InVMT, OVMT, OVMT).
process_body_calls([Call|Body], [CallCopy|BodyCopy],
		   InVMT, OutVMT, OVMT, OutOVMT) :-
    ip_set(current_predicate, Call),
    ip_set(current_predicate_copy, CallCopy),
    process_call(Call, InVMT, MidVMT, Body, OVMT, MidOVMT),
    '$retry_delay_unifys'(MidVMT, Mid1VMT, MidOVMT, Mid1OVMT),
    process_body_calls(Body, BodyCopy, Mid1VMT, OutVMT, Mid1OVMT, OutOVMT).

%% DEBUG
%%process_call(Call, InVMT, OutVMT, Body, OVMT, _) :- errornl(process_call(Call, InVMT, OutVMT, Body, OVMT)),fail.

process_call(Call, InVMT, OutVMT, _Body, OVMT, OutOVMT) :-
    atom(Call), !,
    InVMT = OutVMT,
    OVMT =  OutOVMT.
process_call(Call, _InVMT, _OutVMT, _Body, _OVMT, _OutOVMT) :-
    var(Call),!,
    writeL(stderr, ["Error: cannot use a variable as a goal", nl_]),
    fail.

/*
  process_call(Call, _InVMT, _OutVMT, _Body, _OVMT, _OutOVMT) :-
    collect_vars(Call, CallVars),
    ip_lookup('$local_vars', LocalVars),
    intersect_list(CallVars, LocalVars, InterVars),
    InterVars \= [], !,
    ip_lookup(check_index_, N1),
    global_state_lookup(gs_check_index_, N2),
    process_globals_locals_error(InterVars).
*/
process_call(TypeCheck, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    '$is_a_type_test'(TypeCheck, Term, Type),
    process_call(typeG(Term, Type), InVMT, OutVMT, _, _, _).
process_call(Call, InVMT, OutVMT, Body, OVMT, OVMT) :-
    compound(Call),
    functor(Call, F, _),
    ( var(F) ; compound(F) ), !,
    Call =.. [_|Args],
    %has_type(F, !(T), InVMT, _),
    get_mt_constraints(F, '!', T, InVMT, InVMT, _, [], OutLE),
    bind_type_constraints(OutLE),
    '$add_annotations_to_type'(T, AT),
    ( AT = '<='(AT1) ; AT = '~>>'(AT1) ),
    '$tuple_type_to_list'(AT1, Types),
    phase_one(Call, Types, InVMT),
    ip_set(check_phase, phase_two),
    phase_two_check(Args, Types, InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, OutVMT),
    check_head_output(OVMT, OutVMT, _, Body).
process_call(type(Term, Type), InVMT, OutVMT, _Body, OVMT, OVMT) :-
    moded(Type),
    !,
    Type = M(T),
    (
      M = '!'
    ->
      process_type_call(Term, T, InVMT, MidVMT),
      collect_vars(Term, TermVars),
      set_modes(TermVars, '!', MidVMT, OutVMT)
    ;
      process_type_call(Term, T, InVMT, OutVMT)
    ).
process_call(type(Term, Type), InVMT, OutVMT, _Body, OVMT, OVMT) :-
    !,
    process_type_call(Term, Type, InVMT, OutVMT).
process_call(type(Term, Type, _), InVMT, OutVMT, _Body, OVMT, OVMT) :-
    !,
    process_type_call(Term, Type, InVMT, OutVMT).
process_call(typeG(Term, Type), InVMT, OutVMT, _Body, OVMT, OVMT) :-
    !,
    process_type_call(Term, Type, InVMT, MidVMT),
    collect_vars(Term, TermVars),
    set_modes(TermVars, '!', MidVMT, OutVMT).
process_call(typeG(Term, Type, _), InVMT, OutVMT, _Body, OVMT, OVMT) :-
    !,
    process_type_call(Term, Type, InVMT, MidVMT),
    collect_vars(Term, TermVars),
    set_modes(TermVars, '!', MidVMT, OutVMT).
process_call(isa(Term, Type), InVMT, OutVMT, _Body, OVMT, OVMT) :-
    !,
    process_type_call(Term, Type, InVMT, MidVMT),
    collect_vars(Term, TermVars),
    set_modes(TermVars, '!', MidVMT, OutVMT).
process_call(ground(Term), InVMT, OutVMT, _Body, OVMT, OVMT) :-
    !,
    collect_vars(Term, TermVars),
    set_modes(TermVars, '!', InVMT, OutVMT).
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = (B '=?' A),
    compound(A),
    A = (_ ++ _), !,
    %ip_lookup(type_checker_context, _(_, _Vars)),
    check_if_ground_call(B, InVMT),
    get_mt_constraints_top(B, '!', string, InVMT, InVMT, _, [], OutLE),
    bind_type_constraints(OutLE),
    extract_concat_args(A, AArgs),
    process_string_concat(AArgs, InVMT, OutVMT).
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = (B '=?' A),
    compound(A),
    A = (_ <> _), !,
    %ip_lookup(type_checker_context, _(_, _Vars)),
    check_if_ground_call(B, InVMT),
    get_mt_constraints_top(B, '!', '$list_type'(T), InVMT, InVMT, _, [], OutLE),
    bind_type_constraints(OutLE),
    extract_append_args(A, AArgs),
    process_list_append(AArgs, T, InVMT, OutVMT).
process_call(Pred, InVMT, OutVMT, Body, OVMT, OVMT) :-
    Pred = (B '=?' A), compound(A), A = '@..'(_, _), !,
    ip_set(call_type, ('$tuple_type'(['?'(term), '!'(term)]) <=)),
    get_mt_constraints_top(B, '!', term, InVMT, InVMT, _, [], OutLE),
    ip_set(check_phase, phase_one),
    phase_one_check([A], ['?'(term)], InVMT, OutLE, AConstr),
    bind_type_constraints(AConstr),
    ip_set(check_phase, phase_two),
    phase_two_check([A], ['?'(term)], InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, OutVMT),
    check_head_output(OVMT, OutVMT, _, Body).
%%% !!! Currently only process single call
process_call(Pred, InVMT, OutVMT, Body, OVMT, OutOVMT) :-
    Pred = once(OPred), !,
    process_call(OPred, InVMT, OutVMT, Body, OVMT, OutOVMT).
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = (not B), !,
    %ip_lookup(type_checker_context, _(_, Vars)),
    check_if_ground_call(B, InVMT),
    process_call(B, InVMT, _, [], [], _),
    OutVMT = InVMT.
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = (X \= Y), !,
    %ip_lookup(type_checker_context, _(_, Vars)),
    B = (X = Y),
    check_if_ground_call(B, InVMT),
    process_call(B, InVMT, _, [], [], _),
    OutVMT = InVMT.    
process_call(Pred, InVMT, OutVMT, Body, OVMT, OutOVMT) :-
    Pred = (A = B), !,
    process_unify_call(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body),
    check_head_output(OutOVMT, OutVMT, _, Body).
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = '$exists'(EVars, B),
    ip_lookup_init_empty('$global_vars', GVars),
    intersect_list(EVars, GVars, InterVars),
    (
      InterVars = []
    ->
      true
    ;
      process_locals_globals_error(InterVars)
    ),
    copy_term(B, BCopy),
    %freeze_term(B, FVars),
    '$goals2list'(B, BL),
    %thaw_term(FVars),
    freeze_term(BCopy),
    '$goals2list'(BCopy, BLCopy),
    %thaw_term(FVarsCopy),
    append(EVars, GVars, AllVars),
    ip_set('$global_vars', AllVars),
    process_body_calls(BL, BLCopy, InVMT, OutVMT, OVMT, _),
    ip_set('$global_vars',GVars),
    ip_lookup_init_empty('$local_vars', LVars),
    append(EVars, LVars, ELVars),
    ip_set('$local_vars', ELVars).
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = '$forall'(QV, A1, A2),!,
    ip_lookup(current_predicate, OG),
    ip_lookup(current_predicate_copy, OGCopy),
    ip_lookup_init_empty('$global_vars', GVars),
    intersect_list(QV, GVars, InterVars),
    (
      InterVars = []
    ->
      true
    ;
      process_locals_globals_error(InterVars)
    ),
    collect_vars((A1,A2), AllA1A2Vars),
    intersect_list(AllA1A2Vars, GVars, UsedGVars),
    get_non_ground_vars(UsedGVars, InVMT, NonGround),
    (
      NonGround = []
    ->
      true
    ;
      process_not_mode_error(NonGround)
    ),
    ip_lookup_init_empty('$underscore_vars', UnderVars),
    ip_lookup_init_empty('$local_vars', LocalVars),
    copy_term(A1, A1Copy),
    '$goals2list'(A1, A1L),
    freeze_term(A1Copy),
    '$goals2list'(A1Copy, A1LCopy),
    process_body_calls(A1L, A1LCopy, InVMT, MidVMT, OVMT, _),
    %% Get the current global vars (including those in A1)
    ip_lookup_init_empty('$global_vars', A1GlobalVars),
    copy_term(A2, A2Copy),
    '$goals2list'(A2, A2L),
    freeze_term(A2Copy),
    '$goals2list'(A2Copy, A2LCopy),
    process_body_calls(A2L, A2LCopy, MidVMT, _, OVMT, _),
    %% Get the current global vars (including those in A1 and A2)
    ip_lookup_init_empty('$global_vars', A2GlobalVars),
    diff_list(A2GlobalVars, GVars, A1A2Vars),
    %diff_list(A1A2Vars, GVars, LocalGobalVars),
    diff_list(A1A2Vars, LocalVars, OtherLocalVars),
    diff_list(OtherLocalVars, QV, QVOtherVars),
    diff_list(QVOtherVars, UnderVars, NonUnderLocals),
    (
      NonUnderLocals = []
    ->
      true
    ;
      process_not_underscore_mode_error(NonUnderLocals)
    ),
    %% check for vars in QV that are on the RHS but not the LHS
    %% get the globals of A2 not in A1
    ip_set(current_predicate, OG),
    ip_set(current_predicate_copy, OGCopy),
    diff_list(A2GlobalVars, A1GlobalVars, A2Only),
    intersect_list(A2Only, QV, QVRHSOnly),
    (
      QVRHSOnly = []
    ->
      true
    ;
      process_forall_quant_vars_error(QVRHSOnly)
    ),
    ip_lookup_init_empty('$local_vars', NewLocalVars),
    append(NewLocalVars, QV, QVLocalVars),
    ip_set('$local_vars', QVLocalVars),
    ip_set('$global_vars', GVars),
    OutVMT = InVMT.
process_call(Pred, InVMT, OutVMT, Body, OVMT, OVMT) :-
    Pred = '$forall_actions'(QV, A1, A2),!,
    process_call('$forall'(QV, A1, A2), InVMT, OutVMT, Body, OVMT, _).
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = remember(_A), !,
    check_if_ground_call(Pred, InVMT),
    process_db_calls(Pred, InVMT, _),
    OutVMT = InVMT.
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = rememberA(_A), !,
    check_if_ground_call(Pred, InVMT),
    process_db_calls(Pred, InVMT, _),
    OutVMT = InVMT.
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = remember_for(_A, T), !,
    check_if_ground_call(Pred, InVMT),
    get_mt_constraints(T, '!', num, InVMT, InVMT, _OutVMT, [], _OutLE),
    process_db_calls(Pred, InVMT, _),
    OutVMT = InVMT.
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = rememberA_for(_A, T), !,
    check_if_ground_call(Pred, InVMT),
    get_mt_constraints(T, '!', num, InVMT, InVMT, _OutVMT, [], _OutLE),
    process_db_calls(Pred, InVMT, _),
    OutVMT = InVMT.
process_call(Pred, InVMT, OutVMT, Body, OVMT, OVMT) :-
    Pred = forget(A), !,
    process_db_calls(Pred, InVMT, Types),
    A =.. [_|Args],
    ip_set(check_phase, phase_two),
    phase_two_check(Args, Types, InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, OutVMT),
    check_head_output(OVMT, OutVMT, _, Body).
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = forget_after(_A, T), !,
    check_if_ground_call(Pred, InVMT),
    get_mt_constraints(T, '!', num, InVMT, InVMT, _OutVMT, [], _OutLE),
    process_db_calls(Pred, InVMT, _),
    OutVMT = InVMT.

process_call(Pred, InVMT, OutVMT, Body, OVMT, OVMT) :-
    Pred = replace_by(A, B), !,
    ip_lookup(current_predicate, replace_by(Arg1, Arg2)),
    ip_lookup(current_predicate_copy, replace_by(Arg1Copy, Arg2Copy)),
    ip_set(current_predicate, Arg1),
    ip_set(current_predicate_copy, Arg1Copy),
    '$type_info'(_, Kind, Types, A, _),
    '$add_annotations'(Types, ATypes, '?'),
    ip_set(call_type,'<='('$tuple_type'(ATypes))),
    get_mt_constraints(A, '?', dynamic, InVMT, InVMT, _MidVMT, [], _),
    A =.. [_|Args],
    ip_set(check_phase, phase_two),
    phase_two_check(Args, ATypes, InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, OutVMT),
    check_head_output(OVMT, OutVMT, _, Body),
    ip_set(check_phase, phase_one),
    ip_set(current_predicate, Arg2),
    ip_set(current_predicate_copy, Arg2Copy),
    '$type_info'(_, Kind, BTypes, B, _),
    '$add_annotations'(BTypes, ABTypes, '!'),
    ip_set(call_type,'<='('$tuple_type'(ABTypes))),
    get_mt_constraints(B, '!', dynamic, OutVMT, OutVMT, _, [], _).
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = call(A), !,
    check_ground_inputs_(A, InVMT, OutVMT).
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = do(A), !,
    check_ground_inputs_(A, InVMT, OutVMT).
process_call(Pred, InVMT, OutVMT, _Body, OVMT, OVMT) :-
    Pred = ground_inputs(A), !,
    check_ground_inputs_(A, InVMT, OutVMT).
/*
process_call(Call, InVMT, _OutVMT, _Body, OVMT, OVMT) :-
    Call = (_A in B),
    (
     check_if_ground_call(B, InVMT)
    ->
     fail
    ;
     !,fail
    ).
process_call(Call, InVMT, OutVMT, Body, OVMT, OVMT) :-
    Call = (_A in _B),
    Types = ['?'(string), '!'(string)], 
    ip_set(call_type, ('$tuple_type'(Types) <=)),
    phase_one(Call, Types, InVMT), !,
    Call =.. [_|Args],
    ip_set(check_phase, phase_two),
    phase_two_check(Args, Types, InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, OutVMT),
    check_head_output(OVMT, OutVMT, _, Body).
process_call(Call, InVMT, OutVMT, Body, OVMT, OVMT) :-
    Call = (_A in _B), 
    Types = ['?'(T), '!'('$list_type'(T))],
    ip_set(call_type, ('$tuple_type'(Types) <=)),
    phase_one(Call, Types, InVMT), !,
    Call =.. [_|Args],
    ip_set(check_phase, phase_two),
    phase_two_check(Args, Types, InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, OutVMT),
    check_head_output(OVMT, OutVMT, _, Body).
process_call(Call, InVMT, OutVMT, Body, OVMT, OVMT) :-
    Call = (_A in _B), !,
    Types = ['?'(T), '!'('$set_type'(T))],
    ip_set(call_type, ('$tuple_type'(Types) <=)),
    phase_one(Call, Types, InVMT),
    Call =.. [_|Args],
    ip_set(check_phase, phase_two),
    phase_two_check(Args, Types, InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, OutVMT),
    check_head_output(OVMT, OutVMT, _, Body).
*/    
process_call(Call, InVMT, OutVMT, Body, OVMT, OVMT) :-
    '$annotated_type_info'(Types, Call),
    '$kind_operator'(Call, Op),
    ip_set(call_type, Op('$tuple_type'(Types))),
    phase_one(Call, Types, InVMT),
    Call =.. [_|Args],
    ip_set(check_phase, phase_two),
    phase_two_check(Args, Types, InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, OutVMT),
    check_head_output(OVMT, OutVMT, _, Body).

'$is_a_type_test'(int(Term), Term, int) :- !.
'$is_a_type_test'(nat(Term), Term, nat) :- !.
'$is_a_type_test'(num(Term), Term, num) :- !.
'$is_a_type_test'(string(Term), Term, string) :- !.
'$is_a_type_test'(atom(Term), Term, atom) :- !.
'$is_a_type_test'(atomic(Term), Term, atomic) :- !.
'$is_a_type_test'(belief(Term), Term, belief) :- !.
'$is_a_type_test'(percept(Term), Term, percept) :- !.
'$is_a_type_test'(percept(Term), Term, percept) :- !.
'$is_a_type_test'(relcall(Term), Term, relcall) :- !.
'$is_a_type_test'(actcall(Term), Term, actcall) :- !.
'$is_a_type_test'(trcall(Term), Term, trcall) :- !.
'$is_a_type_test'([Type](Term), Term, '$list_type'(Type)) :-
    !.
'$is_a_type_test'('$set_enum'([Type])(Term), Term, '$set_type'(Type)) :-
    !,
    '$is_a_type'(Type).
'$is_a_type_test'(Type(Term), Term, Type) :-
    is_declared_type(Type, _), !.

process_string_concat([], InVMT, InVMT).
process_string_concat([A|Args], InVMT, OutVMT) :-
    compound(A),
    A = (T ? Pred), !,
    ip_set(check_phase, phase_one),
    phase_one_check([T], ['?'(string)], InVMT, [], _AConstr),
    ip_set(check_phase, phase_two),
    phase_two_check([T], ['?'(string)], InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, MidVMT),
    '$annotated_type_info'(Types, Pred),
    phase_one(Pred, Types, MidVMT),
    Pred =.. [_|PArgs],
    phase_two_check(PArgs, Types, MidVMT, MidVMT, PredLocalVMT),
    merge_VMT(MidVMT, PredLocalVMT, Mid1VMT),
    process_string_concat(Args, Mid1VMT, OutVMT).
process_string_concat([A|Args], InVMT, OutVMT) :-
    ip_set(check_phase, phase_one),
    phase_one_check([A], ['?'(string)], InVMT, [], _AConstr),
    ip_set(check_phase, phase_two),
    phase_two_check([A], ['?'(string)], InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, MidVMT),
    process_string_concat(Args, MidVMT, OutVMT).
process_string_concat([], InVMT, InVMT).

process_list_append([A|Args], Type, InVMT, OutVMT) :-
    compound(A),
    A = (T ? Pred), !,
    ip_set(check_phase, phase_one),
    phase_one_check([T], ['?'('$list_type'(Type))], InVMT, [], _AConstr),
    ip_set(check_phase, phase_two),
    phase_two_check([T], ['?'('$list_type'(Type))], InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, MidVMT),
    '$annotated_type_info'(Types, Pred),
    phase_one(Pred, Types, MidVMT),
    Pred =.. [_|PArgs],
    phase_two_check(PArgs, Types, MidVMT, MidVMT, PredLocalVMT),
    merge_VMT(MidVMT, PredLocalVMT, Mid1VMT),
    process_list_append(Args, Type, Mid1VMT, OutVMT).
process_list_append([A|Args], Type, InVMT, OutVMT) :-
    ip_set(check_phase, phase_one),
    phase_one_check([A], ['?'('$list_type'(Type))], InVMT, [], _AConstr),
    ip_set(check_phase, phase_two),
    phase_two_check([A], ['?'('$list_type'(Type))], InVMT, [], LocalVMT),
    merge_VMT(InVMT, LocalVMT, MidVMT),
    process_list_append(Args, Type, MidVMT, OutVMT).
process_list_append([], _, InVMT, InVMT).



process_type_call(Term, Type, InVMT, OutVMT) :-
    %ip_set(check_phase, phase_two),
    (
      moded(Type)
    ->
      Type = M(T)
    ;
      M = '+?',
      T = Type
    ),
    get_mt_constraints_top1(Term, unify(M), T, [], [], VMT, [], _),
    process_type_call_aux(VMT, InVMT, OutVMT).

process_type_call_aux([], InVMT, InVMT).
process_type_call_aux([Term : Mode(Type)|VMT], InVMT, OutVMT) :-
    update_VMT(type, InVMT, Term, Mode, Type, MidVMT),
    process_type_call_aux(VMT, MidVMT, OutVMT).

process_db_calls(Pred, InVMT, ATypes) :-
    ip_lookup(current_predicate, Call),
    ip_lookup(current_predicate_copy, CallCopy),
    arg(1, Pred, A),
    arg(1, Call, CallArg),
    arg(1, CallCopy, CallCopyArg),
    ip_set(current_predicate, CallArg),
    ip_set(current_predicate_copy, CallCopyArg),
    '$type_info'(_, _Kind, Types, A, _),
    '$add_annotations'(Types, ATypes, '?'),
    ip_set(call_type,'<='('$tuple_type'(ATypes))),
    get_mt_constraints(A, '?', dynamic, InVMT, InVMT, _, [], _OutLE).
    

    

phase_one(Call, Types, VMT) :-
    Call =.. [_|Args],
    ip_set(check_phase, phase_one),
    ip_set(seen_bang, false),
    phase_one_check(Args, Types, VMT, [], OutLE),
    (
     ip_lookup(seen_bang, true)
    ->
     bind_type_constraints(OutLE)
    ;
     phase_one_check2(Args, Types, VMT, [], OutLE2),
     bind_type_constraints(OutLE2)
    ).
%% DEBUG    
%%phase_one_check(Args, Types, VMT, Mid1LE, OutLE) :- errornl(phase_one_check(Args, Types, VMT, Mid1LE, OutLE)),fail.

phase_one_check([], [], _VMT, LE, LE).
phase_one_check([A|Args], [M(T)|Types], VMT, InLE, OutLE) :-
    get_inner_mode_and_type(M(T), '!', _), !,
    ip_set(seen_bang, true),
    get_mt_constraints_top(A, M, T, VMT, [], _LocalVMT, InLE, MidLE),
    phase_one_check(Args, Types, VMT, MidLE, OutLE).
phase_one_check([A|Args], [M(T)|Types], VMT, InLE, OutLE) :-
    collect_vars(A, AVars),
    get_non_ground_vars(AVars, VMT, NGV),
    NGV = [], !,
    ip_set(seen_bang, true),
    get_mt_constraints_top(A, M, T, VMT, [], _LocalVMT, InLE, MidLE),
    phase_one_check(Args, Types, VMT, MidLE, OutLE).
phase_one_check([_|Args], [_|Types], VMT, InLE, OutLE) :-
    phase_one_check(Args, Types, VMT, InLE, OutLE).


phase_one_check2([], [], _VMT, LE, LE).
phase_one_check2([A|Args], [M(T)|Types], VMT, InLE, OutLE) :-
    get_mt_constraints_top(A, M, T, VMT, [], _LocalVMT, InLE, MidLE),
    phase_one_check2(Args, Types, VMT, MidLE, OutLE).


phase_one_check_nonin([], [], _VMT).
phase_one_check_nonin([A|LeftArgs], [M(T)|LeftTypes], VMT) :-
    get_mt_constraints_top(A, M, T, VMT, [], _LocalVMT, _InLE, _MidLE),
    phase_one_check_nonin(LeftArgs, LeftTypes, VMT).

phase_one_check_aux([], _, LE, LE).
phase_one_check_aux([(Term : Mode(Type))|Rest], VMT, InLE, OutLE) :-
    (
      type_member_of(Term : M(T), VMT)
    ->
      get_inner_mode_and_type(Mode(Type), _IMode, IType),
      get_inner_mode_and_type(M(T), _IM, IT),
      (
        Mode = '!', \+matched_groundedness(Mode(Type), M(T))
      ->
        process_mode_error(Mode, Term)
      ;
        type_le(IT, IType, InLE, MidLE)
      ->
        true
      ;
        process_type_error(Term, IType, IT)
      )
    ;
      (
	Mode = '!'
      ->
	process_mode_error(Mode, Term)
      ;
        MidLE = InLE
      )
    ),
    phase_one_check_aux(Rest, VMT, MidLE, OutLE).

    
phase_two_check([], [], _, VMT, VMT).
phase_two_check([_A|Args], [MT|Types], VMT, InVMT, OutVMT) :-
    all_bang_mode(MT),
    !,
    phase_two_check(Args, Types, VMT, InVMT, OutVMT).
phase_two_check([A|Args], [M(T)|Types], VMT, InVMT, OutVMT) :-
    get_mt_constraints_top(A, M, T, VMT, [], LocalVMT, [], _),
    phase_two_check_aux(LocalVMT, InVMT, MidVMT),
    phase_two_check(Args, Types, VMT, MidVMT, OutVMT).

phase_two_check_aux([], VMT, VMT).

phase_two_check_aux([(Term : Mode(Type))|Rest], InVMT, OutVMT) :-
    !,
    update_mode(Mode, Type, M),
    update_VMT(false, InVMT, Term, M, Type, MidVMT),
    phase_two_check_aux(Rest, MidVMT, OutVMT).

process_unify_call(A, B, InVMT, InVMT, _, _, _) :-
    A == B, !,
    process_unification_succeeds_warning(A, B).
process_unify_call(A, B, InVMT, InVMT, _, _, _) :-
    atomic(A), atomic(B), A \= B, !,
    process_unification_failure_warning(A, B).
process_unify_call(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :-
    process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body).


%%process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :- errornl(process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body)),fail.

%% A and B are identical - ignore
process_unify_call_aux(A, B, InVMT, InVMT, OVMT, OVMT, _) :-
    A == B, !.
%% A has mode ?? so can't be instantiated - mode error
process_unify_call_aux(A, _B, InVMT, InVMT, OVMT, OVMT, _) :-
    var(A),
    type_member_of(A:MTA, InVMT),
    get_inner_mode_and_type(MTA, '??', _), !,
    process_mode_error(doubleq_bound, A).
%% Ditto B
process_unify_call_aux(_A, B, InVMT, InVMT, OVMT, OVMT, _) :-
    var(B),
    type_member_of(B:MTB, InVMT),
    get_inner_mode_and_type(MTB, '??', _), !,
    process_mode_error(doubleq_bound, B).
%% Both A and B are nonvars and A is a function call
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    nonvar(A), nonvar(B),
    compound(A),
    '$get_function_type_info'(A, _, _, _), !,
    process_call('$unify'(A, B), InVMT, OutVMT, [], OVMT, OutOVMT),
    '$add_to_unify_delays'(X, A),
    '$add_to_unify_delays'(X, B).
%% Both A and B are nonvars and B is a function call
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    nonvar(A), nonvar(B),
    compound(B),
    '$get_function_type_info'(B, _, _, _), !,
    process_call('$unify'(A, B), InVMT, OutVMT, [], OVMT, OutOVMT),
    '$add_to_unify_delays'(X, A),
    '$add_to_unify_delays'(X, B).

process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    nonvar(B), atomic(A), !,
    (
      A == B
    ->
      InVMT = OutVMT, OVMT = OutOVMT
    ;
      process_unification_failure_warning(A, B)
    ).
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    nonvar(A), atomic(B), !,
    (
      A == B
    ->
      InVMT = OutVMT, OVMT = OutOVMT
    ;
      process_unification_failure_warning(A, B)
    ).
%% neither sides are vars so must be compound
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :-
    compound(A), compound(B), !,
    A =.. AList,
    B =.. BList,
    process_unify_call_list(AList, BList, InVMT, OutVMT, OVMT, OutOVMT, Body).
    
%% At this point at least one side is a variable

%%   A and B are both new vars
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(A), var(B), 
    \+ type_member_of(B:_, InVMT), \+ type_member_of(A:_, InVMT), !,
    OutVMT = InVMT,
    OVMT = OutOVMT,
    update_check_global_vars(A),
    update_check_global_vars(B),
    '$add_to_unify_delays'(A, B).
%%   A is a new variable, B is a known variable
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(A), var(B), 
    \+ type_member_of(A:_, InVMT), type_member_of(B:T, InVMT), !,
    OutVMT = [A:T|InVMT],
    OVMT = OutOVMT,
    update_check_global_vars(A),
    '$add_to_unify_delays'(A, B).
%%   B is a new variable, A is a known variable
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(A), var(B), 
    \+ type_member_of(B:_, InVMT), type_member_of(A:T, InVMT), !,
    OutVMT = [B:T|InVMT],
    OVMT = OutOVMT,
    update_check_global_vars(B),
    '$add_to_unify_delays'(A, B).

%%   A and B are both in InVMT in '!' mode
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(A), var(B), 
    type_member_of(A:MTA, InVMT),
    type_member_of(B:MTB, InVMT),
    get_inner_mode_and_type(MTA, '!', TA),
    get_inner_mode_and_type(MTB, '!', TB),
    !,
    OVMT = OutOVMT,
    (
      type_le(TA, TB)
    ->
      type_remove_from(B:_, InVMT, MidVMT),
      OutVMT = [B : '!'(TA)|MidVMT]
    ;
      type_le(TB, TA)
    ->
      type_remove_from(A:_, InVMT, MidVMT),
      OutVMT = [A : '!'(TB)|MidVMT]
    ;
      process_inconsistent_type_error(A, TA, TB)
    ).
%%   A in VMT in '!' mode  (B a var)
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(A), var(B), 
    type_member_of(A:MTA, InVMT),
    get_inner_mode_and_type(MTA, '!', TA), !,
    type_member_of(B:_(TB), InVMT),
    update_check_global_vars(B),
    OVMT = OutOVMT,
    (
      compatible(TA, TB)
    ->
      true
    ;
      process_inconsistent_type_error(A, TA, TB)
    ),
    (
      type_remove_from(B:_(TB), InVMT, MidVMT)
    ->
      OutVMT = [B:'!'(TA)|MidVMT]
    ;
      OutVMT = [B:'!'(TA)|InVMT]
    ).
%%   B in VMT in '!' mode (A a var)
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(A), var(B), 
    type_member_of(B:MTB, InVMT),
    get_inner_mode_and_type(MTB, '!', _), !,
    process_unify_call_aux(B, A, InVMT, OutVMT, OVMT, OutOVMT, _).
%%   A and B both in VMT - neither in ! mode
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(A), var(B), 
    type_member_of(A:_, InVMT),
    type_member_of(B:_, InVMT), !,
    process_call('$unify'(A, B), InVMT, OutVMT, [], OVMT, OutOVMT),
    '$add_to_unify_delays'(A, B).
%% B is a var, A isn't
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(B), !,
    process_unify_call_aux(B, A, InVMT, OutVMT, OVMT, OutOVMT, _).
%% From here on A is a variable and B isn't
%% B is a "ground" term
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    get_mt_constraints_top(B, '?', TB, InVMT, [], _, [], LE),
    bind_type_constraints(LE),
    collect_vars(B, BVars),
    get_non_ground_vars(BVars, InVMT, NGV),
    ip_lookup_init_empty('$local_vars', Locals),
    diff_list(NGV, Locals, Free),
    update_check_global_vars(A),
    Free = [], !,
    process_unify_call_ground_term(A, TB, InVMT, OutVMT, OVMT, OutOVMT, _).
%% A is in ! mode
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    type_member_of(A:'!'(TA), InVMT), !,
    get_mt_constraints_top(B, '?', TA, InVMT, InVMT, MidVMT, [], LE),
     bind_type_constraints(LE),
    collect_vars(B, BVars),
    set_modes(BVars, '!', MidVMT, OutVMT),
    OVMT = OutOVMT.
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    type_member_of(A:M(TA), InVMT), !,
    %get_inner_mode_and_type(M(TA), _, ITA),
    phase_two_check([B],[M(TA)], InVMT, InVMT, OutVMT),
    OVMT = OutOVMT,
    '$add_to_unify_delays'(A, B).
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    update_check_global_vars(A),
    update_check_global_vars(B),
    '$add_to_unify_delays'(A, B),
    InVMT = OutVMT, OVMT = OutOVMT.
    
process_unify_call_ground_term(A, TB, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    type_member_of(A:AT, InVMT), 
    get_inner_mode_and_type(AT, '!', TA), !,
    (
      compatible(TA, TB)
    ->
      true
    ;
      process_inconsistent_type_error(A, TA, TB)
    ),    
    OVMT = OutOVMT,
    (
      type_le(TA, TB)
    ->
      OutVMT = InVMT
    ;
      type_le(TB, TA)
    ->
      type_remove_from(A:_, InVMT, MidVMT),
      OutVMT = [A : '!'(TB)|MidVMT]
    ;
      process_inconsistent_type_error(A, TA, TB)
    ).
process_unify_call_ground_term(A, TB, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    type_member_of(A:AT, InVMT), !,
    get_inner_mode_and_type(AT, _, TA), !,
    (
      compatible(TA, TB)
    ->
      true
    ;
      process_inconsistent_type_error(A, TA, TB)
    ),    
    OVMT = OutOVMT,
   type_remove_from(A:_, InVMT, MidVMT),
    OutVMT = [A : '!'(TB)|MidVMT].
process_unify_call_ground_term(A, TB, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    OVMT = OutOVMT,
    OutVMT = [A : '!'(TB)|InVMT].

    
'$retry_delay_unifys'(InVMT, OutVMT, OVMT, OutOVMT) :-
    ip_lookup('$delayed_unify', Unifys),
    (
      Unifys = []
    ->
      InVMT = OutVMT, OVMT = OutOVMT
    ;
      sort(Unifys, SUnifys),
      ip_set('$delayed_unify', []),
      '$retry_delay_unifys'(SUnifys, InVMT, OutVMT, OVMT, OutOVMT)
    ).

%%'$retry_delay_unifys'(SUnifys, InVMT, OutVMT, OVMT, OutOVMT) :- errornl('$retry_delay_unifys'(SUnifys, InVMT, OutVMT, OVMT, OutOVMT)),fail.

'$retry_delay_unifys'(SUnifys, InVMT, OutVMT, OVMT, OutOVMT) :-
    '$retry_delay_unifys_aux'(SUnifys, InVMT, MidVMT, OVMT, MidOVMT),
    ip_lookup('$delayed_unify', Unifys),
    sort(Unifys, S2Unifys),
    (
      S2Unifys == SUnifys
    ->
      OutVMT = MidVMT,
      OutOVMT = MidOVMT
    ;
      ip_set('$delayed_unify', []),
      '$retry_delay_unifys'(S2Unifys, MidVMT, OutVMT, MidOVMT, OutOVMT)
    ).
      
'$retry_delay_unifys_aux'([], VMT, VMT, OVMT, OVMT).

'$retry_delay_unifys_aux'([(A=B)|SUnifys], InVMT, OutVMT, OVMT, OutOVMT) :-
    process_unify_call_aux(A, B, InVMT, MidVMT, OVMT, MidOVMT, _),
    '$retry_delay_unifys_aux'(SUnifys, MidVMT, OutVMT, MidOVMT, OutOVMT).

    
'$add_to_unify_delays'(A, B) :-
    ip_lookup('$delayed_unify', Unifys),
    ( var(Unifys) -> Unifys = [] ; true ),
    ip_set('$delayed_unify', [(A=B)|Unifys]).

%% This should only be about vars that are in post ? mode as all
%% other modes will have been processed before getting here
%% (or var 
'$end_unifys_check'([], _, _).
'$end_unifys_check'([(A=B)|Unifys], VMT, OVMT) :-
    type_member_of(A:M(TA), VMT), !,
    get_inner_mode_and_type(M(TA), IM, ITA),
    %get_mt_constraints_top(A, '?', TA, VMT, [], _, [], ALE),
    %bind_type_constraints(ALE),
    get_mt_constraints_top(B, IM, ITA, [], [], XX, [], BLE),
    bind_type_constraints(BLE),
    check_head_output_end(OVMT, XX, _Unbound, []),
    '$end_unifys_check'(Unifys, VMT).

'$update_type'(T, NewT, RT) :-
    moded(T), !,
    T = M(T1),
    '$update_type'(T1, NewT, RT1),
    RT = M(RT1).
'$update_type'(T, NewT, RT) :-
    is_a_constructor_type(T, Constr, T1),
    is_a_constructor_type(NewT, Constr, NewT1),
    '$update_type'(T1, NewT1, RT1),
    add_constructor(Constr, RT1, RT).
'$update_type'(_T, NewT, NewT).
/*   
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(B), var(A),
    \+ type_member_of(A:_, InVMT), \+ type_member_of(A:_, OVMT),!,
    InVMT = OutVMT,
    OVMT = OutOVMT,
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
%% B is in neither VMT and A is a var so we just bind A and B
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(A), var(B),
    \+ type_member_of(B:_, InVMT), \+ type_member_of(B:_, OVMT),!,
    InVMT = OutVMT,
    OVMT = OutOVMT,
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
%% Now we deal with both A and B variables and, from above,
%% they must be in InVMT, OutVMT or both
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :-
    var(A), var(B), !,
    ip_set(unify_check, in),
    unified_check_VMT(A, B, InVMT, OutVMT),
    ip_set(unify_check, out),
    unified_check_VMT(A, B, OVMT, OutOVMT),
    check_head_output_unify(A, B, OVMT, OutVMT, Body),
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
%% Exactly one is a var (and must be in one of the VMTs)
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :-
    var(B), !,
    process_unify_call_aux(B, A, InVMT, OutVMT, OVMT, OutOVMT, Body).
%% Now check when A is a var and B isn't
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    var(A), ground(B), !,
    has_type(B, BType, [], _),
    unified_ground_check_VMT(A, B, BType, InVMT, OutVMT),
    unified_ground_check_VMT(A, B, BType, OVMT, OutOVMT),
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
%%% When A is a var from here on B must be a constructor term containing vars
%% A is in InVMT but not in OVMT
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    var(A), type_remove_from(A:MA(TA), InVMT, AVMT),
    \+ type_member_of(A:_, OVMT), !,
    get_mt_constraints_top(B, unify(MA), TA, InVMT, [], LocalVMT, [], _),
    merge_unify_VMT(AVMT, LocalVMT, OutVMT),
    OVMT = OutOVMT,
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    var(A), type_remove_from(A:MA(TA), InVMT, AVMT), !,
    type_remove_from(A:_, OVMT, MidOVMT),
    get_mt_constraints_top(B, unify(MA), TA, InVMT, [], LocalVMT, [], _),
    merge_unify_VMT(AVMT, LocalVMT, OutVMT),
    merge_unify_VMT(MidOVMT, LocalVMT, OutOVMT),
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :-
    var(A), 
    type_remove_from(A:MA(TA), OVMT, AOVMT), !,
    get_mt_constraints_top(B, unify(MA), TA, InVMT, [], LocalVMT, [], _),
    merge_unify_VMT(AOVMT, LocalVMT, OutOVMT),
    OutVMT = InVMT,
    check_head_output_unify(A, B, OutOVMT, OutVMT, Body),
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
    
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :-
    A =.. AList,
    B =.. BList,
    process_unify_call_list(AList, BList, InVMT, OutVMT, OVMT, OutOVMT, Body).
*/
process_unify_call_list([], [], InVMT, InVMT, OVMT, OVMT, _Body).
process_unify_call_list([A|AList], [B|BList], InVMT, OutVMT,
                        OVMT, OutOVMT, Body) :-
    process_unify_call_aux(A, B, InVMT, MidVMT, OVMT, MidOVMT, Body),
    process_unify_call_list(AList, BList, MidVMT,
                            OutVMT, MidOVMT, OutOVMT, Body).
    
    
%%process_unify_call(A, B, InVMT, _, OVMT, _, _) :- errornl(process_unify_call(A, B, InVMT, OVMT)),fail.
/*
process_unify_call(A, B, InVMT, InVMT, _, _, _) :-
    A == B, !,
    process_unification_succeeds_warning(A, B).
process_unify_call(A, B, InVMT, InVMT, _, _, _) :-
    ground(A), ground(B), A \= B, !,
    process_unification_failure_warning(A, B).
process_unify_call(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :-
    process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body).

%%process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :- errornl(process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body)),fail.

process_unify_call_aux(A, B, InVMT, InVMT, OVMT, OVMT, _) :-
    A == B, !.
%% A is in neither VMT so we just bind A and B
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(A), \+ type_member_of(A:_, InVMT), \+ type_member_of(A:_, OVMT),!,
    InVMT = OutVMT,
    OVMT = OutOVMT,
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
%% B is in neither VMT so we just bind A and B
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _) :-
    var(B), \+ type_member_of(B:_, InVMT), \+ type_member_of(B:_, OVMT),!,
    InVMT = OutVMT,
    OVMT = OutOVMT,
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
%% Now we deal with both A and B variables and, from above,
%% they must be in InVMT, OutVMT or both
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :-
    var(A), var(B), !,
    ip_set(unify_check, in),
    unified_check_VMT(A, B, InVMT, OutVMT),
    ip_set(unify_check, out),
    unified_check_VMT(A, B, OVMT, OutOVMT),
    check_head_output_unify(A, B, OVMT, OutVMT, Body),
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
%% Exactly one is a var (and must be in one of the VMTs)
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :-
    var(B), !,
    process_unify_call_aux(B, A, InVMT, OutVMT, OVMT, OutOVMT, Body).
%% Now check when A is a var and B isn't
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    var(A), ground(B), !,
    has_type(B, BType, [], _),
    unified_ground_check_VMT(A, B, BType, InVMT, OutVMT),
    unified_ground_check_VMT(A, B, BType, OVMT, OutOVMT),
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
%%% When A is a var from here on B must be a constructor term containing vars
%% A is in InVMT but not in OVMT
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    var(A), type_remove_from(A:MA(TA), InVMT, AVMT),
    \+ type_member_of(A:_, OVMT), !,
    get_mt_constraints_top(B, unify(MA), TA, InVMT, [], LocalVMT, [], _),
    merge_unify_VMT(AVMT, LocalVMT, OutVMT),
    OVMT = OutOVMT,
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    var(A), type_remove_from(A:MA(TA), InVMT, AVMT), !,
    type_remove_from(A:_, OVMT, MidOVMT),
    get_mt_constraints_top(B, unify(MA), TA, InVMT, [], LocalVMT, [], _),
    merge_unify_VMT(AVMT, LocalVMT, OutVMT),
    merge_unify_VMT(MidOVMT, LocalVMT, OutOVMT),
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :-
    var(A), 
    type_remove_from(A:MA(TA), OVMT, AOVMT), !,
    get_mt_constraints_top(B, unify(MA), TA, InVMT, [], LocalVMT, [], _),
    merge_unify_VMT(AOVMT, LocalVMT, OutOVMT),
    OutVMT = InVMT,
    check_head_output_unify(A, B, OutOVMT, OutVMT, Body),
    thaw_term((A,B)),
    A = B,
    freeze_term((A,B)).
%% both are compound terms
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    atom(A), !,
    (
      A == B
    ->
      InVMT = OutVMT, OVMT = OutOVMT
    ;
      process_unification_failure_warning(A, B)
    ).
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, _Body) :-
    atom(B), !,
    (
      A == B
    ->
      InVMT = OutVMT, OVMT = OutOVMT
    ;
      process_unification_failure_warning(A, B)
    ).
    
process_unify_call_aux(A, B, InVMT, OutVMT, OVMT, OutOVMT, Body) :-
    A =.. AList,
    B =.. BList,
    process_unify_call_list(AList, BList, InVMT, OutVMT, OVMT, OutOVMT, Body).

process_unify_call_list([], [], InVMT, InVMT, OVMT, OVMT, _Body).
process_unify_call_list([A|AList], [B|BList], InVMT, OutVMT,
                        OVMT, OutOVMT, Body) :-
    process_unify_call_aux(A, B, InVMT, MidVMT, OVMT, MidOVMT, Body),
    process_unify_call_list(AList, BList, MidVMT,
                            OutVMT, MidOVMT, OutOVMT, Body).
*/

unified_check_VMT(A, B, InVMT, OutVMT) :-
    type_remove_from(A:MA(TA), InVMT, MidVMT),
    type_remove_from(B:MB(TB), MidVMT, Mid1VMT), !,
    get_inner_mode_and_type(MA(TA), _IMA, ITA),
    get_inner_mode_and_type( MB(TB), _IMB, ITB),
    min_mode(MA, MB, UM),
    (
     type_le(ITA, ITB)
    ->
      (
        ( MA = '!' ; ip_lookup(unify_check, out) )
      ->
        OutVMT = [A : UM(TA)|Mid1VMT]
      ;
        OutVMT = [A : UM(TB)|Mid1VMT]
      )
    ;
     type_le(ITB, ITA)
    ->
      (
        ( MB = '!' ; ip_lookup(unify_check, out) )
      ->
        OutVMT = [A : UM(TB)|Mid1VMT]
      ;
        OutVMT = [A : UM(TA)|Mid1VMT]
      )
    ;
     process_inconsistent_type_error(A, TA, TB)
    ).
unified_check_VMT(_A, _B, InVMT, InVMT).


unified_ground_check_VMT(A, B, TB, InVMT, OutVMT) :-
    type_remove_from(A:MA(TA), InVMT, MidVMT),!,
    get_inner_mode_and_type(MA(TA), _IMA, ITA),
    (
     type_le(TB, ITA)
    ->
     OutVMT = [A : '!'(TB)|MidVMT]
    ;
     type_le(ITA, TB), MA = '!'
    ->
      process_unification_failure_warning(A, B),
      OutVMT = [A : '!'(TA)|MidVMT]
    ;
     process_inconsistent_type_error(A, TA, TB)
    ).
unified_ground_check_VMT(_A, _B, _TB, VMT, VMT).
    
unified_check_OVMT(A, A, [], []).
unified_check_OVMT(A, B, [A:MA(TA)|OVMT], OutOVMT) :-
    var(B), type_remove_from(B:MB(TB), OVMT, TOVMT), !,
    min_mode(MA, MB, UM),
    get_inner_mode_and_type(MA(TA), _IMA, ITA),
    get_inner_mode_and_type( MB(TB), _IMB, ITB),
    (
     type_le(ITA, ITB)
    ->
     A = B, OutOVMT = [A : UM(TA)|TOVMT]
    ;
     type_le(ITB, ITA)
    ->
     A = B, OutOVMT = [A : UM(TB)|TOVMT]
    ;
     process_inconsistent_type_error(A, TA, TB)
    ).
unified_check_OVMT(A, B, [A:MA(TA)|OVMT], [A:MA(TA)|OVMT]) :-
    var(B), !, A = B.
unified_check_OVMT(A, B, [A:MA(TA)|OVMT], OutOVMT) :-
    get_mt_constraints_top(B, MA, TA, [], [], LocalVMT, [], _),
    A = B,
    merge_VMT(OVMT, LocalVMT, OutOVMT).

check_list_constr(Pred, VMT, VarNames,TVars) :-
    '$freeze_underscores'(Pred, VarNames),
    freeze_term(TVars),
    get_bang_vars(VMT, Bangs),
    freeze_term(Bangs),
    freeze_term(Pred, VVV),
    %% VVV are all the vars that are not underscore vars and not
    %% vars in the term and not in ! mode
    (
     VVV = []
    ->
     true
    ;
     process_not_underscore_mode_error(VVV)
    ).
 
check_list_constr_out(T, VMT, Term) :-
    get_bang_vars(VMT, Bangs),
    freeze_term(Bangs),
    freeze_term(T, VVV),
    %% VVV are all the vars that are not underscore vars and not
    %% vars in the term and not in ! mode
    (
     VVV = []
    ->
     true
    ;
     process_mode_error_list_constr(T, VVV, Term)
    ).

check_if_ground_call(Call, VMT) :-
    collect_vars(Call, AllVars),
    ip_lookup_init_empty('$local_vars', LVars),    
    ip_lookup_init_empty('$underscore_vars', UVars),
    union_list(LVars, UVars, LUVars),
    diff_list(AllVars, LUVars, UsedGVars),
    get_non_ground_vars(UsedGVars, VMT, NonGround),
    (
      NonGround = []
    ->
      true
    ;
      process_not_mode_error(NonGround)
    ).


get_non_ground_vars([], _VMT, []).
get_non_ground_vars([V|Vars], VMT, Out) :-
    (
      type_member_of(V:'!'(_), VMT)
    ->
      Out = Out1
    ;
      Out = [V|Out1]
    ),
    get_non_ground_vars(Vars, VMT, Out1).

split_on_VMT([], _, _, [], []).
split_on_VMT([V|Vs], VMT, OVMT, Old, New) :-   
    (
      ( type_member_of(V:_, VMT) ; type_member_of(V:_, OVMT) )
    ->
      Old = [V|Old1], New = New1
    ;
      Old = Old1, New = [V|New1]
    ),
    split_on_VMT(Vs, VMT, OVMT, Old1, New1).

check_forall_quant_vars([], _A1Vars, _A2Vars, []).
check_forall_quant_vars([V|QV], A1Vars, A2Vars, BadVars) :-
    (
     member_eq(V, A2Vars)
    ->
     member_eq(V, A1Vars)
    ;
     true
    ), !, 
    check_forall_quant_vars(QV, A1Vars, A2Vars, BadVars).
check_forall_quant_vars([V|QV], A1Vars, A2Vars, [V|BadVars]) :-
    check_forall_quant_vars(QV, A1Vars, A2Vars, BadVars).

quant_vars_overlap([], _OuterVars, []).
quant_vars_overlap([V|QV], OuterVars, [V|NonLocal]) :-
    member_eq(V, OuterVars), !,
    quant_vars_overlap(QV, OuterVars, NonLocal).
quant_vars_overlap([_V|QV], OuterVars, NonLocal) :-
    quant_vars_overlap(QV, OuterVars, NonLocal).
/*
  type_le(T1, T2, InConstr, OutConstr) :-
    thaw_term(T1, Vars), Vars \= [], !, 
    T1 = T2, freeze_term(Vars), OutConstr = InConstr.
type_le(T1, T2, InConstr, OutConstr) :-
    thaw_term(T2, Vars), Vars \= [], !, 
    T1 = T2, freeze_term(Vars), OutConstr = InConstr.
type_le(T1, T2, InConstr, OutConstr) :-
    thawed_var(T1), !,
    merge_le(le(T1, T2), InConstr, OutConstr).
type_le(T1, T2, InConstr, OutConstr) :-
    thawed_var(T2), !,
    merge_le(le(T1, T2), InConstr, OutConstr).
type_le(T1, T2, InConstr, OutConstr) :-
    frozen_var(T1), !,
    T1 = T2,
    OutConstr = InConstr.
*/

order_le(le(A, B), le(C, D)) :-
    le(B,A) @< le(D, C).

type_le(T1, T2, InConstr, OutConstr) :-
    type_le_aux(T1, T2, InConstr, MidConstr),
    redo_type_le(MidConstr, OutConstr).

%%redo_type_le(InConstr, OutConstr) :- errornl(redo_type_le(InConstr, OutConstr)),fail.

redo_type_le(InConstr, OutConstr) :-
    sort(InConstr, SConstr, order_le),
    redo_type_le_aux(SConstr, MidConstr),
    sort(MidConstr, SMidConstr, order_le),
    (
     SConstr == SMidConstr
    ->
     OutConstr = SConstr
    ;
     redo_type_le(SMidConstr, OutConstr)
    ).

%% find loops le(T1, T2), le(T2, T3), ..., le(Tn1, Tn), le(Tn, T1)
%% bind all together to same term - else fail
%%redo_type_le_aux(InConstr, OutConstr) :-errornl(redo_type_le_aux(InConstr, OutConstr)),fail.
redo_type_le_aux(InConstr, OutConstr) :-
    member(le(T1, T2), InConstr), T1 == T2, !,
    diff_list(InConstr, [le(T1, T2)], OutConstr).
redo_type_le_aux(InConstr, OutConstr) :-
    member(le(T1, T2), InConstr),
    find_a_loop(InConstr, T1, T2, [], Loop), !,
    T1 = T2,
    '$unify_all'(Loop, T1),
    OutConstr = InConstr.
%% find chains that start from a ground type and end with a ground type
%% T1 = End - bind all
%% T1 > End - fail
redo_type_le_aux(InConstr, OutConstr) :-
    member(le(T1, T2), InConstr), ground(T1),
    find_a_chain(T2, InConstr, Chain, Tn),
    (
     T1 = Tn
    ->
     !,
     '$unify_all'(Chain, T1),
    OutConstr = InConstr
    ;
     type_le_aux(Tn, T1, [], _)
    ->
     !,
     fail
    ;
     type_le_aux(T1, Tn, [], _)
    ->
     T2 = Tn, le1_remove_from(le(T1, T2), InConstr, OutConstr)
    ;
      !,
      fail
    ).
%% remove le(T1, T2) with T1 == T2
redo_type_le_aux(InConstr, OutConstr) :-
    le1_remove_from(le(T1, T2), InConstr, OutConstr),
    T1 == T2, !.
%% le(T, T1), le(T, T2) and T1 and T2 incompatible -> fail
redo_type_le_aux(InConstr, _OutConstr) :-
    member(le(T, T1), InConstr), ground(T1),
    member(le(Tx, T2), InConstr), T == Tx, ground(T2),
    (
     compatible(T1, T2)
    ->
     fail
    ;
     !,
     fail
    ).
%% le(T1, T2), le(T2, T3), le(T1, T3) -> remove le(T1, T3)
redo_type_le_aux(InConstr, OutConstr) :-
    member(le(T1, T2), InConstr),
    member(le(T2x, T3), InConstr), T2x == T2,
    le1_remove_from(le(T1, T3x), InConstr, OutConstr),
    T3x == T3, !.
%% le(T1, T2), le(T1, T3) and T2 =< T3 -> remove le(T1, T3)
redo_type_le_aux(InConstr, OutConstr) :-
    member(le(T1, T2), InConstr), ground(T2),
    le1_remove_from(le(T1, T3), InConstr, OutConstr), T2 \== T3,
    ground(T3), type_le_aux(T2, T3, [], _), !.
%% le(T2, T1), le(T3, T1) and T2 =< T3 -> remove le(T2, T1)
redo_type_le_aux(InConstr, OutConstr) :-
    member(le(T3, T1), InConstr), ground(T3),
    le2_remove_from(le(T2, T1), InConstr, OutConstr), T2 \== T3,
    ground(T2), type_le_aux(T2, T3, [], _), !.
%% le(T1, T2) can be simplified
redo_type_le_aux(InConstr, OutConstr) :-
    delete(le(T1, T2), InConstr, MidConstr),
    type_le_aux(T1, T2, [], C),
    C \== [le(T1, T2)], !,
    append(C, MidConstr, OutConstr).
%% catchall - no change
redo_type_le_aux(InConstr, InConstr).

find_a_loop(_Constr, T1, T2, InLoop, Loop) :-
    T1 == T2, !, Loop = InLoop.
find_a_loop(Constr, T1, T2, InLoop, Loop) :-
    member(le(T3, T4), Constr), T2 == T3, \+member_eq(T4, InLoop), !,
    find_a_loop(Constr, T1, T4, [T4|InLoop], Loop).


find_a_chain(T2, _InConstr, Chain, Tn) :-
    ground(T2), !,
    Chain = [], Tn = T2.
find_a_chain(T2, InConstr, [T4|Chain], Tn) :-
    member(le(T3, T4), InConstr), T2 == T3, !,
    find_a_chain(T4, InConstr, Chain, Tn).


    
    
%% DEBUG
%%type_le_aux(T1, T2, InConstr, OutConstr) :- errornl(type_le_aux(T1, T2, InConstr, OutConstr)),fail.

type_le_aux(_T1, T2, InConstr, OutConstr) :-
    T2 == top, !, 
    OutConstr = InConstr.
type_le_aux(T1, T2, InConstr, OutConstr) :-
    T1 == T2, !, 
    OutConstr = InConstr.
type_le_aux(T1, T2, InConstr, OutConstr) :-
    frozen_var(T1), T2 == term, !,
    OutConstr = InConstr.
type_le_aux(T1, T2, InConstr, OutConstr) :-
    frozen_var(T1), !,
    T1 = T2,
    OutConstr = InConstr.
type_le_aux(T1, T2, InConstr, OutConstr) :-
    frozen_var(T2), !,
    T1 = T2,
    OutConstr = InConstr.
type_le_aux(T1, T2, InConstr, OutConstr) :-
    var(T2), !,
    OutConstr = [le(T1, T2)|InConstr].
type_le_aux(T1, T2, InConstr, OutConstr) :-
    nonvar(T1),
    freeze_term(T1, V),
    '$defined_type'(T1, '$type_macro'(DefT1)),
    thaw_term(V),
    type_le_aux(DefT1 , T2, InConstr, OutConstr), !.
type_le_aux(T1, T2, InConstr, OutConstr) :-
    nonvar(T2),
    freeze_term(T2, V),
    '$defined_type'(T2, '$type_macro'(DefT2)),
    thaw_term(V),
    type_le_aux(T1 , DefT2, InConstr, OutConstr), !.
type_le_aux(T1, (T2D -> T2R), InConstr, OutConstr) :-
    !,
    T1 = (T1D -> T1R),
    type_le_aux(T2D, T1D, InConstr, Mid1Constr),
    type_le_aux(T1R, T2R, Mid1Constr, OutConstr).
type_le_aux(T, '<='(T2), InConstr, OutConstr) :-
    var(T),
    !,
    '$tuple_type_to_list'(T2, Types2),
    length(Types2, N),
    length(Types1, N),
    T = '<='('$tuple_type'(Types1)),
    '$add_annotations'(Types2, ATypes2, '!'),
    map(wrap_var_functor, Types1, ATypes1),
    type_annotated_le(ATypes1, ATypes2, InConstr, OutConstr).
type_le_aux(T, '~>>'(T2), InConstr, OutConstr) :-
    var(T),
    !,
    '$tuple_type_to_list'(T2, Types2),
    length(Types2, N),
    length(Types1, N),
    T = '~>>'('$tuple_type'(Types1)),
    '$add_annotations'(Types2, ATypes2, '!'),
    map(wrap_var_functor, Types1, ATypes1),
    type_annotated_le(ATypes1, ATypes2, InConstr, OutConstr).
type_le_aux(T, '~>'(T2), InConstr, OutConstr) :-
    var(T),
    !,
    '$tuple_type_to_list'(T2, Types2),
    length(Types2, N),
    length(Types1, N),
    T = '~>'('$tuple_type'(Types1)),
    '$add_annotations'(Types2, ATypes2, '!'),
    map(wrap_var_functor, Types1, ATypes1),
    type_annotated_le(ATypes1, ATypes2, InConstr, OutConstr).
type_le_aux(T, '$tuple_type'(T2), InConstr, OutConstr) :-
    !,
   T =  '$tuple_type'(T1),
    list_type_le_aux(T1, T2, InConstr, OutConstr).
type_le_aux(T, '$set_type'(T2), InConstr, OutConstr) :-
    var(T),
    !,
    T =  '$set_type'(T1),
    type_le_aux(T1, T2, InConstr, OutConstr).
type_le_aux(T, '$list_type'(T2), InConstr, OutConstr) :-
    var(T),
    !,
    T =  '$list_type'(T1),
    type_le_aux(T1, T2, InConstr, OutConstr).
type_le_aux(T1, T2, InConstr, OutConstr) :-
    var(T1), !,
    OutConstr = [le(T1, T2)|InConstr].

%% from here on T1 and T2 are both nonvars
type_le_aux(belief, relcall, InConstr, OutConstr) :-
    !, 
    OutConstr = InConstr.
type_le_aux(percept, relcall, InConstr, OutConstr) :-
    !, 
    OutConstr = InConstr.
type_le_aux('$enum_type'(T1), '$enum_type'(T2), InConstr, OutConstr) :-
    !,
    forall(member(X, T1), member(X, T2)), 
    OutConstr = InConstr.
type_le_aux('$enum_type'(T1), atom, InConstr, OutConstr) :-  
    forall( member(A,T1), atom(A)), !, 
    OutConstr = InConstr.
type_le_aux('$enum_type'(T1), string, InConstr, OutConstr) :-  
    forall( member(A,T1), string(A)), !, 
    OutConstr = InConstr.
type_le_aux('$enum_type'(T1), nat, InConstr, OutConstr) :-  
    forall( member(A,T1), is_nat(A)), !, 
    OutConstr = InConstr.
type_le_aux('$enum_type'(T1), int, InConstr, OutConstr) :-  
    forall( member(A,T1), integer(A)), !, 
    OutConstr = InConstr.
type_le_aux('$enum_type'(T1), num, InConstr, OutConstr) :-  
    forall( member(A,T1), number(A)), !, 
    OutConstr = InConstr.
type_le_aux('$enum_type'(T1), atomic, InConstr, OutConstr) :- 
    forall( member(A,T1), (atomic(A);string(A))), !, 
    OutConstr = InConstr.
type_le_aux('$union_type'(T1), T2, InConstr, OutConstr) :-
    is_primitive_or_defined_type(T2), !,
    forall( member(E,T1), type_le_aux(E,T2, [], []) ), 
    OutConstr = InConstr.
 %% probably remove as atomic wouldn't be generated
type_le_aux(atomic, '$union_type'(T2), InConstr, OutConstr) :- 
    forall( member(T,[atom,nat,int,num, string]),
	    (member(ST,T2),type_le_aux(T,ST,[],[]))), 
    OutConstr = InConstr.
type_le_aux('$union_type'(T1), '$union_type'(T2), InConstr, OutConstr) :-
    forall(member(E1,T1), 
	   (member(E2,T2), once(type_le_aux(E1,E2, [], [])))), 
    OutConstr = InConstr.
type_le_aux(T1, '$union_type'(T2), InConstr, OutConstr) :-
    !,
    member(E2,T2), 
    type_le_aux(T1, E2, InConstr, OutConstr). 
type_le_aux((N1 .. M1), (N2 .. M2), InConstr, OutConstr) :-
    N2 =< N1,
    M2 >= M1, 
    OutConstr = InConstr.
type_le_aux((A.._), nat, InConstr, InConstr) :- A >= 0, !.
type_le_aux((_.._), int, InConstr, InConstr).
type_le_aux((_.._), num, InConstr, InConstr).
type_le_aux((_.._), atomic, InConstr, InConstr).
type_le_aux(nat, int, InConstr, InConstr).
type_le_aux(nat, num, InConstr, InConstr).
type_le_aux(int, num, InConstr, InConstr).
type_le_aux(nat, atomic, InConstr, InConstr).
type_le_aux(int, atomic, InConstr, InConstr).
type_le_aux(num, atomic, InConstr, InConstr).
type_le_aux(atom, atomic, InConstr, InConstr).
type_le_aux(string, atomic, InConstr, InConstr).
type_le_aux(bottom, _, InConstr, InConstr).
type_le_aux(T, term, InConstr, InConstr) :- T \= code.
%type_le_aux('$set_type'(T1), [T2], InConstr, OutConstr) :-
%    type_le_aux(T1, T2, InConstr, OutConstr).
type_le_aux('$list_type'(T1), '$list_type'(T2), InConstr, OutConstr) :-
    type_le_aux(T1, T2, InConstr, OutConstr).
type_le_aux('$set_type'(T1), '$set_type'(T2), InConstr, OutConstr) :-
    type_le_aux(T1, T2, InConstr, OutConstr).


type_le_aux('<='(T1), '<='(T2), InConstr, OutConstr) :-
    !,
    '$tuple_type_to_list'(T1, Types1),
    '$tuple_type_to_list'(T2, Types2),
    '$add_annotations'(Types1, ATypes1, '!'),
    '$add_annotations'(Types2, ATypes2, '!'),
    type_annotated_le(ATypes1, ATypes2, InConstr, OutConstr).
type_le_aux('~>>'(T1), '~>>'(T2), InConstr, OutConstr) :-
    !,
    '$tuple_type_to_list'(T1, Types1),
    '$tuple_type_to_list'(T2, Types2),
    '$add_annotations'(Types1, ATypes1, '!'),
    '$add_annotations'(Types2, ATypes2, '!'),
    type_annotated_le(ATypes1, ATypes2, InConstr, OutConstr).
type_le_aux('~>'(T1), '~>'(T2), InConstr, OutConstr) :-
    !,
    '$tuple_type_to_list'(T1, Types1),
    '$tuple_type_to_list'(T2, Types2),
    '$add_annotations'(Types1, ATypes1, '!'),
    '$add_annotations'(Types2, ATypes2, '!'),
    type_annotated_le(ATypes1, ATypes2, InConstr, OutConstr).
/*
type_le_aux(percept, term, _InConstr, _OutConstr) :-
    !,fail.
type_le_aux(belief, term, _InConstr, _OutConstr) :-
    !,fail.
type_le_aux(relcall, term, _InConstr, _OutConstr) :-
    !,fail.
type_le_aux(actcall, term, _InConstr, _OutConstr) :-
    !,fail.
*/
type_le_aux(_T1, T2, InConstr, OutConstr) :-
    T2 = top, !,
    OutConstr = InConstr.
%% Poly types
type_le_aux(T1, T2, InConstr, OutConstr) :-
    compound(T1), compound(T2),
    functor(T1, F1, 1), functor(T2, F2, 1),
    atom(F1), atom(F2), F1 = F2, !,
    arg(1, T1, T1A), arg(1, T2, T2A),
    type_le_aux(T1A, T2A, InConstr, OutConstr).

type_le_aux(T1, T2, InConstr, OutConstr) :-
    '$defined_type'(T1, DT1),
    type_le_aux(DT1, T2, InConstr, OutConstr).
type_le_aux(T1, T2, InConstr, OutConstr) :-
    '$defined_type'(T2, DT2),
    type_le_aux(T1, DT2, InConstr, OutConstr).
type_le_aux(T1, T2, InConstr, OutConstr) :-
    '$defined_type'(T1, '$type_macro'(DT1)),
    type_le_aux(DT1, T2, InConstr, OutConstr).
type_le_aux(T1, T2, InConstr, OutConstr) :-
    '$defined_type'(T2, '$type_macro'(DT2)),
    type_le_aux(T1, DT2, InConstr, OutConstr).


type_annotated_le([], [], InConstr, InConstr).
type_annotated_le([T1|T1R], [T2|T2R], InConstr, OutConstr) :-
    type_annotated_le_aux(T1, T2, InConstr, MidConstr),
    type_annotated_le(T1R, T2R, MidConstr, OutConstr).

type_annotated_le_aux('?'(T1A), '!'(T2A), InConstr, OutConstr) :-
    !,
    type_le_aux(T2A, T1A, InConstr, OutConstr).
type_annotated_le_aux('+?'(T1A), '!'(T2A), InConstr, OutConstr) :-
    !,
    type_le_aux(T2A, T1A, InConstr, OutConstr).
type_annotated_le_aux('?'(T1A), '+?'(T2A), InConstr, InConstr) :-
    !,
    T1A = T2A.
type_annotated_le_aux(MT, MT, InConstr, InConstr) :- !.

list_type_le_aux([], [], Constr, Constr).
list_type_le_aux([T1|T1R], [T2|T2R], InConstr, OutConstr) :-
    type_le_aux(T1, T2, InConstr, MidConstr),
    list_type_le_aux(T1R, T2R, MidConstr, OutConstr).

/*
type_le_constructor_types(A, B, _, _) :- errornl(type_le_constructor_types(A, B)),fail.

type_le_constructor_types('!'(T1), _(T2), InConstr, OutConstr) :-
    is_a_constructor_type(T1, Constr, TT1),
    !,
    is_a_constructor_type(T2, Constr, TT2),
    type_le_constructor_types(TT1, TT2, InConstr, OutConstr).
type_le_constructor_types('!'(T1), _(T2), InConstr, OutConstr) :-
    !,
    type_le_aux(T1, T2, InConstr, OutConstr).
type_le_constructor_types(_(T1), '!'(T2), InConstr, OutConstr) :-
    !,fail.
type_le_constructor_types(_(T1), _(T2), InConstr, InConstr) :-
    !,
    type_le_aux(T1, T2, InConstr, OutConstr).
*/

/*
  Don't need - min mode is enough

unify_mode('!', _, M) :- !, M = '!'.
unify_mode(_, '!', M) :- !, M = '!'.
unify_mode('?', _, M) :- !, M = '?'.
unify_mode(_, '?', M) :- !, M = '?'.
unify_mode('+?', _, M) :- !, M = '+?'.
unify_mode(_, '+?', M) :- !, M = '+?'.
unify_mode('^', '^', M) :- !, M = '^'.
*/

compound_arg_mode('!?', '?') :- !.
compound_arg_mode('!??', '??') :- !.
compound_arg_mode(X, X).

min_mode('!', _, X) :- !, X = '!'.
min_mode(_, '!', X) :- !, X = '!'.
min_mode('?', _, X) :- !, X = '?'.
min_mode(_, '?', X) :- !, X = '?'.
min_mode('??', _, X) :- !, X = '??'.
min_mode(_, '??', X) :- !, X = '??'.
min_mode('+?', '+?', X) :- !, X = '+?'.

/*
max_mode('??', _, X) :- !, X = '??'.
max_mode(_, '??', X) :- !, X = '??'.
max_mode('?', _, X) :- !, X = '?'.
max_mode(_, '?', X) :- !, X = '?'.
max_mode('^', _, X) :- !, X = '?'.
max_mode(_, '^', X) :- !, X = '?'.
max_mode('!', '!', '!').
*/

update_mode('??', _, '??') :- !.
update_mode('+?', _, '+?') :- !.
update_mode(_, _, '!').

/*
  get_max_mode(Term, VMT, Mode) :-
    collect_vars(Term, Vars),
    get_max_mode_aux(Vars, VMT, Mode).

get_max_mode_aux([], _, '!').
get_max_mode_aux([V|Vs], VMT, Mode) :-
    errornl(get_max_mode_aux([V|Vs], VMT, Mode)),
    get_max_mode_aux(Vs, VMT, Mode1),
    (
      type_member_of(V:M(_), VMT)
    ->
      errornl(v__00__(V, M, Mode1)),
      max_mode(M, Mode1, Mode)
    ;
      Mode = '??'
    ),errornl(c__(Mode1, Mode)).
*/

%%%%%%%%%%%%%%%%%
/*
check_head_args([], [], VarModeTypes, VarModeTypes).
check_head_args([A|Args], [M(T)|Types], InVarModeTypes, OutVarModeTypes) :-
    (
      check_type(A, M, T, true, InVarModeTypes, MidVarModeTypes, [], _, [], _)
    ->
      true
    ;
       ip_lookup(type_checker_context, RelAct(Clause, Vars)),
       process_error_context(arg_type_error(A, T, relhead_type(RelAct,Vars, Clause)))
    ),
    check_head_args(Args, Types, MidVarModeTypes, OutVarModeTypes).

check_call(Args, MTypes, InVMT, OutVMT) :-
    errornl(check_call(Args, MTypes, InVMT, OutVMT)),
    check_call_args(Args, MTypes, InVMT, [], OutLvmt, [], OutConstr),
    errornl(check_call_args______(Args, MTypes, InVMT, [], OutLvmt, [], OutConstr)),
    bind_type_constraints(OutConstr),
    merge_vmts(OutLvmt, InVMT, OutVMT),
    errornl(exit_check_call(Args, MTypes, InVMT, OutVMT)).

check_call_args([], [], _VMT, Lvmt, Lvmt, Constr, Constr).
check_call_args([A|Args], [M(T)|Types], VMT, InLvmt, OutLvmt,
		InConstr, OutConstr) :-
    (
      is_mode(M, A, VMT)
    ->
      true
    ;
      errornl(vvvvv), process_mode_error(M, A)
    ),
    (
      check_type(A, M, T, false, VMT, OutVMT, InLvmt, MidLvmt,
                 InConstr, MidConstr)
    ->
      true
    ;
     %global_state_lookup('$error_level',EL),
     %EL1 is EL-1,
     \+'$error'(_),
      errornl(vvvvvvv),
      process_type_error(A, T)
      %ip_lookup(type_checker_context, Context),
      %ip_lookup(current_predicate, CP),
      %process_error_context(arg_type_error(A, T, rel_type(CP,Context)))
    ),
    check_call_args(Args, Types, OutVMT, MidLvmt, OutLvmt,
		MidConstr, OutConstr).
*/
check_clause(Head, Body) :-
    copy_term(Body, BodyCopy),
    check_clause(Head, Body, BodyCopy).

/*
check_body([], VMT, VMT).
check_body([B1|B], InVMT, OutVMT) :-
    '$annotated_type_info'(ATypes, B1),
    B1 =.. [_|Args],
    check_call(Args, ATypes, InVMT, MidVMT),
    check_body(B, MidVMT, OutVMT).


check_final_modes([], []).
check_final_modes([_:'!'(_)|Rest], UB) :- !, check_final_modes(Rest, UB).
check_final_modes([_:'??'(_)|Rest], UB) :- !, check_final_modes(Rest, UB).
check_final_modes([X:_(_)|Rest], [X|UB]) :- !,
    check_final_modes(Rest, UB).
%%%%%%%%%%%%%%%




merge_vmts([], VMT, VMT).
merge_vmts([X:M(T)|Rest], InVMT, OutVMT) :-
    merge_vmt_aux(X, M, T, InVMT, MidVMT),
    merge_vmts(Rest, MidVMT, OutVMT).

merge_vmt_aux(X, M1, T1, InVMT, OutVMT) :-
    type_remove_from(X:M(_), InVMT, MidVMT), !,
    (
      M1 = '??', M = '??'
    ->
      OutVMT = [X:M(T1)|MidVMT]
    ;
      M1 = '??', M \= '!'
    ->
      OutVMT = [X:'?'(T1)|MidVMT]
    ;
      OutVMT = [X:'!'(T1)|MidVMT]
    ).
merge_vmt_aux(X, M1, T1, InVMT, OutVMT) :-
    OutVMT = [X:M1(T1)|InVMT].
*/

type_member_of(X:T, [Y:TY|_]) :- X == Y, !, T = TY.
type_member_of(XT, [_|Xs]) :- type_member_of(XT, Xs).

type_remove_from(X:T, [Y:TY|Ys], Z) :- X == Y, !, T = TY, Z = Ys.
type_remove_from(XT, [Y|Ys], [Y|Z]) :- type_remove_from(XT, Ys, Z).

le1_remove_from(le(X, T), [le(Y,TY)|Ys], Z) :- X == Y, !, T = TY, Z = Ys.
le1_remove_from(XT, [Y|Ys], [Y|Z]) :- le1_remove_from(XT, Ys, Z).
le2_remove_from(le(T, X), [le(TY,Y)|Ys], Z) :- X == Y, !, T = TY, Z = Ys.
le2_remove_from(XT, [Y|Ys], [Y|Z]) :- le2_remove_from(XT, Ys, Z).


already_le(T1, T2, Constr) :-
    member_eq(le(T1, T2), Constr), !.
already_le(T1, T2, Constr) :-
    member(le(TA, TB), Constr), TA == T1, TB \== T1, !,
    already_le(TB, T2, Constr).

different_le(le(A, B)) :- A \== B.

/*
merge_le(C, InConstr, OutConstr) :- errornl(merge_le(C, InConstr, OutConstr)),fail.
merge_le(le(T1, T2), InConstr, OutConstr) :-
    already_le(T1, T2, InConstr), !,
    OutConstr = InConstr.
merge_le(le(T1, T2), InConstr, OutConstr) :-
    var(T2), !,
    (
      le2_remove_from(le(T, T2), InConstr, MidConstr)
    ->
      '$simplify_types'([T1, T], SimpT),
      OutConstr0 = [le(SimpT, T2)|MidConstr]
    ;
      OutConstr0 = [le(T1, T2)|InConstr]
    ),
    check_type_consistency(T2, OutConstr0, OutConstr1),
    filter(different_le, OutConstr1, OutConstr).
merge_le(le(T1, T2), InConstr, OutConstr) :-
    var(T1), !,
    (
      member(le(T11, T), InConstr), T11 == T1
    ->
      (
        type_le_aux(T2, T,[],[])
      ->
        OutConstr0 = [le(T1, T2)|InConstr]
      ;
        type_le_aux(T, T2, [], [])
      ->
        OutConstr0 = [le(T1, T)|InConstr]
      ;
        errornl(type_error_XXXXX___(le(T1, T2), InConstr)),
        fail
      )
    ;
      OutConstr0 = [le(T1, T2)|InConstr]
    ),
    check_type_consistency(T1, OutConstr0, OutConstr).
*/      
      
check_type_consistency(T, InConstr, OutConstr) :-
    member(le(TLow, T1), InConstr), T1 == T,
    member(le(T2, THi), InConstr), T2 == T,
    !,
    (
      type_le_aux(TLow, THi, [], []), type_le_aux(THi, TLow, [], [])
    ->
      %% bounds are equal
      le1_remove_from(le(TLow, T), InConstr, MidConstr),
      le2_remove_from(le(T, THi), MidConstr, OutConstr),
      T = TLow
    ;
      type_le_aux(TLow, THi, [], [])
    ->
      %% consistent constraints
      OutConstr = InConstr
    ;
     ip_lookup(check_var_type_var, Var),
     process_inconsistent_type_error(Var, TLow, THi)
    ).
check_type_consistency(_T, Constr, Constr).

%bind_type_constraints([]) :- !.
%%bind_type_constraints(Constr) :- errornl(bind_type_constraints(Constr)),fail.
bind_type_constraints(Constr) :-
    redo_type_le(Constr, SimpConstr),
    thaw_term(SimpConstr, FVars),
    copy_term((SimpConstr, FVars), (SimpConstrCopy, CFVars)),
    freeze_term(CFVars),
    bind_type_constraints_aux(SimpConstrCopy),
    bind_to_copy_with_simp(SimpConstr, SimpConstrCopy).

%%bind_to_copy_with_simp(A, B) :- errornl(bind_to_copy_with_simp(A, B)),fail.
bind_to_copy_with_simp([], []).
bind_to_copy_with_simp([le(A, B)|Constr], [le(A, D)|ConstrCopy]) :-
    var(B), !,
    '$simplify_the_type'(D, B),
    bind_to_copy_with_simp(Constr, ConstrCopy).
bind_to_copy_with_simp([le(A,_)|Constr], [le(A,_)|ConstrCopy]) :-
    bind_to_copy_with_simp(Constr, ConstrCopy).


%%'$simplify_the_type'(T, V) :- errornl('$simplify_the_type'(T)),fail.
'$simplify_the_type'(V, R) :-
    var(V), !,
    R = V.
'$simplify_the_type'('$union_type'(Types), Type) :-
    !,
    map('$simplify_the_type', Types, SimpTypes),
    sort(SimpTypes, SortedSimpTypes),
    '$collect_simp_types'(SortedSimpTypes, [], SimpSimpTypes),
    /*
    findall(T, (member(T, SortedSimpTypes), 
                   forall((member(T1, SortedSimpTypes), T1 \== T),
                          \+type_le(T, T1))),
            SimpSimpTypes),
      */
    (
      SimpSimpTypes = [T1]
    ->
      Type = T1
    ;
      Type = '$union_type'(SimpSimpTypes)
    ).
'$simplify_the_type'('$list_type'(T), Type) :-
    !,
    '$simplify_the_type'(T, SimpT),
    Type = '$list_type'(SimpT).
'$simplify_the_type'('$set_type'(T), Type) :-
    !,
    '$simplify_the_type'(T, SimpT),
    Type = '$set_type'(SimpT).
'$simplify_the_type'(CType, SimpT) :-
    '$defined_type'(CType, '$constr_enum_type'(_)),
    CType = C(Type), !,
    '$simplify_the_type'(Type, SType),
    SimpT = C(SType).
'$simplify_the_type'(T, T).

'$collect_simp_types'([], _, []).
'$collect_simp_types'([T|Ts], Seen, Simp) :-
    member(T1, Ts), type_le(T, T1), !,
    '$collect_simp_types'(Ts, Seen, Simp).
'$collect_simp_types'([T|Ts], Seen, Simp) :-
    member(T1, Seen), type_le(T, T1), !,
    '$collect_simp_types'(Ts, Seen, Simp).
'$collect_simp_types'([T|Ts], Seen, Simp) :-
    '$collect_simp_types'(Ts, [T|Seen], Simp1),
    Simp = [T|Simp1].

bind_type_constraints_aux([]).
bind_type_constraints_aux([le(T1, T2)|Rest]) :-
    var(T2), !,
    collect_same_2_le(Rest, T2, Collected, Left),
    '$simplify_types'([T1|Collected], T22),
    T2 = T22,
    bind_type_constraints_aux(Left).
bind_type_constraints_aux([le(T, T)|Rest]) :-
    bind_type_constraints_aux(Rest).

collect_same_2_le([], _T2, [], []) :- !.
collect_same_2_le([le(T1, T1x)|Rest], T2, [T1|Collected], Left) :-
    T1x == T2, !,
    collect_same_2_le(Rest, T2, Collected, Left).
collect_same_2_le(Rest, _T2, [], Rest).


get_bang_vars([], []).
get_bang_vars([(X:'!'(_))|Rest], [X|Xs]) :-
    !,
    get_bang_vars(Rest, Xs).
get_bang_vars([_|Rest], Xs) :-
    get_bang_vars(Rest, Xs).

replace_constructors('$list_constr'(_, _), '____') :- !.
replace_constructors('$set_constr'(_, _), '____') :- !.
replace_constructors(X, X).

/*
  is_mode('!', A, VMT) :-
    !,
    transform_subterms(replace_constructors, A, AT),
    get_bang_vars(VMT, Vars),
    freeze_term(Vars),
    '$frozen_term'(AT),
    thaw_term(Vars).
is_mode('^', A, VMT) :-
    !,
    var(A),
    ( type_member_of((A:M(_)), VMT) -> M = '^' ; true ).
is_mode(_, _, _).

set_mode(Mode, X : _(T), X : Mode(T)).

set_bang_modes(VMT, BVMT) :- map(set_mode('!'), VMT, BVMT).

unbound_vars(A, VMT, UB) :-
    get_bang_vars(VMT, Vars),
    freeze_term(Vars),
    freeze_term(A, UB),
    thaw_term((Vars,UB)).
*/

'$kind_operator'(Term, Op) :-
    '$type_info'(_, Kind, _, Term, _), !,
    '$kind_operator_aux'(Kind, Op).
'$kind_operator_aux'(relation, '<=').
'$kind_operator_aux'(action, '~>>').
'$kind_operator_aux'(durative, durative).
'$kind_operator_aux'(discrete, discrete).
    
'$annotated_type_info'(ATypes, Term) :-
    '$type_info'(_, _Kind, Types, Term, _),
    '$add_annotations'(Types, ATypes, '!').

'$add_annotations_to_type'(X, Y) :-
    var(X),
    !,
    X = Y.
'$add_annotations_to_type'(X, Y) :-
    '$has_annotation'(X), !,
    X = Ann(T),
    '$add_annotations_to_type'(T, AT),
    Y = Ann(AT).
'$add_annotations_to_type'('<='(Tuple), '<='(ATuple)) :-
    !,
    '$tuple_type_to_list'(Tuple, Types),
    map('$add_annotations_to_type', Types, ATypes1),
    '$add_annotations'(ATypes1, ATypes, '!'),
    '$list_to_tuple_type'(ATypes, ATuple).
'$add_annotations_to_type'('~>>'(Tuple), '~>>'(ATuple)) :-
    !,
    '$tuple_type_to_list'(Tuple, Types),
    map('$add_annotations_to_type', Types, ATypes1),
    '$add_annotations'(ATypes1, ATypes, '!'),
    '$list_to_tuple_type'(ATypes, ATuple).
'$add_annotations_to_type'('~>'(Tuple), '~>'(ATuple)) :-
    !,
    '$tuple_type_to_list'(Tuple, Types),
    map('$add_annotations_to_type', Types, ATypes1),
    '$add_annotations'(ATypes1, ATypes, '!'),
    '$list_to_tuple_type'(ATypes, ATuple).
'$add_annotations_to_type'((Tuple1 -> Tuple2),
                           (ATuple1 -> ATuple2)) :-
    !,
    '$tuple_type_to_list'(Tuple1, Types1),
    '$tuple_type_to_list'(Tuple2, Types2),
    map('$add_annotations_to_type', Types1, ATypes1),
    map('$add_annotations_to_type', Types2, ATypes2),
    '$list_to_tuple_type'(ATypes1, ATuple1),
    '$list_to_tuple_type'(ATypes2, ATuple2).
'$add_annotations_to_type'(T, T).



extract_append_args(A, Args) :-
    compound(A),
    A = (A1 <> A2),
    !,
    extract_append_args(A1, Args1),
    extract_append_args(A2, Args2),
    append(Args1, Args2, Args).
extract_append_args(A, Args) :-
    Args = [A].

extract_concat_args(A, Args) :-
    compound(A),
    A = (A1 ++ A2),
    !,
    extract_concat_args(A1, Args1),
    extract_concat_args(A2, Args2),
    append(Args1, Args2, Args).
extract_concat_args(A, Args) :-
    Args = [A].

'$check_global_local_vars'([], Globals, Globals, Locals, Locals).
'$check_global_local_vars'([Q|QueryL], Globals, OutGlobals,
			   Locals, OutLocals) :-
    collect_vars(Q, QVars),
    intersect_list(QVars, Locals, IntLocals),
    (
     IntLocals = []
    ->
     true
    ;
     errornl(error_forall_vars_locals(Q, IntLocals)),
     fail
    ),
    '$check_global_local_vars_aux'(Q, Globals, Locals, MidGlobals, MidLocals),
    '$check_global_local_vars'(QueryL, MidGlobals, OutGlobals,
			       MidLocals, OutLocals).

'$check_global_local_vars_aux'('$forall_actions'(V, Q1, Q2), Globals, Locals,
			       MidGlobals, MidLocals) :-
    !,
    '$check_global_local_vars_aux'('$forall'(V, Q1, Q2), Globals, Locals,
				   MidGlobals, MidLocals).
'$check_global_local_vars_aux'('$forall'(V, Q1, Q2), Globals, Locals,
			       MidGlobals, MidLocals) :-
    !,
    intersect_list(V, Globals, IntGlobals),
    (
     IntGlobals = []
    ->
     true
    ;
     errornl(error_forall_vars_quant('$forall'(V, Q1, Q2), IntGlobals)),
     fail
    ),
    '$goals2list'((Q1 & Q2), FGoals),
    union_list(Globals, V, VGlobals),
    '$check_global_local_vars'(FGoals, VGlobals, _, Locals, Mid1Locals),
    union_list(V, Mid1Locals, MidLocals),
    MidGlobals = Globals.
'$check_global_local_vars_aux'(Q, Globals, Locals, MidGlobals, MidLocals) :-
    collect_vars(Q, Vars),
    union_list(Vars, Globals, MidGlobals),
    MidLocals = Locals.
    
'$try_unify_to_empty'([]) :- !.
'$try_unify_to_empty'(_).

ip_lookup_init_empty(A, B) :-
    ip_lookup(A, B),
    '$try_unify_to_empty'(B).

update_check_global_vars(Var) :-
    ip_lookup_init_empty('$local_vars', LocalVars),
    ip_lookup_init_empty('$global_vars', GVars),
    (
      member_eq(Var, LocalVars)
    ->
      process_globals_locals_error(Var)
    ;
      true
    ),
    union_list(GVars, [Var], AllVars),
    ip_set('$global_vars', AllVars).

%%set_modes(A,B,C,D) :- errornl(set_modes(A,B,C,D)),fail.
set_modes([], _, InVMT, InVMT).
set_modes([V|Vars], Mode, InVMT, OutVMT) :-
    type_remove_from(V: _(T), InVMT, MidVMT), !,
    Mid1VMT = [V: Mode(T)|MidVMT],
    set_modes(Vars, Mode, Mid1VMT, OutVMT).
set_modes([V|Vars], Mode, InVMT, OutVMT) :-
    set_modes(Vars, Mode, [V:Mode(term)|InVMT], OutVMT).

'$get_function_args'(Term, Args) :-
    Term =.. [_|AllArgs],
    (
     AllArgs = ['$none_']
    ->
     Args = []
    ;
     Args = AllArgs
    ).

check_ground_inputs_(_Call, InVMT, OutVMT) :-
    InVMT = OutVMT.
     
     
ground_inputs(Call) :-
    (
     var(Call)
    ->
     fail
    ;
     true
    ),
    (
     functor(Call, F, _), var(F)
    ->
     fail
    ;
     true
    ),
    '$type_info'(_, _Kind, Types, Call, _),
    '$add_annotations'(Types, ATypes, '!'),
    Call =.. [_|Args],
    ground_inputs_aux(Args, ATypes), !.

%ground_inputs_aux(A, B) :- errornl(ground_inputs_aux((A, B))),fail.
ground_inputs_aux([], []).
ground_inputs_aux([A|Args], [M(_)|Types]) :-
    (
     ground_inputs_aux2(M, A)
    ->
     true
    ;
     %errornl(arg_mode_error(A, M)),
     fail
    ),
    ground_inputs_aux(Args, Types).

ground_inputs_aux2('!', A) :- !, ground(A).
ground_inputs_aux2(_, _).

/*
get_mode_and_type(M(T), _, M, T) :- !.
get_mode_and_type(T, M, M, T).
*/
is_a_constructor_type(T, Constr, IT) :-
    freeze_term(T, Vars),
    is_a_constructor_type_aux(T, Constr, IT),
    thaw_term(Vars).

is_a_constructor_type_aux('$list_type'(T), '$list_type', T) :- !.
is_a_constructor_type_aux(Type, T1, T) :-
    '$defined_type'(Type, '$constr_enum_type'(_)), !,
    Type = T1(T).

/*
add_constructor_mode(Type, Mode, _CMType) :- errornl(add_constructor_mode(Type, Mode)),fail.
add_constructor_mode(Type, Mode, CMType) :-
    freeze_term(Type, Vars),
    add_constructor_mode_aux(Type, Mode, CMType),
    thaw_term(Type).

add_constructor_mode_aux(Type, Mode, CMType) :-
    is_a_constructor_type_aux(Type, Constr, T),
    no_mode(T), !,
    add_constructor_mode_aux(T, Mode, TM),
    (
      Constr = '$list'
    ->
      CMType = [Mode(TM)]
    ;
      CMType = Constr(Mode(TM))
    ).
add_constructor_mode_aux(Type, Mode, CMType) :-
    is_a_constructor_type(Type, Constr, TM), !,
    TM = M(T),
    add_constructor_mode_aux(T, M, CMT),
    (
      Constr = '$list'
    ->
      CMType = [M(CMT)]
    ;
      CMType = Constr(M(CMT))
    ).
add_constructor_mode_aux(Type, _Mode, Type).
*/

/*
  add_constructor_inner_mode(Type, Mode, CMType) :-
    no_mode(Type), !,
    CMType = Mode(Type).
add_constructor_inner_mode(Type, _Mode, Type).
*/

no_mode(T) :-
    moded(T), !,
    fail.
no_mode(_).

moded(Term) :-
    '$get_atom_functor'(Term, F),
    is_a_mode(F).

is_a_mode('!').
is_a_mode('?').
is_a_mode('+?').
is_a_mode('??').
is_a_mode('!?').

all_bang_mode(F(T)) :-
    atom(F), F = '!',
    (
     is_a_constructor_type(T, _, MT), moded(MT)
    ->
     all_bang_mode(MT)
    ;
     true
    ).

%%update_mode_on_call_exit(X, Y, _) :- errornl(update_mode_on_call_exit(X, Y)),fail.
update_mode_on_call_exit(T1, T2, T3) :-
    freeze_term((T1,T2), Vars),
    update_mode_on_call_exit_aux(T1, T2, T3),
    thaw_term(Vars).

update_mode_on_call_exit_aux('??'(InType), M(_), M(InType)) :- !.
update_mode_on_call_exit_aux('+?'(InType), M(_), M(InType)) :- !.
update_mode_on_call_exit_aux(_(InType), M(Other), '!'(OutType)) :-
    is_a_constructor_type_aux(InType, Constr, T), !,
    is_a_constructor_type_aux(Other, Constr, OT),
    (
      moded(OT)
    ->
      IT = OT
    ;
      IT = M(OT)
    ),
    (
     moded(T)
    ->
     update_mode_on_call_exit_aux(T, IT, OutT)
    ;
     OutT = T
    ),
    add_constructor(Constr, OutT, OutType).
update_mode_on_call_exit_aux(_(InType), _, '!'(OutType)) :-
    OutType = InType.
                        
/*
check_constructor_mode(A,B,C,D,E) :- errornl(check_constructor_mode(A,B)),fail.
check_constructor_mode('!'(Type1), '!'(Type2), Term, MT1, MT2) :-
    is_a_constructor_type(Type1, Constr, T1), !,
    (
      is_a_constructor_type(Type2, Constr, T2)      
    ->
      true
    ;
      process_type_error(Term, MT1, MT2)
    ),
    check_constructor_mode(T1, T2, Term, MT1, MT2).
check_constructor_mode('!'(_Type1), '!'(_Type2), _Term, _MT1, _MT2) :- !.
check_constructor_mode('!'(Type1), _Type2, Term, _MT1, _MT2) :-
    !,
    process_mode_error('!', Term).
check_constructor_mode(_Type1, _Type2, _Term, _MT1, _MT2).
*/

get_inner_mode_and_type(_M(T), IM, Type) :-
    is_a_constructor_type(T, Constr, T1),
    moded(T1), !,
    get_inner_mode_and_type(T1, IM, InnerType),
    add_constructor(Constr, InnerType, Type).
get_inner_mode_and_type(M(T), IM, Type) :-
    is_a_constructor_type(T, Constr, T1), !,
    get_inner_mode_and_type(M(T1), IM, InnerType),
    add_constructor(Constr, InnerType, Type).
get_inner_mode_and_type(M(T), M, T).

%%matched_groundedness(A, B) :- errornl(matched_groundedness(A, B)), fail.
matched_groundedness(A, B) :-
    freeze_term((A,B), Vars),
    matched_groundedness_aux(A, B),
    thaw_term(Vars).

%%matched_groundedness_aux(A, B) :- errornl(matched_groundedness_aux(A, B)), fail.
matched_groundedness_aux('!'(T1), '!'(T2)) :-
    is_a_constructor_type_aux(T1, _, IT1),
    is_a_constructor_type_aux(T2, _, IT2),
    !,
    matched_groundedness_aux(IT1, IT2).
matched_groundedness_aux('!'(_T1), '!'(_T2)) :-
    !.
matched_groundedness_aux('!'(_T1), _(_T2)) :-
    !,
    fail.
matched_groundedness_aux(_, _).

%%check_matching_constructor_types(T1, T2) :- errornl(check_matching_constructor_types(T1, T2)),fail.
check_matching_constructor_types(term, _) :- !.
check_matching_constructor_types(_, term) :- !.
check_matching_constructor_types(T1, T2) :-
    is_a_constructor_type(T1, Constr, IT1), nonvar(T2), !,
    (
      moded(IT1)
    ->
      IT1 = _(IIT1)
    ;
      IIT1 = IT1
    ),
    is_a_constructor_type(T2, Constr, IT2),
    (
      moded(IT2)
    ->
      IT2 = _(IIT2)
    ;
      IIT2 = IT2
    ),   
    check_matching_constructor_types(IIT1, IIT2).
check_matching_constructor_types(T1, T2) :-
    nonvar(T1),
    is_a_constructor_type(T2, _Constr, _IT1),!,fail.
check_matching_constructor_types(_, _).

%%min_mode_constructor_types(Min, T1, T2, MT) :- errornl(min_mode_constructor_types(Min, T1, T2, MT)),fail.
min_mode_constructor_types(Min, T1, T2, MT) :-
    is_a_constructor_type(T1, Constr, IT1),
    is_a_constructor_type(T2, Constr, IT2), !,
    (
     moded(IT1)
    ->
     IT1 = IM1(IIT1)
    ;
     IM1 = Min,
     IIT1 = IT1
    ),
    (
     moded(IT2)
    ->
     IT2 = IM2(IIT2)
    ;
     IM2 = Min,
     IIT2 = IT2
    ),
    min_mode(IM1, IM2, IMin),
    min_mode_constructor_types(IMin, IIT1, IIT2, IMT),
    add_constructor(Constr, IMT, CMT),
    MT = Min(CMT).
min_mode_constructor_types(Min, T1, _T2, Min(T1)).

add_constructor('$list', Term, [Term]) :- !.
add_constructor(Constr, Term, Constr(Term)).

'$is_code'(T) :-
    atom(T),
    '$type_info'(T, Kind, _, _, _),
    ( Kind = function ; Kind = relation ; Kind = action ; Kind = tr_prog ), !.


type_info(Type, Info) :-
    '$type_info'(_, defined_type, Type, TypeDef, _),
    '$type_info_aux'(TypeDef, Info).

'$type_info_aux'('$union_type'(T), unionT_(T)).
'$type_info_aux'('$enum_type'(T), enumT_(T)).
'$type_info_aux'('$tuple_type'(T), tupleT_(T)).
'$type_info_aux'('$constr_enum_type'(T), constrT_(T)).
'$type_info_aux'((N..M), rangeT_(N, M)).